// -*- mode: C++; tab-width: 4; indent-tabs-mode: nil -*- (for GNU Emacs)
//
// Copyright (c) 1985-1998 Microsoft Corporation
//
// This file is part of the Microsoft Research IPv6 Network Protocol Stack.
// You should have received a copy of the Microsoft End-User License Agreement
// for this software along with this release; see the file "license.txt".
// If not, please see http://www.research.microsoft.com/msripv6/license.htm,
// or write to Microsoft Research, One Microsoft Way, Redmond, WA 98052-6399.
//
// Abstract:
//
// Definitions derived from the IPv6 specifications.
//


#ifndef IP6_INCLUDED
#define IP6_INCLUDED 1

//
// Configuration option controlling whether the implementation
// supports the post-RFC1884 addressing scheme (interface token formation,
// solicted-node multicast address formation).
//
// REVIEW: We should just quit supporting the old addressing at some point.
//
#define IPV6_NEW_ADDRESSING 1

#if IPV6_NEW_ADDRESSING
#define INTERFACE_IDENTIFIER_LENGTH     64
#else
#define INTERFACE_IDENTIFIER_LENGTH     48
#endif


//
// Basic types.
// REVIEW: These really belong somewhere else.
//
typedef unsigned char uchar;
typedef unsigned short ushort;
typedef unsigned long ulong;
typedef unsigned int uint;


//
// An IPv6 address is 128 bits long.
//
struct in6_addr {
    union {
        uchar Byte[16];
        ushort Word[8];
        ulong DWord[4];
    } u;
};

#define s6_bytes   u.Byte
#define s6_words   u.Word
#define s6_dwords  u.DWord
#define s6_addr    s6_bytes  // The only member defined in the draft.


//
// IPv6 Header Format.
// See RFC 1883, page 5 (and subsequent draft updates to same).
//
typedef struct IPv6Header {
    ulong VersClassFlow;   // 4 bits Version, 8 Traffic Class, 20 Flow Label.
    ushort PayloadLength;  // Zero indicates Jumbo Payload hop-by-hop option.
    uchar NextHeader;      // Values are superset of IPv4's Protocol field.
    uchar HopLimit;
    struct in6_addr Source;
    struct in6_addr Dest;
} IPv6Header;

//
// Maximum value for the PayloadLength field.
// Note that the 40-byte IPv6 header is NOT included.
//
#define MAX_IPv6_PAYLOAD        65535

//
// Useful constants for working with various fields in the IPv6 header.
//
// NOTE: We keep the Version, Traffic Class and Flow Label fields as a single
// NOTE: 32 bit value (VersClassFlow) in network byte order (big-endian).
// NOTE: Since NT is little-endian, this means all loads/stores to/from this
// NOTE: field need to be byte swapped.
//
#define IP_VER_MASK 0x000000F0  // Version is high 4 bits of VersClassFlow.
#define IP_VERSION 0x00000060   // This is 6 << 28 (after byte swap).
#define IP_TRAFFIC_CLASS_MASK 0x0000F00F  // 0x0FF00000 (after byte swap).

#define MAX_IP_PROTOCOL 256


//
// IPv6 type-length-value (TLV) encoded option types found in some
// extension headers.  The upper two bits of each type are encoded
// so as to specify what action the node should take if it doesn't
// grok the option type.  The third-highest-order bit specifies if
// the option data can change en-route to the final destination.
// See RFC 1883, section 4.2 (pages 9-10) for more information.
//
#define IPv6_OPT_ACTION_MASK 0xc0   // High two bits
#define IPv6_OPT_DYNDATA_MASK 0x20  // Third-highest bit

#define IPv6_OPT_PAD1 0     // Special case -- no length or value fields.
#define IPv6_OPT_PADN 1     // Multi-byte pad, data (if any) is always zero.
#define IPv6_OPT_JUMBO 194  // Hop-by-Hop Jumbo Payload.


//
// Hop-by-Hop Options Header.
// 
typedef struct IPv6HopByHopHeader {
    uchar NextHeader;
    uchar HeaderExtLength;  // in 8-byte units, not counting first 8.
} IPv6HopByHopHeader;


//
// Destination Options Header.
//
typedef struct IPv6DestOptionsHeader{
    uchar NextHeader;
    uchar HeaderExtLength;  // in 8-byte units, not counting first 8.
} IPv6DestOptionsHeader;


//
// Rounting Header (RFC 1883, page 13).
//
typedef struct IPv6RoutingHeader {
    uchar NextHeader;
    uchar HeaderExtLength;  // in 8-byte units, not counting first 8.
    uchar RoutingType;
    uchar SegmentsLeft;     // Number of nodes still left to be visited.
} IPv6RoutingHeader;

#endif // IP6_INCLUDED
