// -*- mode: C++; tab-width: 4; indent-tabs-mode: nil -*- (for GNU Emacs)
//
// Copyright (c) 1998 Microsoft Corporation
//
// This file is part of the Microsoft Research IPv6 Network Protocol Stack.
// You should have received a copy of the Microsoft End-User License Agreement
// for this software along with this release; see the file "license.txt".
// If not, please see http://www.research.microsoft.com/msripv6/license.htm,
// or write to Microsoft Research, One Microsoft Way, Redmond, WA 98052-6399.
//
// Abstract:
//
// Support for running IPv6 over a multicast IPv4 network,
// using Neighbor Discovery.
//


#include "oscfg.h"
#include "ndis.h"
#include "ip6imp.h"
#include "ip6def.h"
#include "llip6if.h"
#include "tdi.h"
#include "tdikrnl.h"
#include "tunnel.h"

//* TunnelV6V4CreateToken
//
//  Given a link-layer address, creates a 64-bit "interface token"
//  in the low eight bytes of an IPv6 address.
//  Does not modify the other bytes in the IPv6 address.
//
void
TunnelV6V4CreateToken(
    void *Context,
    void *LinkAddress,
    IPv6Addr *Address)
{
    uchar *IPAddress = (uchar *)LinkAddress;

#if IPV6_NEW_ADDRESSING
    Address->u.Byte[8] = 0;
    Address->u.Byte[9] = 0;
#endif
    Address->u.Byte[10] = 0;
    Address->u.Byte[11] = 0;
    Address->u.Byte[12] = IPAddress[0];
    Address->u.Byte[13] = IPAddress[1];
    Address->u.Byte[14] = IPAddress[2];
    Address->u.Byte[15] = IPAddress[3];
}

//* TunnelV6V4ReadLinkLayerAddressOption
//
//  Parses a Neighbor Discovery link-layer address option
//  and if valid, returns a pointer to the link-layer address.
//
void *
TunnelV6V4ReadLinkLayerAddressOption(
    void *Context,
    uchar *OptionData)
{
    //
    // Check that the option length is correct.
    //
    if (OptionData[1] != 1)
        return NULL;

    //
    // Check the must-be-zero padding bytes.
    //
    if ((OptionData[2] != 0) || (OptionData[3] != 0))
        return NULL;

    //
    // Return a pointer to the embedded IPv4 address.
    //
    return OptionData + 4;
}

//* TunnelV6V4WriteLinkLayerAddressOption
//
//  Creates a Neighbor Discovery link-layer address option.
//  Our caller takes care of the option type & length fields.
//  We handle the padding/alignment/placement of the link address
//  into the option data.
//
//  (Our caller allocates space for the option by adding 2 to the
//  link address length and rounding up to a multiple of 8.)
//
void
TunnelV6V4WriteLinkLayerAddressOption(
    void *Context,
    uchar *OptionData,
    void *LinkAddress)
{
    uchar *IPAddress = (uchar *)LinkAddress;

    //
    // Place the address after the option type/length bytes
    // and two bytes of zero padding.
    //
    OptionData[2] = 0;
    OptionData[3] = 0;
    OptionData[4] = IPAddress[0];
    OptionData[5] = IPAddress[1];
    OptionData[6] = IPAddress[2];
    OptionData[7] = IPAddress[3];
}

//* TunnelV6V4ConvertMulticastAddress
//
//  Converts an IPv6 multicast address to a link-layer address.
//  Generally this requires hashing the IPv6 address into a set
//  of link-layer addresses, in a link-layer-specific way.
//
void
TunnelV6V4ConvertMulticastAddress(
    void *Context,
    IPv6Addr *Address,
    void *LinkAddress)
{
    uchar *IPAddress = (uchar *)LinkAddress;

#if IPV6_NEW_ADDRESSING
    IPAddress[0] = 239;
    IPAddress[1] = 192; // or 128 or 64??
    IPAddress[2] = Address->u.Byte[14];
    IPAddress[3] = Address->u.Byte[15];
#else
    IPAddress[0] = 231;
    IPAddress[1] = Address->u.Byte[13];
    IPAddress[2] = Address->u.Byte[14];
    IPAddress[3] = Address->u.Byte[15];
#endif
}

//* TunnelV6V4SetMulticastAddressList
//
//  Takes an array of link-layer multicast addresses
//  (from TunnelV6V4ConvertMulticastAddress) from which we should
//  receive packets. Passes them to NDIS.
//
NDIS_STATUS
TunnelV6V4SetMulticastAddressList(
    void *Context,
    void *LinkAddresses,
    uint NumKeep,
    uint NumAdd,
    uint NumDel)
{
    TunnelContext *tc = (TunnelContext *)Context;
    IPAddr *Addresses = (IPAddr *)LinkAddresses;
    NTSTATUS status;
    uint i;

    //
    // We add the multicast addresses to Tunnel.Pseudo.AOFile,
    // instead of tc->AOFile, because we are only receiving
    // from the first address object.
    //
    for (i = 0; i < NumAdd; i++) {
        status = TunnelAddMulticastAddress(Tunnel.Pseudo.AOFile,
                                           tc->Addr,
                                           Addresses[NumKeep + i]);
        if (!NT_SUCCESS(status))
            return (NDIS_STATUS)status;
    }

    for (i = 0; i < NumDel; i++) {
        status = TunnelDelMulticastAddress(Tunnel.Pseudo.AOFile,
                                           tc->Addr,
                                           Addresses[NumKeep + NumAdd + i]);
        if (!NT_SUCCESS(status))
            return (NDIS_STATUS)status;
    }

    return NDIS_STATUS_SUCCESS;
}

//* TunnelV6V4CreateInterface
//
//  Creates a virtual 6-over-4 interface for a specific IPv4 address.
//
Interface *
TunnelV6V4CreateInterface(TunnelAddress *ta)
{
    LLIPBindInfo BindInfo;      // Holds info for IPv6.

    BindInfo.lip_context = &ta->V6V4;
    BindInfo.lip_mtu = TUNNEL_MTU;
    // We do not want IPv6 to reserve space for our link-level header.
    BindInfo.lip_hdrsize = 0;
    BindInfo.lip_addrlen = sizeof(IPAddr);
    BindInfo.lip_addr = (uchar *) &ta->V6V4.Addr;

    BindInfo.lip_transmit = TunnelTransmit;
    BindInfo.lip_token = TunnelV6V4CreateToken;
    BindInfo.lip_rdllopt = TunnelV6V4ReadLinkLayerAddressOption;
    BindInfo.lip_wrllopt = TunnelV6V4WriteLinkLayerAddressOption;
    BindInfo.lip_mcaddr = TunnelV6V4ConvertMulticastAddress;
    BindInfo.lip_mclist = TunnelV6V4SetMulticastAddressList;

    return CreateInterface(&BindInfo);
}
