// -*- mode: C++; tab-width: 4; indent-tabs-mode: nil -*- (for GNU Emacs)
//
// Copyright (c) 1985-1998 Microsoft Corporation
//
// This file is part of the Microsoft Research IPv6 Network Protocol Stack.
// You should have received a copy of the Microsoft End-User License Agreement
// for this software along with this release; see the file "license.txt".
// If not, please see http://www.research.microsoft.com/msripv6/license.htm,
// or write to Microsoft Research, One Microsoft Way, Redmond, WA 98052-6399.
//
// Abstract:
//
// IPv6 private definitions.
//
// This file contains all of the definitions for IPv6 that
// are not visible to outside layers.
//


#ifndef IPv6DEF_INCLUDED
#define IPv6DEF_INCLUDED 1

//
// Per-neighbor information.  We keep address translation and unreachability
// detection info for each of our neighbors that we're in communication with.
//
// A non-zero reference count prevents the NCE from being reclaimed.
// An NCE with zero references may be kept cached.
// A per-interface lock protects all NCEs for that interface.
//
// The Next/Prev fields link NCEs into a circular doubly-linked list.
// They must be first and must match the IF->FirstNCE/LastNCE fields
// to make the casting work out.
//
// The list of NCEs is kept sorted, from most-recently-used to least.
//
typedef struct NeighborCacheEntry {   // a.k.a. NCE
    struct NeighborCacheEntry *Next;  // Next entry on I/F neighbor list.
    struct NeighborCacheEntry *Prev;  // Previous entry on I/F neighbor list.
    IPv6Addr NeighborAddress;         // Address of I/F on neighboring node.
    void *LinkAddress;                // Media address corresponding to above.
    // NB: LinkAddressLength field not needed - use IF->LinkAddressLength.
    ushort IsRouter;                  // Non-zero if neighbor is a router.
    ushort NDState;                   // Neighbor Discovery Protocol state.
    LARGE_INTEGER LastReachable;      // Timestamp (system ticks since boot).
    ushort EventTimer;                // In IPv6Timer ticks (see IPv6Timeout).
    ushort EventCount;                // Count number of various timeouts.
    struct Interface *IF;             // Interface on media with neighbor.
    NDIS_PACKET *WaitQueue;           // Queue of packets waiting on ND.
    long RefCnt;                      // Reference count.
    long Validate;                    // RCE validation counter.
} NeighborCacheEntry;

__inline void
AddRefNCE(NeighborCacheEntry *NCE)
{
    InterlockedIncrement(&NCE->RefCnt);
}

__inline void
ReleaseNCE(NeighborCacheEntry *NCE)
{
    InterlockedDecrement(&NCE->RefCnt);
}

//
// Forces the route cache to redo next-hop determination
// for any RCEs that are using this NCE as the next hop.
//
// Called in two situations: the neighbor is no longer reachable, or
// the neighbor is no longer a router. When used in the latter case,
// we actually invalidate all RCEs using this neighbor, not just those
// RCEs that are using this neighbor as a router.
//
// This is generally not harmful, because next-hop determination will
// just decide again that the neighbor is on-link. However, in the
// rare case that the RCE results from a Redirect to the on-link
// neighbor and the neighbor was also a router and no longer is,
// although the RCE wasn't using the neighbor as a router it will
// still be invalidated and another Redirect will be necessary to
// restore communication.  The network-visible effect is the same as
// if the RCE had been purged for other reasons.
//
// If this should become a performance problem, it is possible to fix
// it by looking up NCE->NeighborAddress in the route cache, finding any
// RCEs with RCE->Destination == NCE->NeighborAddress, and updating
// RCE->ValidateNCE.
//
__inline void
InvalidateRouteCacheForNCE(NeighborCacheEntry *NCE)
{
    InterlockedIncrement(&NCE->Validate);
}

//
// Values for "NDState" above.  See RFC 1970, section 7.3.2 for details.
// Note: only state names are documented, we chose the values used here.
// The values are not arbitrary - a smaller value indicates that the
// LinkAddress is less likely to be reachable. Router selection
// uses this comparison.
//
// In the INCOMPLETE state, the LinkAddress is not valid.
// In all other states, LinkAddress may be used to send packets.
// WaitQueue can only be non-NULL in the INCOMPLETE state,
// but it is always initialized to NULL in the other states.
// EventTimer is also initialized in all states,
// but EventCount may not be.
//
// The INCOMPLETE state has two flavors, dormant and active. If
// EventTimer and EventCount are both zero, then we are not actively
// trying to solicit the link address. If someone tries to send to
// this neighbor, then we start soliciting the link address. If the
// solicitation fails (or if we enter the PROBE state and then fail to
// confirm reachability), then any waiting packets are discarded and
// we reset to INCOMPLETE with zero EventTimer/EventCount. (So with
// the next use of this neighbor, we start soliciting again from scratch.)
//
#define ND_STATE_INCOMPLETE 0
#define ND_STATE_PROBE      1
#define ND_STATE_DELAY      2
#define ND_STATE_STALE      3
#define ND_STATE_REACHABLE  4
#define ND_STATE_PERMANENT  5


//
// Per-net information.  There is one NTE for each source (unicast) address
// assigned to an interface.
//
typedef struct NetTableEntry {        // a.k.a. NTE
    struct NetTableEntry *NextOnIF;   // Next NTE on this interface.
    struct NetTableEntry *NextOnNTL;  // Next NTE on NetTableList.
    IPv6Addr Address;                 // IP address (unicast!) for this net.
    uint PrefixLength;                // Length of address prefix.
    struct Interface *IF;             // Pointer to interface on this net.
    long RefCnt;                      // Reference count.
    RouteCacheEntry *LoopRCE;         // Makes loop-back work.
    ushort DADState;                  // Address configuration state.
    ushort DADCount;                  // How many DAD solicits left to send.
    ushort DADCountLB;                // Loopback prevention count.
    ushort DADTimer;                  // In IPv6Timer ticks (see IPv6Timeout).
    uint ValidLifetime;               // In IPv6Timer ticks (see IPv6Timeout).
    uint PreferredLifetime;           // In IPv6Timer ticks (see IPv6Timeout).
} NetTableEntry;

__inline void
AddRefNTE(NetTableEntry *NTE)
{
    InterlockedIncrement(&NTE->RefCnt);
}

__inline void
ReleaseNTE(NetTableEntry *NTE)
{
    InterlockedDecrement(&NTE->RefCnt);
}

//
// Values for DADState above.
//
#define DAD_STATE_TENTATIVE  0
#define DAD_STATE_PREFERRED  1
#define DAD_STATE_DEPRECATED 2
#define DAD_STATE_INVALID    3
#define DAD_STATE_DUPLICATE  4

__inline int
IsValidNTE(NetTableEntry *NTE)
{
    return ((NTE->DADState == DAD_STATE_PREFERRED) ||
            (NTE->DADState == DAD_STATE_DEPRECATED));
}

#define DAD_LIFETIME_INFINITE 0xffffffff

//
// Note that on the wire, there is no infinite value
// for Neighbor Discovery router lifetimes.
// However internally we need an infinite value.
//
#define ROUTER_LIFETIME_INFINITE 0xffffffff


//
// Local address information.  Maps unicast, multicast, and anycast addresses
// to their corresponding NTE.
//
typedef struct AddressEntry {    // a.k.a. ADE
    struct AddressEntry *Next;   // Linkage on chain.
    IPv6Addr Address;            // Address identifying this entry.
    ushort Type;                 // Address type (unicast, multicast, etc).
    ushort Scope;                // Address scope (link, site, global, etc).
    NetTableEntry *NTE;          // NTE associated with this address.
    uchar MCastRefCount;         // # of user level connections to this group.
    uchar MCastFlag;             // Necessary info about a group.
    ushort MCastTimer;           // Ticks until a membership report is sent.
    struct AddressEntry *NextQL; // QL entry for this group if timer running.
} AddressEntry;

//
// Values for address Type.
//
#define ADE_UNICAST   0x01
#define ADE_ANYCAST   0x02
#define ADE_MULTICAST 0x04

//
// Values for address Scope.
//
#define ADE_RESERVED   0x00
#define ADE_NODE_LOCAL 0x01
#define ADE_LINK_LOCAL 0x02
#define ADE_SITE_LOCAL 0x04
#define ADE_ORG_LOCAL  0x08
#define ADE_GLOBAL     0x10

//
// Values for MCastFlag.
//
#define ADE_USER_MULTICAST_GROUP    0x01  // Groups not added at stack init.
#define ADE_LAST_MULTICAST_REPORTER 0x02    
#define ADE_REPORTABLE_ADDRESS      0x04
#define ADE_USER_REPORTABLE_ADDRESS 0x05  // Both USER and REPORTABLE.

typedef struct LinkLayerMulticastAddress {
    uint RefCnt;
    uchar LinkAddress[];        // The link-layer address follows in memory.
} LinkLayerMulticastAddress;


//
// Information about IPv6 interfaces.  There can be multiple NTEs for each
// interface, but there is exactly one interface per NTE.
//
typedef struct Interface {         // a.k.a. IF
    struct Interface *Next;        // Next interface on chain.

    //
    // Interface to the link layer. The functions all take
    // the LinkContext as their first argument. See comments
    // in llip6if.h.
    //
    void *LinkContext;             // Link layer context.
    void (*CreateToken)(void *Context, void *LinkAddress, IPv6Addr *Address);
    void *(*ReadLLOpt)(void *Context, uchar *OptionData);
    void (*WriteLLOpt)(void *Context, uchar *OptionData, void *LinkAddress);
    void (*ConvertMCastAddr)(void *Context,
                             IPv6Addr *Address, void *LinkAddress);
    void (*Transmit)(void *Context, PNDIS_PACKET Packet,
                     uint Offset, void *LinkAddress);
    NDIS_STATUS (*SetMCastAddrList)(void *Context, void *LinkAddresses,
                                    uint NumKeep, uint NumAdd, uint NumDel);

    uint Index;                    // Node unique index of this I/F.
    uint NTECount;                 // Number of valid NTEs on this I/F.
    NetTableEntry *NTE;            // List of NTEs on this I/F.
    AddressEntry *ADE;             // List of ADEs on this I/F.
    NeighborCacheEntry *FirstNCE;  // List of active neighbors on I/F.
    NeighborCacheEntry *LastNCE;   // Last NCE in the list.
    int DisableND;                 // Is Neighbor Discovery allowed?
    uint NCECount;                 // Number of NCEs in the list.
    uint TrueLinkMTU;              // True maximum MTU for this I/F.
    uint LinkMTU;                  // May be smaller than TrueLinkMTU.
    uint CurHopLimit;              // Default Hop Limit for unicast.
    uint BaseReachableTime;        // Base for random ReachableTime (in ms).
    ULONGLONG ReachableTime;       // Can assume reachable (in system ticks).
    uint RetransTimer;             // NS timeout (in IPv6Timer ticks).
    uint DupAddrDetectTransmits;   // Number of solicits during DAD.
    uint RSCount;                  // Number of Router Solicits sent.
    uint RSTimer;                  // RS timeout (in IPv6Timer ticks).
    long Validate;                 // RCE validation counter. 
    uint LinkAddressLength;        // Length of I/F link-level address.
    uchar *LinkAddress;            // Pointer to link-level address.
    uint LinkHeaderSize;           // Length of link-level header.
    KSPIN_LOCK Lock;               // Interface lock.

    int MCastSynchNeeded;          // Is SetMCastAddrList needed?
    KMUTEX MCastLock;              // Serializes SetMCastAddrList operations
    LinkLayerMulticastAddress *MCastAddresses;  // Current addresses
    uint MCastAddrNum;             // Number of link-layer mcast addresses
    uint MCastAddrNew;             // Number added since SetMCastAddrList
    int LoopbackCapable;           // Can we disable loopback of this IF.
} Interface;

#define SentinelNCE(IF) ((NeighborCacheEntry *)&(IF)->FirstNCE)

//
// Forces the route cache to redo source address (NTE) selection
// for any RCEs that are using this interface.
//
// Called when an NTE on the interface is created, destroyed,
// or changes state in a relevant way.
//
__inline void
InvalidateRouteCacheForNTE(Interface *IF)
{
    InterlockedIncrement(&IF->Validate);
}


//
// We have a periodic timer (IPv6Timer) that causes our IPv6Timeout
// routine to be called IPv6_TICKS_SECOND times per second.  Most of the
// timers and timeouts in this implementation are driven off this routine.
//
// There is a trade-off here between timer granularity/resolution
// and overhead.  The resolution should be subsecond because
// RETRANS_TIMER is only one second.
//
#define IPv6_TICKS_SECOND 2  // Two ticks per second.

#define IPv6_TIMEOUT (1000 / IPv6_TICKS_SECOND)  // In milliseconds.

#define IPv6TimerTicks(seconds) ((seconds) * IPv6_TICKS_SECOND)

extern uint
ConvertSecondsToTicks(uint Seconds);

extern uint
ConvertMillisToTicks(uint Millis);


//
// Various constants from the IPv6 RFCs...
//
// REVIEW: Some of these should be per link-layer type.
// REVIEW: Put them in the Interface structure?
//
#define MAX_RTR_SOLICITATION_DELAY BUGBUG:Unused
#define RTR_SOLICITATION_INTERVAL  IPv6TimerTicks(4)  // 4 seconds.
#define MAX_RTR_SOLICITATIONS      3
#define MAX_MULTICAST_SOLICIT      3  // Total transmissions before giving up.
#define MAX_UNICAST_SOLICIT        3  // Total transmissions before giving up.
#define MAX_ANYCAST_DELAY_TIME     BUGBUG:Unused
#define REACHABLE_TIME             (30 * 1000)  // 30 seconds in milliseconds
#define ICMP_ERROR_SEND_INTERVAL   100  // milliseconds.
#define ICMP_INFO_SEND_INTERVAL    500  // milliseconds.
#define RETRANS_TIMER              IPv6TimerTicks(1)  // 1 second.
#define DELAY_FIRST_PROBE_TIME     IPv6TimerTicks(5)  // 5 seconds.
#define MIN_RANDOM_FACTOR          50   // Percentage of base value.
#define MAX_RANDOM_FACTOR          150  // Percentage of base value.

#define IPv6_MINIMUM_MTU           1280
#define RECALC_REACHABLE_INTERVAL  IPv6TimerTicks(3 * 60 * 60)  // 3 hours.


//
// Support for tagging memory allocations.
//
#ifdef POOL_TAGGING

#ifdef ExAllocatePool
#undef ExAllocatePool
#endif

#define ExAllocatePool(type, size) ExAllocatePoolWithTag(type, size, '6vPI')

#endif // POOL_TAGGING


//
// Under NT, we use the assembly language version of the common core checksum
// routine instead of the C language version.
//
ULONG
tcpxsum(IN ULONG Checksum, IN PUCHAR Source, IN ULONG Length);

#define Cksum(Buffer, Length) ((ushort)tcpxsum(0, (PUCHAR)(Buffer), (Length)))


//
// Protocol (i.e. "Next Header" field) values for included protocols.
//
#define IP_PROTOCOL_HOP_BY_HOP 0  // IPv6 Hop-by-Hop Options Header.
#define IP_PROTOCOL_ROUTING   43  // IPv6 Routing Header.
#define IP_PROTOCOL_FRAGMENT  44  // IPv6 Fragment Header.
#define IP_PROTOCOL_ESP       50  // IPSec Encapsulating Security Payload Hdr.
#define IP_PROTOCOL_AH        51  // IPSec Authentication Hdr.
#define IP_PROTOCOL_ICMPv6    58  // IPv6 Internet Control Message Protocol.
#define IP_PROTOCOL_NONE      59  // No next header - ignore packet remainder.
#define IP_PROTOCOL_DEST_OPTS 60  // IPv6 Destination Options Header.
//
// More to be added when the values are known.
//
#define IP_PROTOCOL_BINDING_UPDATE ??  // BUGBUG: Need value.
#define IP_PROTOCOL_HOME_AGENT ??      // BUGBUG: Need value.

#define IS_EXTENSION_HEADER(x) ((x == IP_PROTOCOL_HOP_BY_HOP) || \
                                (x == IP_PROTOCOL_ROUTING) || \
                                (x == IP_PROTOCOL_FRAGMENT) || \
                                (x == IP_PROTOCOL_DEST_OPTS) || \
                                (x == IP_PROTOCOL_ESP) || \
                                (x == IP_PROTOCOL_AH)) 

//
// Protocol Receive Procedures ("Next Header" handlers) have this prototype.
//
typedef int ProtoRecvProc(NetTableEntry *RecvNTE, IPv6Packet *IP);
typedef int ProtoControlRecvProc(NetTableEntry *RecvNTE, IPv6Packet *IP, 
                                 StatusArg *Arg);

typedef struct ProtocolSwitch {
  ProtoRecvProc *DataReceive;
  ProtoControlRecvProc *ControlReceive;
} ProtocolSwitch;

extern ProtoRecvProc ICMPv6Receive;
extern ProtoRecvProc FragmentReceive;
extern ProtoRecvProc DestinationOptionsReceive;
extern ProtoRecvProc HopByHopOptionsReceive;
// extern ProtoRecvProc EncapsulatingSecurityPayloadReceive;
// extern ProtoRecvProc AuthenticationHeaderReceive;

extern ProtoControlRecvProc ICMPv6ControlReceive;
extern ProtoControlRecvProc ExtHdrControlReceive;


//
// Upper Layer Status Procedures have this prototype.
//
typedef uint (*ULStatusProc)(uchar, IP_STATUS, IPv6Addr, IPv6Addr, IPv6Addr,
                             ulong, void *);


//
// The actual definition of a reassembly structure
// can be found in fragment.h.
//
typedef struct Reassembly Reassembly;


//
// Prototypes for global variables.
//
extern ULONG RouterConfigured;  // Non-zero if we're a router.
extern uchar StrongHost;  // Non-zero if config'd for "Strong Host Model".
extern ULONG DefaultCurHopLimit;
extern ProtocolSwitch ProtocolSwitchTable[];
extern KSPIN_LOCK NetTableListLock;
extern NetTableEntry *NetTableList;  // Pointer to the net table.
extern uint NumNTE;
extern Interface *IFList;  // List of all interfaces on the system.
extern uint NumIF;
extern uint MaxLinkLevelHeader;
extern struct EchoControl *ICMPv6OutstandingEchos;
extern LIST_ENTRY PendingEchoList;  // def needed for initialization.
extern Interface LoopInterface;
extern NeighborCacheEntry *LoopNCE;
extern NetTableEntry *LoopNTE;
extern IPv6Addr UnspecifiedAddr;
extern IPv6Addr LoopbackAddr;
extern IPv6Addr AllNodesOnLinkAddr;
extern IPv6Addr AllRoutersOnLinkAddr;
extern AddressEntry *QueryList;    // group addreses with timers running.
extern Reassembly *ReassemblyList;  // Fragments with timers running.


//
// Some handy inline functions for determining IPv6 address types.
//

__inline int
IsUnspecified(IPv6Addr *Addr)
{
    return IP6_ADDR_EQUAL(Addr, &UnspecifiedAddr);
}

__inline int
IsMulticast(IPv6Addr *Addr)
{
    return Addr->u.Byte[0] == 0xff;
}

__inline int
IsLinkLocal(IPv6Addr *Addr)
{
    return ((Addr->u.Byte[0] == 0xfe) &&
            ((Addr->u.Byte[1] & 0xc0) == 0x80));
}

__inline int
IsLinkLocalMulticast(IPv6Addr *Addr)
{
    return IsMulticast(Addr) && ((Addr->u.Byte[1] & 0xf) <= 2);
}

__inline int
IsSiteLocal(IPv6Addr *Addr)
{
    return ((Addr->u.Byte[0] == 0xfe) &&
            ((Addr->u.Byte[1] & 0xc0) == 0xc0));
}


//
// Function prototypes.
//

extern ulong
Random(void);

extern int
IsV4Compatible(IPv6Addr *Addr);

extern int
HasPrefix(IPv6Addr *Addr, IPv6Addr *Prefix, uint PrefixLength);

extern uint
CommonPrefixLength(IPv6Addr *Addr, IPv6Addr *Addr2);

extern int
ParseNumber(WCHAR *S, uint Base, uint MaxNumber, uint *Number);

extern int
ParseV6Address(WCHAR *Sz, WCHAR **Terminator, IPv6Addr *Addr);

extern void
FormatV6AddressWorker(WCHAR *Sz, IPv6Addr *Addr);

extern WCHAR *
FormatV6Address(IPv6Addr *Addr);

extern void
CopyFromBufferChain(
    void *DstBuffer,
    PNDIS_BUFFER SrcBuffer,
    uint SrcOffset,
    uint Length,
    PNDIS_BUFFER *NextBuffer,
    uint *NextOffset);

extern ushort
ChecksumPacket(PNDIS_PACKET Packet, uint Offset, uchar *Data, uint Length,
               IPv6Header *IP, uchar NextHeader);

extern void
CreateSolicitedNodeMulticastAddress(IPv6Addr *Addr, IPv6Addr *MCastAddr);

extern void
NeighborSolicitSend(NeighborCacheEntry *NCE, IPv6Addr *Source);

extern void
IPv6FreePacket(PNDIS_PACKET Packet);

extern void
IPv6Send0(PNDIS_PACKET Packet, uint Offset, NeighborCacheEntry *NCE,
          IPv6Addr *DiscoveryAddress);

extern void
IPv6Send1(PNDIS_PACKET Packet, uint Offset, IPv6Header *IP,
          uint PayloadLength, RouteCacheEntry *RCE);

extern void
ICMPv6EchoTimeout(void);

extern void
IPULUnloadNotify(void);

extern Interface *
FindInterfaceFromIndex(uint Index);

extern void
IPv6Timeout(PKDPC MyDpcObject, void *Context, void *Unused1, void *Unused2);

extern IPv6Header UNALIGNED *
GetIPv6Header(PNDIS_PACKET Packet, uint Offset);

extern uint
PacketPullup(IPv6Packet *Packet, uint Needed);

extern int
InitNTE(NetTableEntry *NTE);

extern void
AddInterface(Interface *IF);

extern NetTableEntry *
CreateNTE(Interface *IF, IPv6Addr *Address, uint PrefixLength,
          uint ValidLifetime, uint PreferredLifetime);

extern AddressEntry *
CreateADE(Interface *IF, IPv6Addr *Addr, NetTableEntry *NTE);

extern void
DestroyADE(Interface *IF, AddressEntry *ADE);

extern void
DestroyADEs(Interface *IF, NetTableEntry *NTE);

extern NetTableEntry *
FindAddressOnInterface(Interface *IF, IPv6Addr *Addr, uint *AddrType,
                       uint *AddrScope);

extern int
FindInterfaceAddress(Interface *IF, IPv6Addr *Addr);

extern void
DeferSynchronizeMulticastAddresses(Interface *IF);

extern int
IPInit(WCHAR *RegKeyNameParam);

extern int
IPConfigure(WCHAR *RegKeyNameParam);

extern void
AddrConfUpdate(Interface *IF, IPv6Addr *Prefix, uint PrefixLength,
               uint ValidLifetime, uint PreferredLifetime);

extern void
AddrConfTimeout(NetTableEntry *NTE);

extern void
NetTableTimeout(void);

extern int
TunnelInit(WCHAR *RegKeyNameParam);

extern NeighborCacheEntry *
TunnelGetPseudoNCE(void);

extern void
MLDTimeout(void);

extern int
MLDSetMCastAddr(IPv6Addr *Addr, int *InterfaceNo, ULONG IOCTL_CODE);

extern int
IsIFMulticastCapable(Interface *IF);

#endif // IPv6DEF_INCLUDED
