// -*- mode: C++; tab-width: 4; indent-tabs-mode: nil -*- (for GNU Emacs)
//
// Copyright (c) 1985-1998 Microsoft Corporation
//
// This file is part of the Microsoft Research IPv6 Network Protocol Stack.
// You should have received a copy of the Microsoft End-User License Agreement
// for this software along with this release; see the file "license.txt".
// If not, please see http://www.research.microsoft.com/msripv6/license.htm,
// or write to Microsoft Research, One Microsoft Way, Redmond, WA 98052-6399.
//
// Abstract:
//
// Multicast Listener Discovery for Internet Protocol Version 6.
// See draft-ietf-ipngwg-mld-00.txt for details.
//


#include "oscfg.h"
#include "ndis.h"
#include "ip6imp.h"
#include "ip6opts.h"
#include "ip6def.h"
#include "mld.h"
#include "icmp.h"
#include "ntddip6.h"
#include "route.h"


AddressEntry *QueryList;
KSPIN_LOCK QueryListLock;


//* IsIFMulitcastCapable - Determine if an IF can accept MCast addresses
//
//  When a multicast address is added to an interface, the interface must be
//  updated so that packets addressed to the group will be received by the
//  interface.  This is done using two routines specified in the IF record.
//  The addresses of both of these routines must be present in order for the 
//  update to complete successfully.
//
int
IsIFMulticastCapable(
    Interface *IF)  // The interface in question.
{
    if (IF->ConvertMCastAddr == NULL || IF->SetMCastAddrList == NULL)
        return FALSE;
    else
        return TRUE;
}


//* FindAddr - find the interface and ADE entry for a given address.
//
//  Scan all interfaces looking for an address.  If not found, set
//  IF and ADE to NULL.
//
//  The caller must lock the IF before calling this function.
//
void
FindAddr(
    IPv6Addr *Addr,      // The address to find.
    Interface **IF,      // I/F the address belongs to, or NULL if not found.
    AddressEntry **ADE)  // ADE the address belongs to, or NULL if not found.
{
    Interface *TmpIF;
    AddressEntry *TmpADE;
    KIRQL OldIrql;
    char Found;
    
    *IF = NULL;
    *ADE = NULL;
    Found = 0;

    // REVIEW: the IFList should be locked?  Currently assumes caller has.

    for (TmpIF = IFList; TmpIF != NULL && !Found; TmpIF = TmpIF->Next) {
        //
        // For each interface, lock the interface, then scan all ADE 
        // entries looking for the desired address.
        // 
        KeAcquireSpinLock(&TmpIF->Lock, &OldIrql);
        for (TmpADE = TmpIF->ADE; TmpADE != NULL && !Found;
             TmpADE = TmpADE->Next) {
            if (IP6_ADDR_EQUAL(Addr, &TmpADE->Address)) {
                *IF = TmpIF;
                *ADE = TmpADE;
                Found = 1;
            }
        }
        KeReleaseSpinLock(&TmpIF->Lock, OldIrql);
    }
}


//* FindADE - find an ADE entry for the given interface.
//
//  If the address is assigned to the interface,
//  returns the ADE.  Otherwise returns NULL.
//
//  The caller must lock the IF before calling this function.
//
AddressEntry *
FindADE(
    Interface *IF,
    IPv6Addr *Addr)
{
    AddressEntry *ADE;

    // Check if address is assigned to the interface using the
    // interface's ADE list.
    for (ADE = IF->ADE; ADE != NULL; ADE = ADE->Next) {
        if (IP6_ADDR_EQUAL(Addr, &ADE->Address))
            break;
    }

    // Done
    return ADE;
}


//* AddToQueryList - Add an ADE entry to the QueryList.
//
//  Add an entry to the front of the QueryList. 
//  The caller should already have the QueryList and the IF locked.
//
//  Note that this routine increments the NTE reference count.  It is
//  not decremented until the ADE is removed from the QueryList either
//  by the timer expiring or being manually removed.
//                   
void
AddToQueryList(
   AddressEntry *ADE)  // The ADE to add to the query list.
{
    ADE->NextQL = QueryList;
    QueryList = ADE;
    AddRefNTE(ADE->NTE);
}


//* RemoveFromQueryList - Remove an  entry from the QueryList.
//
//  Remove an entry for the linked list of MLD query requests.
//  The caller should already have the QueryList and the IF locked.
//  
//  This routine also decrements the NTE reference count which was incremented
//  when the ADE was added to the QueryList.
//
void
RemoveFromQueryList(
    AddressEntry *ADE)  // The ADE entry to remove from the query list.
{
    AddressEntry  *PrevADE, *ThisADE;

    // Find the entry in the linked list, and adjust the pointers to 
    // remove it from the list.
    for (PrevADE = NULL, ThisADE = QueryList; ThisADE != NULL;
         ThisADE = ThisADE->NextQL ) {
        if (ThisADE == ADE) {
            // Found the entry to delete.
            if (PrevADE == NULL)
                QueryList = ThisADE->NextQL;
            else
                PrevADE->NextQL = ThisADE->NextQL;
            ThisADE->NextQL = NULL;
            break;
        }
    
        // Keep searching...
        PrevADE = ThisADE;
    }

    // Decrement the NTE reference count.
    ReleaseNTE(ADE->NTE);
}


//* MLDQuery - Process the receipt of a Group Query MLD message. 
//
//  Queries for a specific group should be sent to the group address
//  in question.  Genral queries are sent to the all nodes address, and 
//  have the group address set to zero.
//  Here we need to add the group to the list of groups waiting to send 
//  membership reports.  Then set the timer value in the ADE entry to a
//  random value determines by the incoming query.
//
int
MLDQuery(
    NetTableEntry *RecvNTE,  // NTE packet was received on.
    IPv6Packet *Packet,      // Packet handed to us by ICMPReceive.
    uchar Type,              // Type of ICMP message.
    uchar Code)              // ICMP Code field (type specific).
{
    AddressEntry *ADE;
    ushort MaxResponseDelay;
    ushort Timer;
    KIRQL OldIrql1, OldIrql2;
    MLDMessage *Message;

    //
    // Verify that we have enough contiguous data to overlay a MLDMessage
    // structure on the incoming packet.  Then do so.
    //
    if (Packet->ContigSize < sizeof(MLDMessage)) {
        if (PacketPullup(Packet, sizeof(MLDMessage)) == 0) {
            // Pullup failed.
            DbgPrint("MLDQuery: Packet too small to contain MLD message\n");
            return 0;  // Drop packet.
        }
    }
    Message = (MLDMessage *)Packet->Data;

    // Get the maximum response value from the received MLD message.
    MaxResponseDelay = net_short(Message->MaxResponseDelay);  // Milliseconds.

    // Lock the interface.
    KeAcquireSpinLock(&RecvNTE->IF->Lock, &OldIrql1);

    //
    // Lock the QueryList so that the timer routine doesn't alter
    // the QueryList while we're here.
    //
    KeAcquireSpinLock(&QueryListLock, &OldIrql2);

    // Loop through the ADE list and update the timer for the desired
    // groups.  Note that a general query specifies group address 0, and
    // sets the timer for all groups.
    for (ADE = RecvNTE->IF->ADE; ADE != NULL; ADE = ADE->Next) {
        // Only process ADE entries for this NTE.
        if (ADE->NTE != RecvNTE)
            continue;

        // Skip this ADE if the corresponding address is not multicast 
        // reportable.
        if (!(ADE->MCastFlag & ADE_REPORTABLE_ADDRESS))
            continue;
        
        // Skip this group address if the scope is reserved (0) or
        // node local (1).
        if (ADE->Scope == ADE_RESERVED || ADE->Scope == ADE_NODE_LOCAL)
            continue;

        // If group specific query, see if this is the correct group.
        if (!IP6_ADDR_EQUAL(&Message->GroupAddr, &UnspecifiedAddr) && 
            !IP6_ADDR_EQUAL(&Message->GroupAddr, &ADE->Address))
            continue;
        
        //
        // If the timer is currently off or if the maximum requested response
        // delay is less than the current timer value, draw a random value on
        // the interval (0, MaxResponseDelay) and update the timer to reflect
        // this value.
        //
        if (ADE->MCastTimer == 0 || MaxResponseDelay < ADE->MCastTimer) {
            // Add this ADE to the QueryList.
            if (ADE->MCastTimer == 0)
                AddToQueryList(ADE);
            
            // Update the timer value.
            ADE->MCastTimer = (ushort)
                ConvertMillisToTicks((Random() % MaxResponseDelay) + 1);

#if DBG
            // BUGBUG: this is for debugging purposes only.
            KdPrint(("MLDQuery: ADE->MCastTimer = %d\n", ADE->MCastTimer));
            if (ADE->MCastTimer > 10)
                ADE->MCastTimer = 10;
#endif
        }
    }

    KeReleaseSpinLock(&QueryListLock, OldIrql2);
    KeReleaseSpinLock(&RecvNTE->IF->Lock, OldIrql1);

    return IP_SUCCESS;
}


//* MLDReport - Process the receipt of a Group Report MLD message. 
//
//  When another host on the local link sends a group report, we receive
//  a copy if we also belong to the group.  If we have a timer running for
//  this group, we can turn it off now.
int
MLDReport(
    NetTableEntry *RecvNTE,  // NTE packet was received on.
    IPv6Packet *Packet,      // Packet handed to us by ICMPReceive.
    uchar Type,              // Type of ICMP message.
    uchar Code)              // ICMP Code field (type specific).
{
    AddressEntry *ADE;
    KIRQL OldIrql1, OldIrql2;
    MLDMessage *Message;
    NetTableEntry *TempNTE;
    uint TempType;
    uint TempScope;

    //
    // Verify that we have enough contiguous data to overlay a MLDMessage
    // structure on the incoming packet.  Then do so.
    //
    if (Packet->ContigSize < sizeof(MLDMessage)) {
        if (PacketPullup(Packet, sizeof(MLDMessage)) == 0) {
            // Pullup failed.
            DbgPrint("MLDReport: Packet too small to contain MLD message\n");
            return 0;  // Drop packet.
        }
    }
    Message = (MLDMessage *)Packet->Data;

    // Lock the ADE structure.
    KeAcquireSpinLock(&RecvNTE->IF->Lock, &OldIrql1);

    //
    // First things first.  NDIS versions < 5 loopback multicast packets 
    // automatically, so we want to simply drop any reports that we ourself
    // sent.  We'll look to see if the source address of the report is 
    // one of the unicast addresses on the receiving IF.  If so, drop it.
    //
    TempNTE = FindAddressOnInterface(RecvNTE->IF, &Packet->IP->Source, 
                                     &TempType, &TempScope);
    if (TempNTE != NULL) {
        //
        // OK, the source address is this host.  Decrement the NTE ref count
        // and bail out.
        //
        ReleaseNTE(TempNTE);
        KeReleaseSpinLock(&RecvNTE->IF->Lock, OldIrql1);
        KdPrint(("TCPIP6: MLDReport: dropped loopback report\n"));
        return IP_SUCCESS;
    }
    
    //
    // Lock the QueryList so that the timer routine doesn't remove
    // this ADE from the QueryList while we're here.
    //
    KeAcquireSpinLock(&QueryListLock, &OldIrql2);

    // Search for the ADE entry.  If found turn off the timer and 
    // the last reporter flag.
    ADE = FindADE(RecvNTE->IF, &Message->GroupAddr);

    if (ADE != NULL && ADE->MCastTimer != 0) {
        ADE->MCastTimer = 0;
        ADE->MCastFlag &= !ADE_LAST_MULTICAST_REPORTER;
        RemoveFromQueryList(ADE);
    }

    KeReleaseSpinLock(&QueryListLock, OldIrql2);
    KeReleaseSpinLock(&RecvNTE->IF->Lock, OldIrql1);

    return IP_SUCCESS;
}


//* MLDSendReport Send an MLD membership report.
//
//  This function is called either when a host first joins a multicast group or
//  at some point after a membership query message was received, and the timer
//  for this host has expired.
//
void
MLDSendReport(
    NetTableEntry *NTE,   // The NTE to send on.
    IPv6Addr *GroupAddr)  // The group address to report.
{
    Interface      *IF = NTE->IF;
    NDIS_STATUS    Status;
    PNDIS_PACKET   Packet;
    PNDIS_BUFFER   Buffer;
    uint           Offset;
    uchar          *Mem;
    uint           MemLen;
    uint           PayloadLength;
    IPv6Header     *IP;
    ICMPv6Header   *ICMP;
    MLDMessage  *MLD;
    MLDRouterAlertOption *RA;


    // Allocate a header for the packet.
    NdisAllocatePacket(&Status, &Packet, IPv6PacketPool);
    if (Status != NDIS_STATUS_SUCCESS) {
        KdPrint(("MLDSendReport: Couldn't allocate packet header!?!\n"));
        return;
    }
    PC(Packet)->CompletionHandler = IPv6FreePacket;

    // Allocate memory for buffer.
    Offset = MaxLinkLevelHeader;

    // REVIEW: what about router alert option?
    PayloadLength = sizeof(MLDRouterAlertOption) + sizeof(ICMPv6Header)
        + sizeof(MLDMessage);
    MemLen = Offset + sizeof(IPv6Header) + PayloadLength;
    Mem = ExAllocatePool(NonPagedPool, MemLen);
    if (Mem == NULL) {
        KdPrint(("MLDSendReport: Couldn't allocate memory!?!\n"));
        return;
    }

    // Allocate a buffer header and put it on the packet chain.
    NdisAllocateBuffer(&Status, &Buffer, IPv6BufferPool, Mem, MemLen);
    if (Status != NDIS_STATUS_SUCCESS) {
        KdPrint(("MLDSendReport: Couldn't allocate buffer header!?!\n"));
        return;
    }
    NdisChainBufferAtFront(Packet, Buffer);


    // Prepare IP header of reply packet.
    IP = (IPv6Header *)(Mem + Offset);
    IP->VersClassFlow = IP_VERSION;
    IP->NextHeader = IP_PROTOCOL_HOP_BY_HOP;
    IP->HopLimit = MLD_HOP_LIMIT; 
    IP->Dest = *GroupAddr;    // The group address.
    IP->Source = NTE->Address;

    // Prepare the router alert option
    RA = (MLDRouterAlertOption *)(IP + 1);
    RA->Header.NextHeader = IP_PROTOCOL_ICMPv6;
    RA->Header.HeaderExtLength = 0;
    RA->Option.Type = OPT6_ROUTER_ALERT;
    RA->Option.Length = 2;
    RA->Option.Value = MLD_ROUTER_ALERT_OPTION_TYPE;
    RA->Pad.Type = 1;
    RA->Pad.DataLength = 0;

    // Prepare ICMP header.
    ICMP = (ICMPv6Header *)(RA + 1);
    ICMP->Type = ICMPv6_MULTICAST_LISTENER_REPORT;
    ICMP->Code = 0;

    // Prepare the MLD message.
    MLD = (MLDMessage *)(ICMP + 1);
    MLD->MaxResponseDelay = 0;
    MLD->Unused = 0;
    MLD->GroupAddr = *GroupAddr;

    ICMPv6Send(NTE, Packet, Offset,
               Offset + sizeof(IPv6Header) + sizeof(MLDRouterAlertOption),
               IP, PayloadLength, ICMP);

    return;
}


//* MLDSendDone Send an MLD done message.
//
//  This function is called when a host quits a multicast group AND this was
//  the last host on the local link to report interest in the group.  A host 
//  quits when either the upper layer explicitly quits or when the interface
//  is deleted.
//
void
MLDSendDone(
    NetTableEntry *NTE,   // The NTE to send on.
    IPv6Addr *GroupAddr)  // The group address to leave.
{
    Interface *IF = NTE->IF;
    NDIS_STATUS Status;
    PNDIS_PACKET Packet;
    PNDIS_BUFFER Buffer;
    uint Offset;
    uchar *Mem;
    uint MemLen;
    uint PayloadLength;
    IPv6Header *IP;
    ICMPv6Header *ICMP;
    MLDMessage *MLD;
    MLDRouterAlertOption *RA;


    // Allocate a header for the packet.
    NdisAllocatePacket(&Status, &Packet, IPv6PacketPool);
    if (Status != NDIS_STATUS_SUCCESS) {
        KdPrint(("MLDSendDone: Couldn't allocate packet header!?!\n"));
        return;
    }
    PC(Packet)->CompletionHandler = IPv6FreePacket;

    // Allocate memory for buffer.
    Offset = MaxLinkLevelHeader;

    // REVIEW: what about router alert option?
    PayloadLength = sizeof(MLDRouterAlertOption) + sizeof(ICMPv6Header)
        + sizeof(MLDMessage);
    MemLen = Offset + sizeof(IPv6Header) + PayloadLength;
    Mem = ExAllocatePool(NonPagedPool, MemLen);
    if (Mem == NULL) {
        KdPrint(("MLDSendDone: Couldn't allocate memory!?!\n"));
        return;
    }

    // Allocate a buffer header and put it on the packet chain.
    NdisAllocateBuffer(&Status, &Buffer, IPv6BufferPool, Mem, MemLen);
    if (Status != NDIS_STATUS_SUCCESS) {
        KdPrint(("MLDSendDone: Couldn't allocate buffer header!?!\n"));
        return;
    }
    NdisChainBufferAtFront(Packet, Buffer);


    // Prepare IP header of reply packet.
    IP = (IPv6Header *)(Mem + Offset);
    IP->VersClassFlow = IP_VERSION;
    IP->NextHeader = IP_PROTOCOL_HOP_BY_HOP;
    IP->HopLimit = MLD_HOP_LIMIT; 
    IP->Dest = AllRoutersOnLinkAddr;
    IP->Source = NTE->Address;

    // Prepare the router alert option.
    RA = (MLDRouterAlertOption *)(IP + 1);
    RA->Header.NextHeader = IP_PROTOCOL_ICMPv6;
    RA->Header.HeaderExtLength = 0;
    RA->Option.Type = OPT6_ROUTER_ALERT;
    RA->Option.Length = 2;
    RA->Option.Value = MLD_ROUTER_ALERT_OPTION_TYPE;
    RA->Pad.Type = 1;
    RA->Pad.DataLength = 0;

    // Prepare ICMP header.
    ICMP = (ICMPv6Header *)(RA + 1);
    ICMP->Type = ICMPv6_MULTICAST_LISTENER_DONE;
    ICMP->Code = 0;

    // Prepare the MLD message.
    MLD = (MLDMessage *)(ICMP + 1);
    MLD->MaxResponseDelay = 0;
    MLD->Unused = 0;
    MLD->GroupAddr = *GroupAddr;
    
    ICMPv6Send(NTE, Packet, Offset,
               Offset + sizeof(IPv6Header) + sizeof(MLDRouterAlertOption),
               IP, PayloadLength, ICMP);

    return;
}


//* MLDAddMCastAddr - Add a multicast group to the specified interface.
//
//  This function is called when a user level program has asked to join a 
//  multicast group.  First we check to see if the group address is already
//  on the interface.  If not, then we create a new ADE entry and call the
//  lower layer routines to update the interface.
int
MLDAddMCastAddr(
  Interface *IF,       // The interface which should receive group traffic.
  IPv6Addr *Addr,      // The multicast address.        
  NetTableEntry *NTE)  // The NTE for this multicast address.
{
    AddressEntry *ADE;
    int SendReport = FALSE;
    KIRQL OldIrql1;
    KIRQL OldIrql2;

    if (!IsMulticast(Addr)) {
        KdPrint(("MLDAddMCastAddr: Not mcast addr\n"));
        return IP_BAD_REQ;
    }

    if (!IsIFMulticastCapable(IF)) {
        KdPrint(("MLDAddMCastAddr: IF cannot add a mcast addr\n"));
        return IP_BAD_REQ;
    }


    // Lock the interface.
    KeAcquireSpinLock(&IF->Lock, &OldIrql1);

    //
    // Find the ADE associated with this address.  
    // If the ADE is found, then make sure the address is not one of the 
    // special multicast addresses automatically added when the interface 
    // is created.  If the ADE is not found create a new one.
    //
    ADE = FindADE(IF, Addr);
    if (ADE != NULL) {
        if (!(ADE->MCastFlag & ADE_USER_MULTICAST_GROUP)) {
            KdPrint(("MLDAddMCastAddr: not a user mcast addr\n"));
            KeReleaseSpinLock(&IF->Lock, OldIrql1);
            return IP_SUCCESS;
        }
    } else {
        ADE = CreateADE(IF, Addr, NTE);
        if (ADE == NULL) {
            // can't allocate a new ADE entry.
            KdPrint(("MLDAddMCastAddr: can't alloc ADE memory\n"));
            KeReleaseSpinLock(&IF->Lock, OldIrql1);
            return IP_GENERAL_FAILURE;
        }
        ADE->MCastFlag = ADE_USER_REPORTABLE_ADDRESS;
    }

    //
    // Increment the reference count so we know how many user sockets are
    // interested in this group address.
    //
    ADE->MCastRefCount++;
    
    //
    // If this is a new multicast address then we need to update the
    // interface to receive this multicast address, send a 
    // MLD ADD_MEMBER message and set the timer so that a 
    // second ADD_MEMBER message is sent soon.
    // Also, we need to lock the QueryList so that the timer routine
    // doesn't change things while we're here.
    //
    if (ADE->MCastRefCount == 1) {
        KeAcquireSpinLock(&QueryListLock, &OldIrql2);
        AddToQueryList(ADE);
        if (ADE->Scope != ADE_RESERVED && ADE->Scope != ADE_NODE_LOCAL) {
            ADE->MCastTimer = MLD_UNSOLICITED_REPORT_INTERVAL;  // REVIEW: should be random value?
            ADE->MCastFlag |= ADE_LAST_MULTICAST_REPORTER;
            SendReport = TRUE;

            // Increment the NTE ref count so it doesn't disappear when we
            // release the IF lock.
            AddRefNTE(NTE);
        }
        KeReleaseSpinLock(&QueryListLock, OldIrql2);
    }

    // Done with the ADE data structure.
    KeReleaseSpinLock(&IF->Lock, OldIrql1);

    if (IF->MCastSynchNeeded)
        DeferSynchronizeMulticastAddresses(IF);

    if (SendReport) {
        MLDSendReport(NTE, Addr);    // send first of two reports

        // Now that we're done with the NTE, decrement the ref count.
        ReleaseNTE(NTE);
    }

    return IP_SUCCESS;
}


//* MLDDropMCastAddr - remove a musticast address from an interface.
//
//  This function is called when a user has indicated that they are no
//  longer interested in a multicast group.  So, decrement the reference 
//  count, and if it reaches zero, remove the address from the ADE list, 
//  and tell the interface to stop receiving the corresponding link 
//  address.
//
int
MLDDropMCastAddr(
    Interface *IF,   // The interface to remove this group from.
    IPv6Addr *Addr)  // The multicast address to drop.
{
    AddressEntry    *ADE;
    int             SendDrop = FALSE;
    uint            RefCount;
    NetTableEntry   *NTE;
    KIRQL           OldIrql1;
    KIRQL           OldIrql2;

    if (!IsMulticast(Addr)) {
        KdPrint(("MLDDropMCastAddr: Not mcast addr\n"));
        return IP_BAD_REQ;
    }

    // Lock the IF data structure.
    KeAcquireSpinLock(&IF->Lock, &OldIrql1);

    // Find the IF and ADE for this multicast address.
    ADE = FindADE(IF, Addr);
    
    // Find the ADE entry for this interface.  If it doesn't exist assume
    // its already deleted.
    if (ADE == NULL) {
        KeReleaseSpinLock(&IF->Lock, OldIrql1);
        return IP_SUCCESS;
    }

    // Make sure it's OK to drop this multicast address.
    if (!(ADE->MCastFlag & ADE_USER_MULTICAST_GROUP)) {
        KdPrint(("MLDDropMCastAddr: Not user mcast addr\n"));
        KeReleaseSpinLock(&IF->Lock, OldIrql1);
        return IP_SUCCESS;
    }

    // Lock the QueryList so the timer routine doesn't change things while
    // we're here.
    KeAcquireSpinLock(&QueryListLock, &OldIrql2);

    // Decrement the reference count.  If it goes to zero we need to remove 
    // the ADE entry.
    RefCount = --ADE->MCastRefCount;
    
    if (RefCount == 0) {
        // Record whether we need to send a leave message for this group.
        SendDrop = ADE->MCastFlag & ADE_LAST_MULTICAST_REPORTER;
        NTE = ADE->NTE;

        // Increment the NTE ref count so it doesn't disappear when we
        // release the IF lock.
        AddRefNTE(NTE);

        DestroyADE(IF, ADE);
    }

    // We're done with the IF data structure.
    KeReleaseSpinLock(&QueryListLock, OldIrql2);
    KeReleaseSpinLock(&IF->Lock, OldIrql1);

    // If the reference count is down to zero, tell the interface not to 
    // receive these packets.            
    if (RefCount == 0)
        DeferSynchronizeMulticastAddresses(IF);
    
    // If we are dropping this group and we were the last host on the local 
    // link to report interest in the group, then send a quit message.
    if (SendDrop) {
        MLDSendDone(NTE, Addr);

        // Now that we're done with the NTE, decrement the ref count.
        ReleaseNTE(NTE);
    }

    return IP_SUCCESS;
}


//* MLDSetMCastAddr - common add/drop routine called from the transport layer.
//
//  This function is the common entry point from the transport layer when a
//  user has resquested to add or drop a multicast group via setsockopt().  
//
//  Additionally this function is called if an application makes a direct 
//  DeviceIOControl() call into the IP stack.  This feature may disappear over
//  time however since the correct way to invoke this functionality is vis
//  the sockets API.
int
MLDSetMCastAddr(
    IPv6Addr *Addr,    // The multicast address.
    int *InterfaceNo,  // The interface number, if 0 replaced with IF used.
    ULONG Code)        // The IOCTL value
{
    Interface *IF;
    IPV6_MREQ *Query;
    RouteCacheEntry *RCE;
    KIRQL OldIrql;
    NetTableEntry *NTE;

    switch (Code) {
    case IOCTL_IPV6_ADD_MEMBERSHIP:
        // Determine the interface to add this address on.
        if (*InterfaceNo == 0) {
            //
            // The kernel needs to determine the interface to use, so we'll ask
            // the routing subsystem.
            //
            if ((RouteToDestination(Addr, NULL, &RCE)) 
                != IP_SUCCESS) {
                return STATUS_INVALID_PARAMETER;
            }

            // Use the interface associated with the RCE found in the routing
            // table.
            IF = RCE->NTE->IF;
            *InterfaceNo = IF->Index;
            ReleaseRCE(RCE);
        } else {
            //
            // Use the interface requested by the application.
            //
            IF = FindInterfaceFromIndex(*InterfaceNo);
            if (IF == NULL) {
                return STATUS_INVALID_PARAMETER; 
            }
        }

        // 
        // Find a link-local address on this interface to use as the source
        // address.
        //
        KeAcquireSpinLock(&IF->Lock, &OldIrql);
        for (NTE = IF->NTE; NTE != NULL; NTE = NTE->NextOnIF) {
            if (IsLinkLocal(&NTE->Address))
                // Found a link local address.
                break;
        }
        KeReleaseSpinLock(&IF->Lock, OldIrql);

        if (NTE == NULL) {
            KdPrint(("MLDSetMCastAddr: could not find a lick-local "
                     "address\n"));
            return STATUS_INVALID_PARAMETER;
        }
           

        if (MLDAddMCastAddr(IF, Addr, NTE) == IP_SUCCESS)
            return STATUS_SUCCESS;
        else
            return STATUS_UNSUCCESSFUL;


    case IOCTL_IPV6_DROP_MEMBERSHIP:
        //
        // Use the interface requested by the application.
        //
        IF = FindInterfaceFromIndex(*InterfaceNo);
        if (IF == NULL) {
            return STATUS_INVALID_PARAMETER; 
        }
        // Remove the group address from the interface.
        if (MLDDropMCastAddr(IF, Addr) == IP_SUCCESS)
            return STATUS_SUCCESS;
        else
            return STATUS_UNSUCCESSFUL;
    }
}


//* MLDTimeout - Handle an MLD timer event.
//
//  This function is called periodically by IPv6Timeout.  Decrement each 
//  waiting query request.  If the timer reaches zero, then send a
//  multicast member report.
//
void
MLDTimeout(void)
{
    AddressEntry *ADE, *PrevADE, *NextADE;
    MLDReportList *ReportList, *Report, *NextReport;
    
    if (QueryList == NULL) {
        KdPrint(("MLDTimeout: nothing to do\n"));
        return;
    }

    // Lock the QueryList.
    KeAcquireSpinLockAtDpcLevel(&QueryListLock);

    ReportList = NULL;

    // Scan the Query List and save any expired entries and remove them from
    // the query list.
    for(PrevADE = NULL, ADE = QueryList; ADE != NULL; ADE = NextADE) {
        NextADE = ADE->NextQL;
        --ADE->MCastTimer;
        if (ADE->MCastTimer != 0) {
            // PrevADE is the previous NON-EXPIRED ADE on the QL.
            PrevADE = ADE;
            continue;
        }

        // This entry has expired, we need to send a report.

        // Unchain this entry from the query list...
        if (PrevADE == NULL)
            QueryList = NextADE;
        else
            PrevADE->NextQL = NextADE;
        ADE->NextQL = NULL;

        // and add it to the list of reports to send.
        Report = ExAllocatePool(NonPagedPool, sizeof *Report);
        if (Report != NULL) {
            Report->Next = ReportList;
            Report->GroupAddr = ADE->Address;
            Report->NTE = ADE->NTE;
            ReportList = Report;
        }
    }
    KeReleaseSpinLockFromDpcLevel(&QueryListLock);

    // Now send any required reports.
    for (Report = ReportList; Report != NULL; Report = NextReport) {
        NextReport = Report->Next;
        MLDSendReport(Report->NTE, &Report->GroupAddr);
        
        //
        // Now we need to decrement the NTE reference count, which was
        // incremented when the ADE was added to the QueryList.
        //
        ReleaseNTE(Report->NTE);
        KdPrint(("MLDTimeout: new NTE RefCnt: %d\n",Report->NTE->RefCnt));
        ExFreePool(Report);
    }
    
    // We're done.
}


#pragma BEGIN_INIT

//* MLDInit - Initialize MLD.
//
//  This bit of code initializes MLD generally.  There is also some amount
//  of work done on a per-NTE basis that we do when each one is initialized.
//
void
MLDInit(void)
{
    KeInitializeSpinLock(&QueryListLock);
    QueryList = NULL;
}

#pragma END_INIT
