// -*- mode: C++; tab-width: 4; indent-tabs-mode: nil -*- (for GNU Emacs)
//
// Copyright (c) 1998 Microsoft Corporation
//
// This file is part of the Microsoft Research IPv6 Network Protocol Stack.
// You should have received a copy of the Microsoft End-User License Agreement
// for this software along with this release; see the file "license.txt".
// If not, please see http://www.research.microsoft.com/msripv6/license.htm,
// or write to Microsoft Research, One Microsoft Way, Redmond, WA 98052-6399.
//
// Abstract:
//
// Neighbor Discovery (ND) Protocol for Internet Protocol Version 6.
// Logically a part of ICMPv6, but put in a seperate file for clarity.
// See RFC 1970 for details.
//
// NOTE: Updated for the post-1970 Neighbor Discovery drafts.
//


#include "oscfg.h"
#include "ndis.h"
#include "ip6imp.h"
#include "ip6def.h"
#include "icmp.h"
#include "neighbor.h"
#include "route.h"
#include "alloca.h"

//
//  REVIEW: What is a reasonable value for NeighborCacheLimit?
//  Should probably be sized relative to physical memory
//  and link characteristics.
//
//  The route cache holds RCEs, which in turn hold references
//  to NCEs. If the limit for RCE cache is large relative to the
//  limit for NCE caching, we will waste much time trying to find
//  unused NCEs. We may need a mechanism to flush unused RCEs
//  when we run short of NCEs.
//
//  Also note that we never reclaim NCEs (no matter how stale)
//  until we hit the limit. Not clear if this is good or bad.
//  And we never free NCEs at all.
//
//  We cache & reclaim NCEs on a per-interface basis.
//  Theoretically it would be better to use a global LRU list.
//  However this would introduce added overhead (making NCEs bigger)
//  and locking. At this point I believe that it isn't worth it.
//
uint NeighborCacheLimit = 4;

//* NeighborCacheInit
//
//  Initialize the neighbor cache for an interface.
//
void
NeighborCacheInit(Interface *IF)
{
    IF->FirstNCE = IF->LastNCE = SentinelNCE(IF);
    IF->NCECount = 0;
}

//* CreateOrReuseNeighbor
//
//  Creates a new NCE for an interface.
//  Attempts to reuse an existing NCE if the interface
//  already has too many NCEs.
//
//  Called with the interface lock held.
//
//  Returns NULL if a new NCE could not be created.
//  Returns the NCE with RefCnt, IF, LinkAddress, WaitQueue, Validate
//  fields initialized. (Actually Validate doesn't need
//  to be initialized, an arbitrary starting value is fine.)
//
NeighborCacheEntry *
CreateOrReuseNeighbor(Interface *IF)
{
    NeighborCacheEntry *NCE;

    if (IF->NCECount >= NeighborCacheLimit) {
        //
        // Try to reclaim an unused NCE.
        // Search backwards (starting with the LRU NCEs)
        // for an NCE with zero references.
        //
        for (NCE = IF->LastNCE; NCE != SentinelNCE(IF); NCE = NCE->Prev) {

            if ((NCE->RefCnt == 0) &&
                (NCE->WaitQueue == NULL)) {
                //
                // Reuse this NCE.
                // Just need to unlink it.
                //
                NCE->Next->Prev = NCE->Prev;
                NCE->Prev->Next = NCE->Next;

                return NCE;
            }
        }
    }

    NCE = (NeighborCacheEntry *) ExAllocatePool(NonPagedPool, sizeof *NCE +
                                                IF->LinkAddressLength);
    if (NCE == NULL)
        return NULL;

    IF->NCECount++;

    NCE->RefCnt = 0;
    NCE->LinkAddress = (void *)(NCE + 1);
    NCE->IF = IF;
    NCE->WaitQueue = NULL;

    return NCE;
}

//* CreatePermanentNeighbor
//
//  Creates a permanent neighbor - one that does not participate
//  in neighbor discovery or neighbor unreachability detection.
//
//  Callable from thread or DPC context.
//
//  Returns NULL if the NCE could not be created.
//
NeighborCacheEntry *
CreatePermanentNeighbor(Interface *IF, IPv6Addr *Addr, void *LinkAddress)
{
    KIRQL OldIrql;
    NeighborCacheEntry *NCE;

    KeAcquireSpinLock(&IF->Lock, &OldIrql);

    //
    // Get a new entry for this neighbor.
    //
    NCE = CreateOrReuseNeighbor(IF);
    if (NCE == NULL) {
        KeReleaseSpinLock(&IF->Lock, OldIrql);
        return NULL;
    }

    //
    // Initialize the neighbor cache entry.
    //
    NCE->RefCnt = 1; // To make it permanent.
    NCE->NeighborAddress = *Addr;
    RtlCopyMemory(NCE->LinkAddress, LinkAddress, IF->LinkAddressLength);
    NCE->IsRouter = FALSE;
    NCE->NDState = ND_STATE_PERMANENT;
    NCE->EventTimer = 0;

    //
    // Link new entry into chain of NCEs on this interface.
    //
    NCE->Next = IF->FirstNCE;
    NCE->Next->Prev = NCE;
    NCE->Prev = SentinelNCE(IF);
    NCE->Prev->Next = NCE;

    //
    // Do not include permanent neighbors in NCECount.
    //
    IF->NCECount--;

    KeReleaseSpinLock(&IF->Lock, OldIrql);

    return NCE;
}

//* FindOrCreateNeighbor
//
//  Searches an interface's neighbor cache for an entry.
//  Creates the entry (but does NOT send initial solicit)
//  if an existing entry is not found.
//
//  Takes interface locks; may be called with route cache lock held.
//  Callable from thread or DPC context.
//
//  We avoid sending a solicit here, because this function
//  is called while holding locks that make that a bad idea.
//
//  Returns NULL only if an NCE could not be created.
//
NeighborCacheEntry *
FindOrCreateNeighbor(Interface *IF, IPv6Addr *Addr)
{
    KIRQL OldIrql;
    NeighborCacheEntry *NCE;

    KeAcquireSpinLock(&IF->Lock, &OldIrql);

    for (NCE = IF->FirstNCE; NCE != SentinelNCE(IF); NCE = NCE->Next) {

        if (IP6_ADDR_EQUAL(Addr, &NCE->NeighborAddress)) {
            //
            // Found matching entry.
            //
            goto ReturnNCE;
        }
    }

    //
    // Get a new entry for this neighbor.
    //
    NCE = CreateOrReuseNeighbor(IF);
    if (NCE == NULL) {
        KeReleaseSpinLock(&IF->Lock, OldIrql);
        return NULL;
    }

    //
    // The RefCnt, IF, LinkAddress, WaitQueue fields are already initialized.
    //
    NCE->NeighborAddress = *Addr;
#if DBG
    RtlZeroMemory(NCE->LinkAddress, IF->LinkAddressLength);
#endif
    NCE->IsRouter = FALSE;
    NCE->NDState = ND_STATE_INCOMPLETE;
    NCE->EventTimer = NCE->EventCount = 0;

    // 
    // We don't want to perform neighbor discovery on multicast addresses,
    // so we set the ND state accordingly.  Also, we create the link-layer
    // multicast address for IP multicast addresses.
    //
    if (IsMulticast(Addr)) {
        (IF->ConvertMCastAddr)(IF->LinkContext, Addr, NCE->LinkAddress);
        NCE->NDState = ND_STATE_PERMANENT;  // This means no ND.
    }

    //
    // Link new entry into chain of NCEs on this interface.
    // Put the new entry at the end, because until it is
    // used to send a packet it has not proven itself valuable.
    //
    NCE->Prev = IF->LastNCE;
    NCE->Prev->Next = NCE;
    NCE->Next = SentinelNCE(IF);
    NCE->Next->Prev = NCE;


  ReturnNCE:
    AddRefNCE(NCE);
    KeReleaseSpinLock(&IF->Lock, OldIrql);
    return NCE;
}

//* NeighborCacheUpdate - Update link address information about a neighbor.
//
//  Called when we've received possibly new information about one of our
//  neighbors, the source of a Neighbor Solicitation or Router Advertisement.
//  Note that our receipt of this packet DOES NOT imply forward reachability
//  to this neighbor, so we do not update our LastReachable timer.
//
//  Callable from DPC context, not from thread context.
//
//  The LinkAddress might be NULL, which means we can only process IsRouter.
//
//  If IsRouter is FALSE, then we don't know whether the neighbor
//  is a router or not. If it's TRUE, then we know it is a router.
//
void
NeighborCacheUpdate(NeighborCacheEntry *NCE, // The neighbor.
                    void *LinkAddress,       // Corresponding media address.
                    int IsRouter)            // Do we know it's a router.
{
    Interface *IF = NCE->IF;
    PNDIS_PACKET Packet = NULL;

    KeAcquireSpinLockAtDpcLevel(&IF->Lock);

    //
    // Check to see if the link address changed.
    //
    if ((LinkAddress != NULL) &&
        ((NCE->NDState == ND_STATE_INCOMPLETE) ||
         RtlCompareMemory(LinkAddress, NCE->LinkAddress,
                          IF->LinkAddressLength) != IF->LinkAddressLength)) {
        //
        // Link-level address changed.  Update cache entry with the
        // new one and change state to STALE as we haven't verified
        // forward reachability with the new address yet.
        //
        RtlCopyMemory(NCE->LinkAddress, LinkAddress, IF->LinkAddressLength);
        NCE->EventTimer = 0; // Cancel any outstanding timeout.
        NCE->NDState = ND_STATE_STALE;

        //
        // Flush the queue of waiting packets.
        // (Only relevant if we were in the INCOMPLETE state.)
        //
        if (NCE->WaitQueue != NULL) {

            Packet = NCE->WaitQueue;
            NCE->WaitQueue = NULL;
        }
    }

    //
    // If we know that the neighbor is a router,
    // remember that fact.
    //
    if (IsRouter)
        NCE->IsRouter = TRUE;

    KeReleaseSpinLockFromDpcLevel(&IF->Lock);

    //
    // If we can now send a packet, do so.
    // (Without holding a lock.)
    //
    if (Packet != NULL) {
        uint Offset;

        Offset = PC(Packet)->pc_offset;
        IPv6Send0(Packet, Offset, NCE, &(PC(Packet)->DiscoveryAddress));
    }
}

//* NeighborCacheSearch
//
//  Searches the neighbor cache for an entry that matches
//  the neighbor IPv6 address. If found, returns the link-level address.
//  Returns FALSE to indicate failure.
//
//  Callable from DPC context, not from thread context.
//
int
NeighborCacheSearch(
    Interface *IF,
    IPv6Addr *Neighbor,
    void *LinkAddress)
{
    NeighborCacheEntry *NCE;

    KeAcquireSpinLockAtDpcLevel(&IF->Lock);

    for (NCE = IF->FirstNCE; NCE != SentinelNCE(IF); NCE = NCE->Next) {

        if (IP6_ADDR_EQUAL(Neighbor, &NCE->NeighborAddress)) {
            //
            // Entry found. Return it's cached link-address,
            // if it's valid.
            //
            if (NCE->NDState == ND_STATE_INCOMPLETE) {
                //
                // No valid link address.
                //
                break;
            }

            //
            // The entry has a link-level address.
            // Must copy it with the lock held.
            //
            RtlCopyMemory(LinkAddress, NCE->LinkAddress,
                          IF->LinkAddressLength);
            KeReleaseSpinLockFromDpcLevel(&IF->Lock);
            return TRUE;
        }
    }

    KeReleaseSpinLockFromDpcLevel(&IF->Lock);
    return FALSE;
}

//* NeighborCacheAdvert
//
//  Updates the neighbor cache in response to an advertisement.
//  If no matching entry is found, ignores the advertisement.
//  (See RFC 1970 section 7.2.5.)
//
//  Callable from DPC context, not from thread context.
//
void
NeighborCacheAdvert(
    Interface *IF,
    IPv6Addr *TargetAddress,
    void *LinkAddress,
    ulong Flags)
{
    NeighborCacheEntry *NCE;
    PNDIS_PACKET Packet = NULL;
    int PurgeRouting = FALSE;

    KeAcquireSpinLockAtDpcLevel(&IF->Lock);

    for (NCE = IF->FirstNCE; NCE != SentinelNCE(IF); NCE = NCE->Next) {
        if (IP6_ADDR_EQUAL(TargetAddress, &NCE->NeighborAddress)) {

            //
            // Pick up link-level address from the advertisement,
            // if we don't have one yet or if override is set.
            //
            if ((LinkAddress != NULL) &&
                ((NCE->NDState == ND_STATE_INCOMPLETE) ||
                 ((Flags & ND_NA_FLAG_OVERRIDE) &&
                  RtlCompareMemory(LinkAddress, NCE->LinkAddress,
                   IF->LinkAddressLength) != IF->LinkAddressLength))) {

                RtlCopyMemory(NCE->LinkAddress, LinkAddress,
                              IF->LinkAddressLength);

                NCE->EventTimer = 0; // Cancel any outstanding timeout.
                NCE->NDState = ND_STATE_STALE;

                //
                // Flush the queue of waiting packets.
                //
                if (NCE->WaitQueue != NULL) {

                    Packet = NCE->WaitQueue;
                    NCE->WaitQueue = NULL;

                    //
                    // Need to keep ref on NCE after we unlock.
                    //
                    AddRefNCE(NCE);
                }

                goto AdvertisementMatchesCachedAddress;
            }

            if ((NCE->NDState != ND_STATE_INCOMPLETE) &&
                ((LinkAddress == NULL) ||
                 RtlCompareMemory(LinkAddress, NCE->LinkAddress,
                  IF->LinkAddressLength) == IF->LinkAddressLength)) {
                ushort WasRouter;

            AdvertisementMatchesCachedAddress:

                //
                // If this is a solicited advertisement
                // for our cached link-layer address,
                // then we have confirmed reachability.
                //
                if (Flags & ND_NA_FLAG_SOLICITED) {
                    NCE->EventTimer = 0;  // Cancel any outstanding timeout.
                    KeQueryTickCount(&NCE->LastReachable);  // Timestamp it.
                    NCE->NDState = ND_STATE_REACHABLE;
                }

                //
                // If this is an advertisement
                // for our cached link-layer address,
                // then update IsRouter.
                //
                WasRouter = NCE->IsRouter;
                NCE->IsRouter = ((Flags & ND_NA_FLAG_ROUTER) != 0);
                if (WasRouter && !NCE->IsRouter) {
                    //
                    // This neighbor used to be a router, but is no longer.
                    //
                    PurgeRouting = TRUE;

                    //
                    // Need to keep ref on NCE after we unlock.
                    //
                    AddRefNCE(NCE);
                }
            }
            else {
                //
                // This is not an advertisement
                // for our cached link-layer address;
                // it is an unsolicited advertisement.
                // Give NUD a little push, if not already in progress.
                //

                if (Flags & ND_NA_FLAG_SOLICITED)
                    DbgPrint("NeighborCacheAdvert: solicited non-matching?\n");

                if (NCE->NDState == ND_STATE_REACHABLE)
                    NCE->NDState = ND_STATE_STALE;
            }

            //
            // Only one NCE should match.
            //
            break;
        }
    }

    KeReleaseSpinLockFromDpcLevel(&IF->Lock);

    //
    // If we can now send a packet, do so.
    // (Without holding a lock.)
    //
    if (Packet != NULL) {
        uint Offset;

        Offset = PC(Packet)->pc_offset;
        IPv6Send0(Packet, Offset, NCE, &(PC(Packet)->DiscoveryAddress));
        ReleaseNCE(NCE);
    }

    //
    // If need be, purge the routing data structures.
    //
    if (PurgeRouting) {
        InvalidateRouter(NCE);
        ReleaseNCE(NCE);
    }
}

//* NeighborCacheEntryTimeout - handle an event timeout on an NCE.
//
//  NeighborCacheTimeout calls this routine when
//  an NCE's EventTimer expires.
//
//  Called with the interface lock held.
//
void
NeighborCacheEntryTimeout(
    NeighborCacheEntry *NCE)
{
    //
    // Neighbor Discovery has timeouts for initial neighbor
    // solicitation retransmissions, the delay state, and probe
    // neighbor solicitation retransmissions.  All of these share
    // the same EventTimer, and are distinguished from each other
    // by the NDState.
    //
    switch (NCE->NDState) {
    case ND_STATE_INCOMPLETE:
        //
        // Retransmission timer expired.  Check to see
        // if we've sent more than the maximum.
        //
        if (NCE->EventCount++ > MAX_MULTICAST_SOLICIT) {
            //
            // Failed to initiate connectivity to neighbor.
            // Reset to dormant INCOMPLETE state.
            //
            NCE->EventCount = 0;
            if (NCE->WaitQueue != NULL) {
                IPv6SendComplete(NULL, NCE->WaitQueue,
                                 NDIS_STATUS_REQUEST_ABORTED);
                NCE->WaitQueue = NULL;
            }

            InvalidateRouteCacheForNCE(NCE);

        } else {
            PNDIS_PACKET Packet;
            IPv6Addr *SourceAddr;

            //
            // Retransmit initial solicit, taking source address
            // from the waiting packet.
            // BUGBUG - Should check that the source address
            // is assigned to the interface. (Might not be in weak model.)
            //

            NCE->EventTimer = NCE->IF->RetransTimer;
            Packet = NCE->WaitQueue;
            if (Packet != NULL) {
                IPv6Header UNALIGNED *IP;

                IP = GetIPv6Header(Packet, PC(Packet)->pc_offset);
                SourceAddr = &IP->Source;
            } else
                SourceAddr = NULL;
            NeighborSolicitSend(NCE, SourceAddr);
        }
        break;

    case ND_STATE_DELAY:
        //
        // Delay timer expired.  Enter PROBE state, send
        // first probe, and start retransmission timer.
        //
        NCE->NDState = ND_STATE_PROBE;
        NCE->EventCount = 0;
        NCE->EventTimer = NCE->IF->RetransTimer;

        // Drop into PROBE state to send first probe...

    case ND_STATE_PROBE:
        //
        // Retransmission timer expired.  Check to see
        // if we've sent more than the maximum.
        //
        if (NCE->EventCount++ > MAX_UNICAST_SOLICIT) {
            //
            // Failed to initiate connectivity to neighbor.
            // Reset to dormant INCOMPLETE state.
            //
            NCE->NDState = ND_STATE_INCOMPLETE;
            NCE->EventCount = 0;

            InvalidateRouteCacheForNCE(NCE);

        } else {
            //
            // Retransmit probe solicitation.
            // NB: We do not have a waiting packet from which
            // to take a source address.
            //
            NCE->EventTimer = NCE->IF->RetransTimer;
            NeighborSolicitSend(NCE, NULL);
        }
        break;

    default:
        //
        // Should never happen.  Other states don't utilize
        // the EventTimer.
        //
        DbgPrint("NeighborCacheEntryTimeout: ND Event Timer in bad state\n");
        break;
    }
}

//* NeighborCacheTimeout
//
//  Called periodically from IPv6Timeout to handle timeouts
//  in the interface's neighbor cache.
//
//  Callable from DPC context, not from thread context.
//
//  Note that NeighborCacheTimeout performs an unbounded
//  amount of work while holding the interface lock.
//  NeighborCacheEntryTimeout can send packets.
//  This is difficult to avoid, because if we unlock
//  then the list can be scrambled.
//
//  One possible strategy that would help, if this is a problem,
//  would be to have a second singly-linked list of NCEs
//  that require action.  With one traversal we reference NCEs
//  and create the action list.  Then we could traverse the action
//  list at our leisure, taking/dropping the interface lock.
//
//  On the other hand, this is all moot on a uniprocessor
//  because our locks are spinlocks and we are already at DPC level.
//  That is, on a uniprocessor KeAcquireSpinLockAtDpcLevel is a no-op.
//
void
NeighborCacheTimeout(Interface *IF)
{
    NeighborCacheEntry *NCE;

    KeAcquireSpinLockAtDpcLevel(&IF->Lock);
    for (NCE = IF->FirstNCE; NCE != SentinelNCE(IF); NCE = NCE->Next) {

        if (NCE->EventTimer != 0) {
            //
            // Timer is running.  Decrement and check for expiration.
            //
            if (--NCE->EventTimer == 0) {
                //
                // Timer went off.
                //
                NeighborCacheEntryTimeout(NCE);
            }
        }
    }
    KeReleaseSpinLockFromDpcLevel(&IF->Lock);
}


//* RouterSolicitReceive - Handle Router Solicitation messages.
//
//  BUGBUG: Add router support later.
//
int
RouterSolicitReceive(
    NetTableEntry *RecvNTE,     // NTE that received the packet.
    IPv6Packet *Packet,         // Packet handed to us by ICMPv6Receive.
    uchar Type,                 // Value of Type field in ICMPv6 header.
    uchar Code)                 // Value of Code field in ICMPv6 header.
{
    DbgPrint("IPv6 ND: Received a Router Solicitation\n");
    return 0;
}


//* RouterAdvertReceive - Handle Router Advertisement messages.
//
//  Validate message, update Default Router list, On-Link Prefix list,
//  perform address auto-configuration.
//  See sections 6.1.2, 6.3.4 of RFC 1970.
//
int
RouterAdvertReceive(
    NetTableEntry *RecvNTE,     // NTE that received the packet.
    IPv6Packet *Packet,         // Packet handed to us by ICMPv6Receive.
    uchar Type,                 // Value of Type field in ICMPv6 header.
    uchar Code)                 // Value of Code field in ICMPv6 header.
{
    Interface *IF = RecvNTE->IF;
    uint Flags, CurHopLimit, RouterLifetime;
    uint AdvReachableTime, RetransTimer;
    void *SourceLinkAddress;
    NeighborCacheEntry *NCE;


    //
    // Validate the advertisement.
    // By the time we get here, any IPv6 Authentication Header will have
    // already been checked, as will have the ICMPv6 checksum.  Still need
    // to check the length, source address, hop limit, and ICMP code.
    //
    if (Packet->IP->HopLimit != 255) {
        //
        // Packet was forwarded by a router, therefore it cannot be
        // from a legitimate neighbor.  Drop the packet.
        //
        DbgPrint("IPv6 ND: Received a routed router advertisement\n");
        return 0;
    }
    if (Code != 0) {
        //
        // Bogus/corrupted router advertisement.  Drop the packet.
        //
        DbgPrint("IPv6 ND: Received a corrupted router advertisement\n");
        return 0;
    }
    if (!IsLinkLocal(&Packet->IP->Source)) {
        //
        // Source address should always be link-local. Drop the packet.
        //
        DbgPrint("IPv6 ND: Non-link-local source in router advertisement\n");
        return 0;
    }

    //
    // Pull CurHopLimit, Flags, RouterLifetime,
    // AdvReachableTime, RetransTimer out of the packet.
    //
    if (Packet->ContigSize < 12)
        if (PacketPullup(Packet, 12) == 0) {
            //
            // Packet too short to contain minimal RA.
            //
            DbgPrint("IPv6 ND: Received a too short router advertisement\n");
            return 0;
        }
    CurHopLimit = *(((uchar *)Packet->Data)++);
    Flags = *(((uchar *)Packet->Data)++);
    RouterLifetime = net_short(*(((ushort *)Packet->Data)++));
    AdvReachableTime = net_long(*(((ulong *)Packet->Data)++));
    RetransTimer = net_long(*(((ulong *)Packet->Data)++));
    Packet->ContigSize -= 12;
    Packet->TotalSize -= 12;
    Packet->Position += 12;

    //
    // The code below assumes a contigous buffer for all the options
    // (the remainder of the packet).  If that isn't currently the
    // case, do a pullup for the whole thing.
    //
    // REVIEW: This is less than perfectly efficient, but we shouldn't
    // REVIEW: do it often, and the code below would need serious
    // REVIEW: reworking to allow for pullups only when needed.
    //
    if (Packet->ContigSize != Packet->TotalSize)
        if (PacketPullup(Packet, Packet->TotalSize) == 0) {
            // Can only fail if we run out of memory.
            return 0;
        }

    ASSERT(Packet->ContigSize == Packet->TotalSize);

    //
    // Look for a source link-layer address option.
    // Also sanity-check the options before doing anything permanent.
    //
    SourceLinkAddress = NULL;
    while (Packet->ContigSize) {
        uchar OptionLength;

        //
        // Validate the option length.
        //
        if ((Packet->ContigSize < 8) ||
            ((OptionLength = *((uchar *)Packet->Data + 1) << 3) == 0) ||
            (OptionLength > Packet->ContigSize)) {
            //
            // Invalid option length.  Drop the packet.
            //
            DbgPrint("IPv6 ND: Received RA option with with bogus length\n");
            return 0;
        }

        switch (*(uchar *)Packet->Data) {
        case ND_OPTION_SOURCE_LINK_LAYER_ADDRESS:
            //
            // Parse the link-layer address option.
            //
            SourceLinkAddress = (*IF->ReadLLOpt)(IF->LinkContext,
                                                 (uchar *)Packet->Data);
            if (SourceLinkAddress == NULL) {
                //
                // Invalid option format.  Drop the packet.
                //
                DbgPrint("IPv6 ND: Received RA with bogus ll option\n");
                return 0;
            }
            //
            // Note that if there are multiple options for some bogus reason,
            // we use the last one.  We sanity-check all the options.
            //
            break;

        case ND_OPTION_MTU:
            //
            // Sanity-check the option.
            //
            if (OptionLength != 8) {
                //
                // Invalid option format.  Drop the packet.
                //
                DbgPrint("IPv6 ND: Received RA with bogus mtu option\n");
                return 0;
            }
            break;

        case ND_OPTION_PREFIX_INFORMATION: {
            NDOptionPrefixInformation *option =
                (NDOptionPrefixInformation *)Packet->Data;

            //
            // Sanity-check the option.
            //
            if ((OptionLength != 32) ||
                (option->PrefixLength > 128)) {
                //
                // Invalid option format.  Drop the packet.
                //
                DbgPrint("IPv6 ND: Received RA with bogus prefix "
                         "information option\n");
                return 0;
            }
            break;
        }
        }

        //
        // Move forward to next option.
        //
        (uchar *)Packet->Data += OptionLength;
        Packet->ContigSize -= OptionLength;
    }

    //
    // Reset Data pointer & ContigSize.
    //
    (uchar *)Packet->Data -= Packet->TotalSize;
    Packet->ContigSize += Packet->TotalSize;

    //
    // Get the Neighbor Cache Entry for the source of this RA.
    //
    NCE = FindOrCreateNeighbor(IF, &Packet->IP->Source);
    if (NCE == NULL) {
        //
        // Couldn't find or create NCE.  Drop the packet.
        //
        return 0;
    }

    //
    // Update the Neighbor Cache Entry for the source of this RA.
    //
    NeighborCacheUpdate(NCE, SourceLinkAddress, TRUE);

    //
    // Update the Default Router List.
    //
    RouterLifetime = ConvertSecondsToTicks(RouterLifetime);
    if (RouterLifetime == ROUTER_LIFETIME_INFINITE)
        RouterLifetime--;
    RouterListUpdate(NCE, RouterLifetime);

    //
    // Update the hop limit for the interface.
    // NB: We rely on loads/stores of the CurHopLimit field being atomic.
    //
    if (CurHopLimit != 0) {
        IF->CurHopLimit = CurHopLimit;
    }

    KeAcquireSpinLockAtDpcLevel(&IF->Lock);

    //
    // Stop sending router solications for this interface.
    // Note that we should always send at least one router solicitation,
    // even if we receive an unsolicited router advertisement first.
    //
    if (RouterLifetime != 0) {
        if (IF->RSCount > 0)
            IF->RSTimer = 0;
    }

    //
    // Update the BaseReachableTime and ReachableTime.
    // NB: We use a lock for coordinated updates, but other code
    // reads the ReachableTime field without a lock.
    //
    if ((AdvReachableTime != 0) &&
        (AdvReachableTime != IF->BaseReachableTime)) {

        IF->BaseReachableTime = AdvReachableTime;
        IF->ReachableTime = CalcReachableTime(AdvReachableTime);
    }

    KeReleaseSpinLockFromDpcLevel(&IF->Lock);

    //
    // Update the RetransTimer field.
    // NB: We rely on loads/stores of this field being atomic.
    //
    if (RetransTimer != 0)
        IF->RetransTimer = ConvertMillisToTicks(RetransTimer);

    //
    // Process any LinkMTU, PrefixInformation options.
    //

    while (Packet->ContigSize) {
        uchar OptionLength;

        //
        // The option length was validated in the first pass
        // over the options, above.
        //
        OptionLength = *((uchar *)Packet->Data + 1) << 3;

        switch (*(uchar *)Packet->Data) {
        case ND_OPTION_MTU: {
            NDOptionMTU *option = (NDOptionMTU *)Packet->Data;
            uint LinkMTU = net_long(option->MTU);

            //
            // Update the LinkMTU field.
            // NB: We rely on loads/stores of this field being atomic.
            //
            if ((LinkMTU >= IPv6_MINIMUM_MTU) &&
                (LinkMTU <= IF->TrueLinkMTU))
                IF->LinkMTU = LinkMTU;
            break;
        }

        case ND_OPTION_PREFIX_INFORMATION: {
            NDOptionPrefixInformation *option =
                (NDOptionPrefixInformation *)Packet->Data;
            uint PrefixLength, ValidLifetime, PreferredLifetime;

            //
            // Silently ignore the link-local prefix.
            // BUGBUG - Is this actually the required check?
            //
            if (IsLinkLocal(&option->Prefix))
                break;

            PrefixLength = option->PrefixLength;
            ValidLifetime = net_long(option->ValidLifetime);
            ValidLifetime = ConvertSecondsToTicks(ValidLifetime);
            PreferredLifetime = net_long(option->PreferredLifetime);
            PreferredLifetime = ConvertSecondsToTicks(PreferredLifetime);

            //
            // Normal case, we will either update the list of on-link
            // prefixes, perform address auto-configuration, or both.
            //

            if (option->Flags & ND_PREFIX_FLAG_ON_LINK)
                PrefixListUpdate(IF, &option->Prefix, PrefixLength,
                                 ValidLifetime);

            if (option->Flags & ND_PREFIX_FLAG_AUTONOMOUS) {
                if (PreferredLifetime > ValidLifetime) {
                    //
                    // MAY log a system management error.
                    //
                    DbgPrint("ND RA: bogus valid/preferred lifetimes\n");
                }
                else if (PrefixLength + INTERFACE_IDENTIFIER_LENGTH != 128) {
                    //
                    // MAY log a system management error.
                    //
                    DbgPrint("ND RA: bogus prefix length\n");
                }
                else {
                    AddrConfUpdate(IF, &option->Prefix, PrefixLength,
                                   ValidLifetime, PreferredLifetime);
                }
            }
            break;
        }
        }

        //
        // Move forward to next option.
        //
        (uchar *)Packet->Data += OptionLength;
        Packet->ContigSize -= OptionLength;
    }

    ReleaseNCE(NCE);
    return 0;
}


//* NeighborSolicitReceive - Handle Neighbor Solicitation messages.
//
//  Validate message, update ND cache, and reply with Neighbor Advertisement.
//  See section 7.2.4 of RFC 1970.
//
int
NeighborSolicitReceive(
    NetTableEntry *RecvNTE,     // NTE that received the packet.
    IPv6Packet *Packet,         // Packet handed to us by the link layer.
    uchar Type,                 // Value of Type field in ICMPv6 header.
    uchar Code)                 // Value of Code field in ICMPv6 header.
{
    Interface *IF = RecvNTE->IF;
    IPv6Addr *TargetAddress;
    void *SourceLinkAddress;
    NDIS_STATUS Status;
    NDIS_PACKET *AdvertPacket;
    NDIS_BUFFER *Buffer;
    uint Offset;
    void *Mem;
    int MemLen;
    IPv6Header *AdvertIP;
    ICMPv6Header *AdvertICMP;
    ulong Flags;
    void *AdvertTargetOption;
    IPv6Addr *AdvertTargetAddress;
    void *DestLinkAddress;
    NetTableEntry *TargetNTE;
    uint TargetType, TargetScope;

    //
    // Validate the solicitation.
    // By the time we get here, any IPv6 Authentication Header will have
    // already been checked, as will have the ICMPv6 checksum.  Still need
    // to check the IP Hop Limit, and the ICMP code and length.
    //
    if (Packet->IP->HopLimit != 255) {
        //
        // Packet was forwarded by a router, therefore it cannot be
        // from a legitimate neighbor.  Drop the packet.
        //
        DbgPrint("IPv6 ND: Received a routed neighbor solicitation\n");
        return 0;
    }
    if ((Code != 0) || (Packet->TotalSize < 20)) {
        //
        // Bogus/corrupted neighbor solicitation message.  Drop the packet.
        //
        DbgPrint("IPv6 ND: Received a corrupted neighbor solicitation\n");
        return 0;
    }

    //
    // Skip over 4 byte "Reserved" field, ignoring whatever may be in it.
    // Get a pointer to the target address which follows, then skip over that.
    //
    (uchar *)Packet->Data += 4;
    TargetAddress = (IPv6Addr *) Packet->Data;
    (uchar *) Packet->Data += sizeof(IPv6Addr);
    Packet->ContigSize -= 4 + sizeof(IPv6Addr);
    Packet->TotalSize -= 4 + sizeof(IPv6Addr);

    //
    // See if we're the target of the solicitation.  If the target address is
    // not a unicast or anycast address assigned to receiving interface,
    // then we must silently drop the packet.
    //
    TargetNTE = FindAddressOnInterface(IF, TargetAddress,
                                       &TargetType, &TargetScope);
    if (TargetNTE == NULL) {
        DbgPrint("IPv6 ND: Did not find target address\n");
        return 0;
    }
    else if (TargetType == ADE_MULTICAST) {
        DbgPrint("IPv6 ND: Target address is not uni/anycast\n");
        goto Return;
    }

    //
    // We may have a source link-layer address option present.
    // Check for it and silently ignore all others.
    //
    // BUGBUG: This code assumes that options fit within current packet buffer.
    //
    SourceLinkAddress = NULL;
    while (Packet->TotalSize) {
        uchar OptionLength;

        //
        // Validate the option length.
        //
        if ((Packet->TotalSize < 8) ||
            ((OptionLength = *((uchar *)Packet->Data + 1) << 3) == 0) ||
            (OptionLength > Packet->TotalSize)) {
            //
            // Invalid option length.  Drop the packet.
            //
            DbgPrint("IPv6 ND: Received option with bogus length\n");
            goto Return;
        }

        if (*((uchar *)Packet->Data) == ND_OPTION_SOURCE_LINK_LAYER_ADDRESS) {
            //
            // Parse the link-layer address option.
            //
            SourceLinkAddress = (*IF->ReadLLOpt)(IF->LinkContext,
                                                 (uchar *)Packet->Data);
            if (SourceLinkAddress == NULL) {
                //
                // Invalid option format. Drop the packet.
                //
                DbgPrint("IPv6 ND: Received bogus ll option\n");
                goto Return;
            }
            //
            // Note that if there are multiple options for some bogus reason,
            // we use the last one. We must sanity-check all option lengths.
            //
        }

        //
        // Move forward to next option.
        //
        (uchar *)Packet->Data += OptionLength;
        Packet->ContigSize -= OptionLength;
        Packet->TotalSize -= OptionLength;
    }

    //
    // We've received and parsed a valid neighbor solicitation.
    //
    // First, we check our Duplicate Address Detection state.
    // (See RFC 1971 section 5.4.3.)
    //
    if (!IsValidNTE(TargetNTE)) {

        if ((TargetNTE->DADState == DAD_STATE_TENTATIVE) &&
            IP6_ADDR_EQUAL(TargetAddress, &TargetNTE->Address) &&
            IsUnspecified(&Packet->IP->Source)) {
            //
            // This appears to be a DAD solicit for our own tentative address.
            // However it might be a loopback of the DAD solicit
            // that we sent ourselves. Use DADCountLB to detect this.
            //
            if (TargetNTE->DADCountLB == 0) {
                //
                // We've received more DAD solicits than we sent.
                // Must be a duplicate address out there.
                //
                TargetNTE->DADState = DAD_STATE_DUPLICATE;
                TargetNTE->DADTimer = 0;
                DbgPrint("NeighborSolicitReceive: DAD found duplicate!\n");
            }
            else {
                //
                // Remember that we received a DAD solicit,
                // presumably a loopback of our own packet.
                //
                TargetNTE->DADCountLB--;
            }
        }

        //
        // In any case, we should not respond to the solicit.
        //
        goto Return;
    }

    if (!IsValidNTE(RecvNTE))
        goto Return;

    //
    // Update the Neighbor Cache Entry for the source of this solicit.
    // In this case, only bother if the SourceLinkAddress is present;
    // if it's not, NeighborCacheUpdate won't do anything.
    //
    if (!IsUnspecified(&Packet->IP->Source) && (SourceLinkAddress != NULL)) {
        NeighborCacheEntry *NCE;

        NCE = FindOrCreateNeighbor(IF, &Packet->IP->Source);
        if (NCE != NULL) {
            NeighborCacheUpdate(NCE, SourceLinkAddress, FALSE);
            ReleaseNCE(NCE);
        }
    }

    //
    // Send a solicited neighbor advertisement back to the source.
    //
    // Interface to send on is 'IF' (I/F solicitation received on).
    //

    //
    // Allocate a packet header for advertisement.
    //
    NdisAllocatePacket(&Status, &AdvertPacket, IPv6PacketPool);
    if (Status != NDIS_STATUS_SUCCESS) {
        DbgPrint("IPv6 ND: Couldn't allocate packet header!?!\n");
        goto Return;
    }
    PC(AdvertPacket)->CompletionHandler = IPv6FreePacket;

    //
    // Allocate memory for buffer.
    // BUGBUG: We're assuming here that there are no intermediate
    // BUGBUG: headers we'll need to put in our advertisement message.
    // In addition to the 24 bytes for the solicit proper, leave space
    // for target link-layer address option (round option length up to
    // 8-byte multiple).
    //
    Packet->TotalSize = 24 + ((IF->LinkAddressLength + 2 + 7) & ~7);
    Offset = IF->LinkHeaderSize;
    MemLen = Offset + sizeof(IPv6Header) + Packet->TotalSize;
    Mem = ExAllocatePool(NonPagedPool, MemLen);
    if (Mem == NULL) {
        DbgPrint("IPv6 ND: Couldn't allocate memory!?!\n");
        NdisFreePacket(AdvertPacket);
        goto Return;
    }
    
    //
    // Allocate a buffer header and put it on the packet chain.
    //
    NdisAllocateBuffer(&Status, &Buffer, IPv6BufferPool, Mem, MemLen);
    if (Status != NDIS_STATUS_SUCCESS) {
        DbgPrint("IPv6 ND: Couldn't allocate buffer header!?!\n");
        ExFreePool(Mem);
        NdisFreePacket(AdvertPacket);
        goto Return;
    }
    NdisChainBufferAtFront(AdvertPacket, Buffer);

    //
    // Prepare IP header of advertisement.
    //
    AdvertIP = (IPv6Header *)((uchar *)Mem + Offset);
    AdvertIP->VersClassFlow = IP_VERSION;
    AdvertIP->PayloadLength = net_short((ushort)Packet->TotalSize);
    AdvertIP->NextHeader = IP_PROTOCOL_ICMPv6;
    AdvertIP->HopLimit = 255;

    //
    // Prepare ICMP header.
    //
    AdvertICMP = (ICMPv6Header *)(AdvertIP + 1);
    AdvertICMP->Type = ICMPv6_NEIGHBOR_ADVERT;
    AdvertICMP->Code = 0;
    AdvertICMP->Checksum = 0;
    Flags = 0;
    if (RouterConfigured)
        Flags |= ND_NA_FLAG_ROUTER;

    //
    // Check source of solicitation to see where we should direct our
    // advertisement (and determine what type of advertisement to send).
    //
    if (IsUnspecified(&Packet->IP->Source)) {
        //
        // Solicitation came from an unspecified address (presumably a node
        // undergoing initialization), so we need to multicast our response.
        // We also don't set the solicited flag since we can't specify the
        // specific node our advertisement is directed at.
        //
        AdvertIP->Dest = AllNodesOnLinkAddr;

        DestLinkAddress = alloca(IF->LinkAddressLength);
        (*IF->ConvertMCastAddr)(IF->LinkContext, &AllNodesOnLinkAddr,
                                DestLinkAddress);
    } else {
        //
        // We know who sent the solicitation, so we can respond by unicasting
        // our solicited advertisement back to the soliciter.
        //
        AdvertIP->Dest = Packet->IP->Source;
        Flags |= ND_NA_FLAG_SOLICITED;

        //
        // Find link-level address for above.  We should have it cached (note
        // that it will be cached if it just came in on this solicitation).
        //
        if (SourceLinkAddress != NULL) {
            //
            // This is faster than checking the ND cache and
            // it should be the common case.
            //
            DestLinkAddress = SourceLinkAddress;
        }
        else {
            DestLinkAddress = alloca(IF->LinkAddressLength);

            if (!NeighborCacheSearch(IF, &Packet->IP->Source, DestLinkAddress)) {
                //
                // No entry found in ND cache.
                // REVIEW: Queuing for ND does not seem correct -
                // just drop the solicit.
                //
                DbgPrint("IPv6 ND: Didn't find NCE!?!\n");
                ExFreePool(Mem);
                NdisFreeBuffer(Buffer);
                NdisFreePacket(AdvertPacket);
                goto Return;
            }
        }
    }

    //
    // Take our source address from the target NTE.
    //
    AdvertIP->Source = TargetNTE->Address;

    //
    // Take the target address from the solicitation.
    //
    AdvertTargetAddress = (IPv6Addr *)((char *)AdvertICMP +
                                       sizeof(ICMPv6Header) + sizeof(ulong));
    *AdvertTargetAddress = *TargetAddress;

    //
    // Always include target link-layer address option.
    //
    AdvertTargetOption = (void *)((char *)AdvertTargetAddress +
                                  sizeof(IPv6Addr));

    ((uchar *)AdvertTargetOption)[0] = ND_OPTION_TARGET_LINK_LAYER_ADDRESS;
    ((uchar *)AdvertTargetOption)[1] = (IF->LinkAddressLength + 2 + 7) >> 3;

    (*IF->WriteLLOpt)(IF->LinkContext, AdvertTargetOption, IF->LinkAddress);

    //
    // We set the override flag unless the target is an anycast address.
    //
    if (TargetType != ADE_ANYCAST)
        Flags |= ND_NA_FLAG_OVERRIDE;

    //
    // Fill in Flags field in advertisement.
    //
    *(ulong *)((char *)AdvertICMP + sizeof(ICMPv6Header)) = net_long(Flags);

    //
    // Calculate the ICMPv6 checksum.  It covers the entire ICMPv6 message
    // starting with the ICMPv6 header, plus the IPv6 pseudo-header.
    //
    AdvertICMP->Checksum = ChecksumPacket(
        AdvertPacket, Offset + sizeof *AdvertIP, NULL, Packet->TotalSize,
        AdvertIP, IP_PROTOCOL_ICMPv6);

    //
    // Call the interface's send routine to transmit the packet.
    //
    // BUGBUG: For anycast target addresses, should delay.
    //
    (*IF->Transmit)(IF->LinkContext,
                    AdvertPacket, Offset, DestLinkAddress);

  Return:
    ReleaseNTE(TargetNTE);
    return 0;
}


//* NeighborAdvertReceive - Handle Neighbor Advertisement messages.
//
// Validate message and update neighbor cache if necessary.
//
int
NeighborAdvertReceive(
    NetTableEntry *RecvNTE,     // NTE that received the packet.
    IPv6Packet *Packet,         // Packet handed to us by ICMPv6Receive.
    uchar Type,                 // Value of Type field in ICMPv6 header.
    uchar Code)                 // Value of Code field in ICMPv6 header.
{
    Interface *IF = RecvNTE->IF;
    ulong Flags;
    IPv6Addr *TargetAddress;
    void *TargetLinkAddress;
    NeighborCacheEntry *NCE;
    NetTableEntry *TargetNTE;
    uint TargetType, TargetScope;

    //
    // Validate the advertisement.
    // By the time we get here, any IPv6 Authentication Header will have
    // already been checked, as will have the ICMPv6 checksum.  Still need
    // to check the IP Hop Limit, and the ICMP code and length.
    //
    if (Packet->IP->HopLimit != 255) {
        //
        // Packet was forwarded by a router, therefore it cannot be
        // from a legitimate neighbor.  Drop the packet.
        //
        DbgPrint("IPv6 ND: Received a routed neighbor advertisement\n");
        return 0;
    }
    if ((Code != 0) || (Packet->TotalSize < 20)) {
        //
        // Bogus/corrupted neighbor advertisement message.  Drop the packet.
        //
        DbgPrint("IPv6 ND: Received a corrupted neighbor advertisement\n");
        return 0;
    }

    //
    // Next 4 bytes contains 3 bits of flags and 29 bits of "Reserved".
    // Pull this out as a 32 bit long, we're required to ignore the "Reserved"
    // bits.  Get a pointer to the target address which follows, then skip
    // over that.
    //
    Flags = net_long(*(ulong *)Packet->Data);
    (uchar *)Packet->Data += 4;
    TargetAddress = (IPv6Addr *)Packet->Data;
    (uchar *)Packet->Data += sizeof(IPv6Addr);
    Packet->ContigSize -= 4 + sizeof(IPv6Addr);
    Packet->TotalSize -= 4 + sizeof(IPv6Addr);

    //
    // Check that the target address is not a multicast address.
    //
    if (IsMulticast(TargetAddress)) {
        //
        // Nodes are not supposed to send neighbor advertisements for
        // multicast addresses.  We're required to drop any such advertisements
        // we receive.
        //
        DbgPrint("IPv6 ND: Received advertisement for a multicast address\n");
        return 0;
    }

    //
    // Look for a target link-layer address option.
    //
    // BUGBUG: This code assumes that options fit within current packet buffer.
    //
    TargetLinkAddress = NULL;
    while (Packet->TotalSize) {
        uchar OptionLength;

        //
        // Validate the option length.
        //
        if ((Packet->TotalSize < 8) ||
            ((OptionLength = *((uchar *)Packet->Data + 1) << 3) == 0) ||
            (OptionLength > Packet->TotalSize)) {
            //
            // Invalid option length.  Drop the packet.
            //
            DbgPrint("IPv6 ND: Received option with bogus length\n");
            return 0;
        }

        if (*((uchar *)Packet->Data) == ND_OPTION_TARGET_LINK_LAYER_ADDRESS) {
            //
            // Parse the link-layer address option.
            //
            TargetLinkAddress = (*IF->ReadLLOpt)(IF->LinkContext,
                                                 (uchar *)Packet->Data);
            if (TargetLinkAddress == NULL) {
                //
                // Invalid option format. Drop the packet.
                //
                DbgPrint("IPv6 ND: Received bogus ll option\n");
                return 0;
            }
            //
            // Note that if there are multiple options for some bogus reason,
            // we use the last one. We must sanity-check all option lengths.
            //
        }
        //
        // Move forward to next option.
        //
        (uchar *)Packet->Data += OptionLength;
        Packet->ContigSize -= OptionLength;
        Packet->TotalSize -= OptionLength;
    }

    //
    // We've received and parsed a valid neighbor advertisement.
    //
    // First, we check our Duplicate Address Detection state.
    // (See RFC 1971 section 5.4.4.)
    //
    TargetNTE = FindAddressOnInterface(IF, TargetAddress,
                                       &TargetType, &TargetScope);
    if (TargetNTE != NULL) {
        if (!IsValidNTE(TargetNTE)) {
            if ((TargetNTE->DADState == DAD_STATE_TENTATIVE) &&
                IP6_ADDR_EQUAL(TargetAddress, &TargetNTE->Address)) {
                //
                // Someone out there is already using our address;
                // they responded to our solicit.
                //
                TargetNTE->DADState = DAD_STATE_DUPLICATE;
                TargetNTE->DADTimer = 0;
                DbgPrint("NeighborAdvertReceive: DAD found duplicate!\n");
            }

            //
            // In any case, no further processing is required.
            //
            ReleaseNTE(TargetNTE);
            return 0;
        }

        ReleaseNTE(TargetNTE);
    }

    if (!IsValidNTE(RecvNTE))
        return 0;

    //
    // Process the advertisement.
    //
    NeighborCacheAdvert(IF, TargetAddress, TargetLinkAddress, Flags);
    return 0;
}


//* RedirectReceive - Handle Redirect messages.
//
//  See RFC 1970, sections 8.1 and 8.3.
//
int
RedirectReceive(
    NetTableEntry *RecvNTE,  // NTE that received the packet.
    IPv6Packet *Packet,      // Packet handed to us by ICMPv6Receive.
    uchar Type,              // Value of Type field in ICMPv6 header.
    uchar Code)              // Value of Code field in ICMPv6 header.
{
    Interface *IF = RecvNTE->IF;
    IPv6Addr *TargetAddress;
    void *TargetLinkAddress;
    IPv6Addr *DestinationAddress;
    NeighborCacheEntry *TargetNCE;

    //
    // Validate the redirect.
    // By the time we get here, any IPv6 Authentication Header will have
    // already been checked, as will have the ICMPv6 checksum.  Still need
    // to check the IP Hop Limit, and the ICMP code and length.
    //
    if (Packet->IP->HopLimit != 255) {
        //
        // Packet was forwarded by a router, therefore it cannot be
        // from a legitimate neighbor.  Drop the packet.
        //
        DbgPrint("IPv6 ND: Received a routed redirect\n");
        return 0;
    }
    if (Code != 0) {
        //
        // Bogus/corrupted redirect message.  Drop the packet.
        //
        DbgPrint("IPv6 ND: Received a corrupted redirect\n");
        return 0;
    }

    //
    // Check that the source address is a link-local address.
    // We need a link-local source address to identify the router
    // that sent the redirect.
    //
    if (!IsLinkLocal(&Packet->IP->Source)) {
        //
        // Drop the packet.
        //
        DbgPrint("IPv6 ND: Received redirect from non-link-local source\n");
        return 0;
    }

    //
    // Skip over the 4 byte reserved field.
    //
    (uchar *)Packet->Data += 4;
    Packet->ContigSize -= 4;
    Packet->TotalSize -= 4;
    Packet->Position += 4;

    //
    // Next 16 bytes contain the target address, and
    // after that the 16 byte redirected destination address.
    //
    // REVIEW: Tradeoff between single pullup check for both and
    // REVIEW: possible pullup avoidance of doing each individually.
    //
    if (Packet->ContigSize < 2 * sizeof(IPv6Addr))
        if (PacketPullup(Packet, 2 * sizeof(IPv6Addr)) == 0) {
            //
            // Packet too short to contain minimal redirect.
            //
            DbgPrint("IPv6 ND: Received a too short redirect\n");
            return 0;
        }
    TargetAddress = ((IPv6Addr *)Packet->Data)++;
    DestinationAddress = ((IPv6Addr *)Packet->Data)++;
    Packet->ContigSize -= 2 * sizeof(IPv6Addr);
    Packet->TotalSize  -= 2 * sizeof(IPv6Addr);
    Packet->Position += 2 * sizeof(IPv6Addr);

    //
    // Check that the destination address is not a multicast address.
    //
    if (IsMulticast(DestinationAddress)) {
        //
        // Drop the packet.
        //
        DbgPrint("IPv6 ND: Received redirect for multicast address\n");
        return 0;
    }

    //
    // Check that either the target address is link-local
    // (redirecting to a router) or the target and the destination
    // are the same (redirecting to an on-link destination).
    //
    if (!IsLinkLocal(TargetAddress) &&
        !IP6_ADDR_EQUAL(TargetAddress, DestinationAddress)) {
        //
        // Drop the packet.
        //
        DbgPrint("IPv6 ND: Received redirect with non-link-local target\n");
        return 0;
    }

    //
    // The code below assumes a contigous buffer for all the options
    // (the remainder of the packet).  If that isn't currently the
    // case, do a pullup for the whole thing.
    //
    // REVIEW: This is less than perfectly efficient, but we shouldn't
    // REVIEW: do it often.  Rework to do check/pullup on per-field basis?
    //
    if (Packet->ContigSize != Packet->TotalSize)
        if (PacketPullup(Packet, Packet->TotalSize) == 0) {
            // Can only fail if we run out of memory.
            return 0;
        }

    ASSERT(Packet->ContigSize == Packet->TotalSize);

    //
    // Look for a target link-layer address option,
    // checking that all included options have a valid length.
    //
    TargetLinkAddress = NULL;
    while (Packet->ContigSize) {
        uchar OptionLength;

        //
        // Validate the option length.
        //
        if ((Packet->ContigSize < 8) ||
            ((OptionLength = *((uchar *)Packet->Data + 1) << 3) == 0) ||
            (OptionLength > Packet->ContigSize)) {
            //
            // Invalid option length.  Drop the packet.
            //
            DbgPrint("IPv6 ND: Received option with bogus length\n");
            return 0;
        }

        if (*(uchar *)Packet->Data == ND_OPTION_TARGET_LINK_LAYER_ADDRESS) {
            //
            // Parse the link-layer address option.
            //
            TargetLinkAddress = (*IF->ReadLLOpt)(IF->LinkContext,
                                                 (uchar *)Packet->Data);
            if (TargetLinkAddress == NULL) {
                //
                // Invalid option format. Drop the packet.
                //
                DbgPrint("IPv6 ND: Received bogus ll option\n");
                return 0;
            }
            //
            // Note that if there are multiple options for some bogus reason,
            // we use the last one. We must sanity-check all option lengths.
            //
        }
        //
        // Move forward to next option.
        //
        (uchar *)Packet->Data += OptionLength;
        Packet->ContigSize -= OptionLength;
    }

    //
    // We have a valid redirect message (except for checking that the
    // source of the redirect is the current first-hop neighbor
    // for the destination - RedirectRouteCache does that).
    //
    // First we update the neighbor cache, getting an NCE for the target
    // of the redirect.  Then we update the route cache.
    //

    //
    // Get the Neighbor Cache Entry for the target.
    //
    TargetNCE = FindOrCreateNeighbor(IF, TargetAddress);
    if (TargetNCE == NULL) {
        //
        // Couldn't find or create NCE. Drop the packet.
        //
        return 0;
    }

    //
    // Update the Neighbor Cache Entry for the target.
    // We know the target is a router if the target address
    // is not the same as the destination address.
    //
    NeighborCacheUpdate(TargetNCE, TargetLinkAddress,
                        !IP6_ADDR_EQUAL(TargetAddress, DestinationAddress));

    //
    // Update the route cache to reflect this redirect.
    //
    RedirectRouteCache(&Packet->IP->Source, DestinationAddress,
                       RecvNTE, TargetNCE);

    ReleaseNCE(TargetNCE);
    return 0;
}


//* RouterSolicitSend
//
//  Sends a Router Solicitation.
//  The solicit is always sent to the all-routers multicast address.
//  Chooses a valid source address for the interface.
//
void
RouterSolicitSend(Interface *IF)
{
    NDIS_STATUS Status;
    NDIS_PACKET *Packet;
    NDIS_BUFFER *Buffer;
    int PayloadLength;
    uint Offset;
    void *Mem;
    int MemLen;
    IPv6Header *IP;
    ICMPv6Header *ICMP;
    void *SourceOption;
    void *LLDest;

    //
    // Allocate a header for solicitation.
    //
    NdisAllocatePacket(&Status, &Packet, IPv6PacketPool);
    if (Status != NDIS_STATUS_SUCCESS) {
        DbgPrint("IPv6 ND: Couldn't allocate packet header!?!\n");
        return;
    }
    PC(Packet)->CompletionHandler = IPv6FreePacket;

    //
    // Allocate memory for buffer.
    // BUGBUG: We're assuming here that there are no intermediate
    // BUGBUG: headers we'll need to put in our solicitation message.
    // In addition to the 8 bytes for the solicit proper, leave space
    // for source link-layer address option (round option length up to
    // 8-byte multiple).
    //
    PayloadLength = 8 + ((IF->LinkAddressLength + 2 + 7) & ~7);
    Offset = IF->LinkHeaderSize;
    MemLen = Offset + sizeof(IPv6Header) + PayloadLength;
    Mem = ExAllocatePool(NonPagedPool, MemLen);
    if (Mem == NULL) {
        DbgPrint("IPv6 ND: Couldn't allocate memory!?!\n");
        NdisFreePacket(Packet);
        return;
    }

    //
    // Allocate a buffer header and put it on the packet chain.
    //
    NdisAllocateBuffer(&Status, &Buffer, IPv6BufferPool, Mem, MemLen);
    if (Status != NDIS_STATUS_SUCCESS) {
        DbgPrint("IPv6 ND: Couldn't allocate buffer header!?!\n");
        ExFreePool(Mem);
        NdisFreePacket(Packet);
        return;
    }
    NdisChainBufferAtFront(Packet, Buffer);

    //
    // Prepare IP header of solicitation.
    //
    IP = (IPv6Header *)((uchar *)Mem + Offset);
    IP->VersClassFlow = IP_VERSION;
    IP->PayloadLength = net_short((ushort)PayloadLength);
    IP->NextHeader = IP_PROTOCOL_ICMPv6;
    IP->HopLimit = 255;

    //
    // If the interface does not have a valid address yet,
    // use the unspecified address. FindInterfaceAddress
    // returns the link-local address if it is valid.
    //
    if (!FindInterfaceAddress(IF, &IP->Source))
        IP->Source = UnspecifiedAddr;
    IP->Dest = AllRoutersOnLinkAddr;

    //
    // Prepare ICMP header.
    //
    ICMP = (ICMPv6Header *)(IP + 1);
    ICMP->Type = ICMPv6_ROUTER_SOLICIT;
    ICMP->Code = 0;
    ICMP->Checksum = 0;

    //
    // Must zero the reserved field.
    //
    *(ulong *)(ICMP + 1) = 0;

    //
    // Always include source link-layer address option.
    //
    SourceOption = (void *)((ulong *)(ICMP + 1) + 1);

    ((uchar *)SourceOption)[0] = ND_OPTION_SOURCE_LINK_LAYER_ADDRESS;
    ((uchar *)SourceOption)[1] = (IF->LinkAddressLength + 2 + 7) >> 3;

    (*IF->WriteLLOpt)(IF->LinkContext, SourceOption, IF->LinkAddress);

    //
    // Calculate the ICMPv6 checksum.  It covers the entire ICMPv6 message
    // starting with the ICMPv6 header, plus the IPv6 pseudo-header.
    //
    ICMP->Checksum = ChecksumPacket(
        Packet, Offset + sizeof *IP, NULL, PayloadLength,
        IP, IP_PROTOCOL_ICMPv6);

    //
    // Calculate the link-level destination address.
    // (The IPv6 destination is a multicast address.)
    //
    LLDest = alloca(IF->LinkAddressLength);
    (*IF->ConvertMCastAddr)(IF->LinkContext, &IP->Dest, LLDest);

    //
    // Call the interface's send routine to transmit the packet.
    //
    (*IF->Transmit)(IF->LinkContext, Packet, Offset, LLDest);
}

//* RouterSolicitTimeout
//
//  Called periodically from IPv6Timeout to handle timeouts
//  for sending router solications for an interface.
//
//  Callable from DPC context, not from thread context.
//
void
RouterSolicitTimeout(Interface *IF)
{
    int SendRouterSolicit = FALSE;

    KeAcquireSpinLockAtDpcLevel(&IF->Lock);

    if (IF->RSTimer != 0) {
        //
        // Timer is running.  Decrement and check for expiration.
        //
        if (--IF->RSTimer == 0) {
            //
            // Timer went off. Generate a router solicitation.
            //
            IF->RSCount++;
            SendRouterSolicit = TRUE;

            //
            // Re-arm the timer, if we should keep trying.
            //
            if (IF->RSCount < MAX_RTR_SOLICITATIONS)
                IF->RSTimer = RTR_SOLICITATION_INTERVAL;
        }
    }

    KeReleaseSpinLockFromDpcLevel(&IF->Lock);

    if (SendRouterSolicit)
        RouterSolicitSend(IF);
}

//* NeighborSolicitSend0
//
//  Low-level version of NeighborSolicitSend -
//  uses explicit source/destination/target addresses
//  instead of an NCE.
//
void
NeighborSolicitSend0(
    Interface *IF,      // Interface for the solicit
    void *LLDest,       // Link-level destination
    IPv6Addr *Source,   // IP-level source
    IPv6Addr *Dest,     // IP-level destination
    IPv6Addr *Target)   // IP-level target of the solicit
{
    NDIS_STATUS Status;
    NDIS_PACKET *Packet;
    NDIS_BUFFER *Buffer;
    int PayloadLength;
    uint Offset;
    void *Mem;
    int MemLen;
    IPv6Header *IP;
    ICMPv6Header *ICMP;
    IPv6Addr *TargetAddress;
    void *SourceOption;

    //
    // Allocate a header for solicitation.
    //
    NdisAllocatePacket(&Status, &Packet, IPv6PacketPool);
    if (Status != NDIS_STATUS_SUCCESS) {
        DbgPrint("IPv6 ND: Couldn't allocate packet header!?!\n");
        return;
    }
    PC(Packet)->CompletionHandler = IPv6FreePacket;

    //
    // Allocate memory for buffer.
    // BUGBUG: We're assuming here that there are no intermediate
    // BUGBUG: headers we'll need to put in our solicitation message.
    // In addition to the 24 bytes for the solicit proper, leave space
    // for source link-layer address option (round option length up to
    // 8-byte multiple).
    //
    PayloadLength = 24 + ((IF->LinkAddressLength + 2 + 7) & ~7);
    Offset = IF->LinkHeaderSize;
    MemLen = Offset + sizeof(IPv6Header) + PayloadLength;
    Mem = ExAllocatePool(NonPagedPool, MemLen);
    if (Mem == NULL) {
        DbgPrint("IPv6 ND: Couldn't allocate memory!?!\n");
        NdisFreePacket(Packet);
        return;
    }
    
    //
    // Allocate a buffer header and put it on the packet chain.
    //
    NdisAllocateBuffer(&Status, &Buffer, IPv6BufferPool, Mem, MemLen);
    if (Status != NDIS_STATUS_SUCCESS) {
        DbgPrint("IPv6 ND: Couldn't allocate buffer header!?!\n");
        ExFreePool(Mem);
        NdisFreePacket(Packet);
        return;
    }
    NdisChainBufferAtFront(Packet, Buffer);

    //
    // Prepare IP header of solicitation.
    //
    IP = (IPv6Header *)((uchar *)Mem + Offset);
    IP->VersClassFlow = IP_VERSION;
    IP->PayloadLength = net_short((ushort)PayloadLength);
    IP->NextHeader = IP_PROTOCOL_ICMPv6;
    IP->HopLimit = 255;
    IP->Source = *Source;
    IP->Dest = *Dest;

    //
    // Prepare ICMP header.
    //
    ICMP = (ICMPv6Header *)(IP + 1);
    ICMP->Type = ICMPv6_NEIGHBOR_SOLICIT;
    ICMP->Code = 0;
    ICMP->Checksum = 0;

    //
    // Must zero the reserved field.
    //
    *(ulong *)(ICMP + 1) = 0;

    //
    // Initialize the target address.
    //
    TargetAddress = (IPv6Addr *)((ulong *)(ICMP + 1) + 1);
    *TargetAddress = *Target;

    //
    // Always include source link-layer address option.
    //
    SourceOption = (void *)(TargetAddress + 1);

    ((uchar *)SourceOption)[0] = ND_OPTION_SOURCE_LINK_LAYER_ADDRESS;
    ((uchar *)SourceOption)[1] = (IF->LinkAddressLength + 2 + 7) >> 3;

    (*IF->WriteLLOpt)(IF->LinkContext, SourceOption, IF->LinkAddress);

    //
    // Calculate the ICMPv6 checksum.  It covers the entire ICMPv6 message
    // starting with the ICMPv6 header, plus the IPv6 pseudo-header.
    //
    ICMP->Checksum = ChecksumPacket(
        Packet, Offset + sizeof *IP, NULL, PayloadLength,
        IP, IP_PROTOCOL_ICMPv6);

    //
    // Call the interface's send routine to transmit the packet.
    //
    // BUGBUG: For anycast target addresses, should delay.
    //
    (*IF->Transmit)(IF->LinkContext, Packet, Offset, LLDest);
}

//* NeighborSolicitSend - Send a Neighbor Solicitation message.
//
//  Chooses source/destination/target addresses depending
//  on the NCE state and sends the solicit packet.
//
void
NeighborSolicitSend(
    NeighborCacheEntry *NCE,  // Neighbor to solicit.
    IPv6Addr *Source)         // Source address to send from (optional).
{
    Interface *IF = NCE->IF;
    IPv6Addr Dest;
    void *LLDest;

    //
    // Check Neighbor Discovery protocol state of our neighbor in order to
    // determine whether we should multicast or unicast our solicitation.
    //
    if (NCE->NDState == ND_STATE_INCOMPLETE) {
        //
        // This is our initial solicitation to this neighbor, so we don't
        // have a link-layer address cached.  Multicast our solicitation
        // to the solicited-node multicast address corresponding to our
        // neighbor's address.
        //
        CreateSolicitedNodeMulticastAddress(&NCE->NeighborAddress, &Dest);

        LLDest = alloca(IF->LinkAddressLength);
        (*IF->ConvertMCastAddr)(IF->LinkContext, &Dest, LLDest);
    } else {
        //
        // We have a cached link-layer address that has gone stale.
        // Probe this address via a unicast solicitation.
        //
        Dest = NCE->NeighborAddress;

        LLDest = NCE->LinkAddress;
    }

    //
    // If we were given a specific source address to use then do so (normal
    // case for our initial multicasted solicit), otherwise use any one
    // of our outgoing interface's that is handy.
    //
    if (Source == NULL)
        Source = &IF->NTE->Address;

    //
    // Now that we've determined the source/destination/target addresses,
    // we can actually send the solicit.
    //
    NeighborSolicitSend0(IF, LLDest, Source, &Dest, &NCE->NeighborAddress);
}


//* DADSolicitSend - Sends a DAD Neighbor Solicit.
//
//  Like NeighborSolicitSend, but specialized for DAD.
//
void
DADSolicitSend(NetTableEntry *NTE)
{
    Interface *IF = NTE->IF;
    IPv6Addr Dest;
    void *LLDest;

    //
    // Multicast our solicitation to the solicited-node multicast
    // address corresponding to our own address.
    //
    CreateSolicitedNodeMulticastAddress(&NTE->Address, &Dest);

    //
    // Convert the IP-level multicast destination address
    // to a link-level multicast address.
    //
    LLDest = alloca(IF->LinkAddressLength);
    (*IF->ConvertMCastAddr)(IF->LinkContext, &Dest, LLDest);

    //
    // Now that we've created the destination addresses,
    // we can actually send the solicit.
    //
    NeighborSolicitSend0(IF, LLDest, &UnspecifiedAddr, &Dest, &NTE->Address);
}


//* DADTimeout - handle a Duplicate Address Detection timeout
//
//  Called from IPv6Timeout to handle DAD timeouts on an NTE.
//
void
DADTimeout(NetTableEntry *NTE)
{
    Interface *IF = NTE->IF;
    int SendDADSolicit = FALSE;

    KeAcquireSpinLockAtDpcLevel(&IF->Lock);

    if (NTE->DADTimer != 0) {
        //
        // Timer is running.  Decrement and check for expiration.
        //
        if (--NTE->DADTimer == 0) {
            //
            // Timer went off.
            //

            ASSERT(NTE->DADState == DAD_STATE_TENTATIVE);

            if (NTE->DADCount == 0) {
                //
                // The address has passed Duplicate Address Detection.
                // Transition to a valid state.
                //
                NTE->DADState = DAD_STATE_PREFERRED;
            }
            else {
                //
                // Time to send another solicit.
                //
                NTE->DADCount--;
                NTE->DADCountLB++;
                NTE->DADTimer = IF->RetransTimer;
                SendDADSolicit = TRUE;
            }
        }
    }

    KeReleaseSpinLockFromDpcLevel(&IF->Lock);

    if (SendDADSolicit)
        DADSolicitSend(NTE);
}

//* CalcReachableTime
//
//  Calculate a pseudo-random ReachableTime from BaseReachableTime
//  (this prevents synchronization of Neighbor Unreachability Detection
//  messages from different hosts), and convert it to units of system
//  interval ticks (cheaper to do once here than at every packet send).
//
ULONGLONG
CalcReachableTime(
    uint BaseReachableTime)  // Learned from Router Advertisements (in ms).
{
    ULONG Factor;
    ULONGLONG ReachableTime;

    //
    // Calculate a uniformly-distributed random value between
    // MIN_RANDOM_FACTOR and MAX_RANDOM_FACTOR of the BaseReachableTime.
    // To keep the arithmetic integer, *_RANDOM_FACTOR (and thus the
    // 'Factor' variable) are defined as percentage values.
    //
    Factor = MAX_RANDOM_FACTOR - MIN_RANDOM_FACTOR;  // Spread.
    Factor = (ULONG)((ULONGLONG)Random() * Factor >> 32);  // Randomize spread.
    Factor += MIN_RANDOM_FACTOR;

    //
    // Now that we have a random value picked out of our percentage spread,
    // take that percentage of the BaseReachableTime.
    //
    // BaseReachableTime has a maximum value of 3,600,000 milliseconds
    // (see RFC 1970, section 6.2.1), so Factor would have to exeed 1100 %
    // in order to overflow a 32-bit unsigned integer.
    //
    BaseReachableTime = (BaseReachableTime * Factor) / 100;

    //
    // Convert from milliseconds (which is what BaseReachableTime is in) to
    // system interval timer ticks (which is what we keep ReachableTime in).
    //
    // REVIEW: Do we have compiler support for direct 64-bit integer divide on
    // REVIEW: all architectures?  Or should we be using a macro below?
    //
    // REVIEW: Worth it to call KeQueryTimeIncrement just once at init time
    // REVIEW: and cache the value?  Or not?  We don't get here very often.
    //
    ReachableTime = BaseReachableTime * 10000;  // Convert to 100-nanoseconds.
    ReachableTime = ReachableTime / KeQueryTimeIncrement();  // Then to ticks.

    return ReachableTime;
}
