// -*- mode: C++; tab-width: 4; indent-tabs-mode: nil -*- (for GNU Emacs)
//
// Copyright (c) 1985-1998 Microsoft Corporation
//
// This file is part of the Microsoft Research IPv6 Network Protocol Stack.
// You should have received a copy of the Microsoft End-User License Agreement
// for this software along with this release; see the file "license.txt".
// If not, please see http://www.research.microsoft.com/msripv6/license.htm,
// or write to Microsoft Research, One Microsoft Way, Redmond, WA 98052-6399.
//
// Abstract:
//
// Receive routines for Internet Protocol Version 6.
//


#include "oscfg.h"
#include "ndis.h"
#include "ip6imp.h"
#include "ip6def.h"
#include "ip6opts.h"
#include "icmp.h"
#include "route.h"
#include "fragment.h"

uchar StrongHost = TRUE;

KSPIN_LOCK ReassemblyListLock;     // protects Reassembly List
Reassembly *ReassemblyList = NULL;

void ReassembleDatagram(Reassembly *Reass, IPv6Packet *Packet,
                        uint OrigLen, NetTableEntry *NTE);

IPv6Packet *CreateFragmentPacket(Reassembly *Reass);

void SendICMPv6Error(NetTableEntry *RecvNTE, IPv6Packet *Packet, 
                     uint Type, uint Code, uint Pointer);

int OptionProcessing(uchar *Buffer, IPv6Packet *Packet, NetTableEntry *NTE,
                     uint OptLength, Options *Opts);

extern void TCPRcvComplete(void);

extern int ICMPv6Reply(NetTableEntry *RecvNTE, IPv6Packet *Packet,
                       ICMPv6MsgType *Err);

ProtoRecvProc DestinationOptionsReceive;

//* IPv6ReceiveComplete - Handle a receive complete.
//
//  Called by the lower layer when receives are temporarily done.
//
void
IPv6ReceiveComplete(void)
{
    // REVIEW: Original IP implementation had code here to call every
    // REVIEW: UL protocol's receive complete routine (yes, all of them) here.

    TCPRcvComplete();  // BUGBUG: shouldn't be hard-wired.
}


//* IPv6Receive - Receive an incoming IPv6 datagram.
//
//  This is the routine called by the link layer module when an incoming IPv6
//  datagram is to be processed.  We validate the datagram and decide what to
//  do with it.
//
//  NB: The datagram may either be held in a NDIS_PACKET allocated by the
//  link-layer or the interface driver (in which case 'Packet->NdisPacket'
//  is non-NULL and 'Data' points to the first data buffer in the buffer
//  chain), or the datagram may still be held by NDIS (in which case
//  'Packet->NdisPacket' is NULL and 'Data' points to a buffer containing
//  the entire datagram).
//
//  NB: Unlike the IPv4 Receive, we are not told if the packet was sent
//  to a link-level unicast address or not. For example, we do NOT
//  check for link-level multi/broadcasts to IPv6 unicast destinations.
//
int                       // Returns: Reference count of packets held.
IPv6Receive(
    void *MyContext,      // Context value we gave to the link layer.
    IPv6Packet *Packet)   // Packet handed to us by the link layer.
{
    Interface *IF = (Interface *)MyContext;  // Interface received on.
    uint PayloadLength;  // Size of IPv6 payload in bytes.
    NetTableEntry *NTE;
    uint Type, Scope;
    int PacketRefs;
    int (*Handler)();

    //
    // Sanity-check InitialSize & TotalSize.
    // Higher-level code in the receive path relies on these conditions.
    //
    ASSERT((Packet->NdisPacket == NULL) ?
           (Packet->ContigSize == Packet->TotalSize) :
           (Packet->ContigSize <= Packet->TotalSize));

    //
    // Make sure we have enough contiguous bytes for an IPv6 header, otherwise
    // attempt to pullup that amount.  Then stash away a pointer to the header
    // and also remember the offset into the packet at which it begins (needed
    // to calculate an offset for certain ICMP error messages).
    //
    if (Packet->ContigSize < sizeof(IPv6Header)) {
        if (PacketPullup(Packet, sizeof(IPv6Header)) == 0) {
            // Pullup failed.
            KdPrint(("IPv6 Receive: Packet too small to contain IPv6 hdr\n"));
            return 0;  // Drop packet.
        }
    }
    Packet->IP = (IPv6Header UNALIGNED *)Packet->Data;
    Packet->IPPosition = Packet->Position;

    //
    // Check the IP version is correct and that the Hop Limit isn't zero.
    //
    if (((Packet->IP->VersClassFlow & IP_VER_MASK) != IP_VERSION) ||
        (Packet->IP->HopLimit == 0)) {

        // Silently discard the packet.
        KdPrint(("IPv6 Receive: Header error caused packet drop\n"));
        return 0;
    }

    //
    // Grab the Payload Length field out of the packet.
    // Swap bytes if necessary.
    //
    PayloadLength = net_short(Packet->IP->PayloadLength);

    //
    // Check that the amount of payload the IPv6 header thinks is present
    // can actually fit inside the packet data area that Ndis handed us.
    // Note that a Payload Length of zero indicates a Jumbo Payload option.
    //
    if (PayloadLength + sizeof(IPv6Header) > Packet->TotalSize) {
        // Silently discard the packet.
        KdPrint(("IPv6 Receive: Header's PayloadLength doesn't match data\n"));
        return 0;
    }

    //
    // Find an NTE on this interface that can accept the packet.
    //
    NTE = FindAddressOnInterface(IF, &Packet->IP->Dest, &Type, &Scope);

    //
    // If we haven't found an NTE yet, and we're running under the
    // "Weak Host Model" then check addresses assigned to our other interfaces.
    // BUGBUG: Should also skip interfaces of different "site" scope?
    //
    if ((NTE == NULL) && !StrongHost && !IsLinkLocal(&Packet->IP->Dest) &&
        !IsLinkLocalMulticast(&Packet->IP->Dest)) {
        Interface *OtherIF;

        //
        // Check other interfaces.  Don't bother rechecking this interface.
        // BUGBUG - The interface list should be locked.
        //
        for (OtherIF = IFList; OtherIF != NULL; OtherIF = OtherIF->Next) {
            if (OtherIF != IF) {

                NTE = FindAddressOnInterface(OtherIF, &Packet->IP->Dest,
                                             &Type, &Scope);
                if (NTE != NULL)
                    break;
            }
        }
    }

    //
    // If we're a router, figure out if we should forward the packet.
    //
    // Note that if we're a router, some multicast packets should be forwarded
    // as well as received locally.
    //
    // BUGBUG: For now, bail if the packet isn't for us.
    //
    if (NTE == NULL) {
        //
        // Packet's destination doesn't match any of our addresses.
        // Note that this is very frequent with the tunnel interfaces...
        //
#if 0
        DbgPrint("IPv6 Receive: no ADE for dest address %ws\n",
                 FormatV6Address(&IP->Dest));
#endif
        return 0;
    }

    //
    // Packet is for this node.
    // Note: We may only be an intermediate node and not the packet's final
    // destination.
    //

    //
    // Skip over IPv6 header (note we keep our pointer to it).
    //
    // REVIEW: Do this higher up (before call out to forwarding code)?
    // REVIEW: Create a macro/inline for this class of operation?
    //
    Packet->Position += sizeof(IPv6Header);
    (uchar *)Packet->Data += sizeof(IPv6Header);
    Packet->ContigSize -= sizeof(IPv6Header);

    //
    // Check for Jumbo Payload option.
    // REVIEW: Do this higher up (before call out to forwarding code)?
    //
    if (PayloadLength == 0) {
        //
        // Packet's Payload Length field indicates a Jumbo Payload option
        // is present in a Hop-by-Hop options extension header.  Verify that
        // and leave our TotalSize value as the amount physically received
        // until we can get the real payload length from the option.
        //
        // BUGBUG: This only checks that we have a Hop-by-Hop extension
        // BUGBUG: header and not that we have a valid Jumbo Payload option.
        // BUGBUG: Add Packet->Flag to indicate we saw 0 PayloadLength?
        //
        if (Packet->IP->NextHeader != IP_PROTOCOL_HOP_BY_HOP)
            return 0;
        Packet->TotalSize -= sizeof(IPv6Header);
    } else {
        //
        // We don't have a Jumbo Payload option (this is the common case).
        // Set our IPv6 packet size according to the payload length.
        // If the resulting packet is smaller than the link-level packet,
        // ignore the excess data.  It's assumed to be frame padding.
        //
        Packet->TotalSize = PayloadLength;
        if (Packet->ContigSize > PayloadLength)
            Packet->ContigSize = PayloadLength;
    }

    //
    // Switch out to next header handler.
    //
    // NB: We do NOT check NTE->DADStatus here.
    // That is the responsibility of higher-level protocols.
    //
    // REVIEW: Should loop through all next header handlers here,
    // REVIEW: rather than recursively call down in each handler?
    //
    Handler = ProtocolSwitchTable[Packet->IP->NextHeader].DataReceive;

    if (Handler == NULL) {
        //
        // If we don't have a handler for this header type,
        // send an ICMP error message and drop the packet.
        // ICMP Pointer value is the offset from the start of the
        // incoming packet's IPv6 header to the offending field.
        //
        KdPrint(("IPv6 Receive: Next Header type %d isn't supported.\n",
                 Packet->IP->NextHeader));
        SendICMPv6Error(NTE, Packet,
                        ICMPv6_PARAMETER_PROBLEM, 
                        ICMPv6_UNRECOGNIZED_NEXT_HEADER,
                        FIELD_OFFSET(IPv6Header, NextHeader));
        ReleaseNTE(NTE);
        return 0;
    }

    PacketRefs = (*Handler)(NTE, Packet);

    ReleaseNTE(NTE);

    return PacketRefs;
}


//* FragmentReceive - Handle a IPv6 datagram fragment.
// 
//  This is the routine called by IPv6 when it receives a fragment of an
//  IPv6 datagram, i.e. a next header value of 44.  Here we attempt to 
//  reassemble incoming fragments into complete IPv6 datagrams.
//
int
FragmentReceive(
    NetTableEntry *RecvNTE,     // NTE that we received the packet on.
    IPv6Packet *Packet)         // Packet handed to us by link layer.
{
    FragmentHeader *Frag;
    Reassembly *Reass;
    ushort FragOffset;
    PacketShim *Shim, *ThisShim, **MoveShim;
    void *Mem;

    //
    // Check that the receiving NTE is valid.
    // Packets needed for auto-configuration should never have fragment
    // headers, so just drop fragments until we're configured.
    //
    if (!IsValidNTE(RecvNTE)) {
        KdPrint(("FragmentReceive: invalid NTE\n"));
        return 0;  // Drop packet.
    }

    // 
    // If a jumbo payload option was seen, send an ICMP error.
    // Set ICMP pointer to the offset of the fragment header.
    //
    if (Packet->Flags & PACKET_JUMBO_OPTION) {
        SendICMPv6Error(RecvNTE, Packet,
                        ICMPv6_PARAMETER_PROBLEM, 
                        ICMPv6_ERRONEOUS_HEADER_FIELD, 
                        Packet->Position - Packet->IPPosition);
        return 0; // discard packet
    }

    //
    // Verify that we have enough contiguous data to overlay a FragmentHeader
    // structure on the incoming packet.  Then do so.
    //
    if (Packet->ContigSize < sizeof(FragmentHeader)) {
        if (PacketPullup(Packet, sizeof(FragmentHeader)) == 0) {
            // Pullup failed.
            KdPrint(("FragmentReceive: Incoming packet too small"
                     " to contain fragment header\n"));
            return 0;  // Drop packet.
        }
    }
    Frag = (FragmentHeader UNALIGNED *) Packet->Data;
    FragOffset = net_short(Frag->OffsetFlag) & FRAGMENT_OFFSET_MASK;
 

    //
    // Lookup this fragment triple (Source Address, Destination Address, and
    // Identification field) to see if we've already received other fragments
    // of this packet.
    //
    Reass = FragmentLookup(&Packet->IP->Source, &Packet->IP->Dest, Frag->Id);

    if (Reass == NULL) {
        //
        // This is the first fragment of this datagram we've received.
        // Allocate a reassembly structure to keep track of the pieces.
        //
        Reass = ExAllocatePool(NonPagedPool, sizeof(struct Reassembly));
        if (Reass == NULL) {
            DbgPrint("FragmentReceive: Couldn't allocate memory!?!\n");
            return 0;
        }

        RtlCopyMemory((uchar *)&(Reass->IPHdr), (uchar *)Packet->IP, 
                      IPv6_HDR_LEN);
        Reass->Id = Frag->Id;
        Reass->ContigList = NULL;
        Reass->GapList = NULL;
        Reass->Timer = DEFAULT_REASSEMBLY_TIMEOUT;
        Reass->Marker = 0;
        Reass->DataLength = 0;  
        Reass->UnfragmentLength = 0;
        Reass->UnfragData = NULL;
        Reass->RecvNTE = NULL;
      
        // Add new Reassembly struct to front of the ReassemblyList.
        Reass->Next = ReassemblyList;
        ReassemblyList = Reass;
    }

    //
    // Skip over fragment header.
    //
    (uchar *)Packet->Data += sizeof *Frag;
    ADJUST_PACKET_PARAMS(Packet, sizeof *Frag);

    // 
    // Send ICMP error if this fragment causes the total packet length 
    // to exceed the maximum packet size.
    // BUGBUR: Are we handling the unfragmentable length correctly here?
    // BUGBUG: What if this fragment overlaps previous fragments?
    //
    if (Reass->UnfragmentLength + Reass->Marker + Packet->TotalSize > 
                                                        MAX_IPv6_PAYLOAD) {
        SendICMPv6Error(RecvNTE, Packet,
                        ICMPv6_PARAMETER_PROBLEM, 
                        ICMPv6_ERRONEOUS_HEADER_FIELD,
                        (Packet->Position - Packet->IPPosition) - 
                        sizeof(FragmentHeader) +
                        FIELD_OFFSET(FragmentHeader, OffsetFlag));
        return 0;
    }

    //
    // If this is the last fragment (more fragments bit not set), then
    // remember the total data length, else, check the length is a multiple 
    // of 8 bytes.
    //
    if ((net_short(Frag->OffsetFlag) & FRAGMENT_FLAG_MASK) == 0) {
        Reass->DataLength = FragOffset + Packet->TotalSize;
    } else {
        if (Packet->TotalSize % 8) {
            //
            // Length is not multiple of 8, send ICMP error with a pointer 
            // value equal to offset of payload len field in IP hdr.
            //
            SendICMPv6Error(RecvNTE, Packet,
                            ICMPv6_PARAMETER_PROBLEM, 
                            ICMPv6_ERRONEOUS_HEADER_FIELD,
                            FIELD_OFFSET(IPv6Header, PayloadLength));
            return 0; // Drop packet.
        }
    }    

    //
    // We can't hold onto the incoming packet data indefinitely,
    // so copy the data contents into our own storage buffer.
    //
    // REVIEW: Combine this with shim allocation below?
    //
    Mem = ExAllocatePool(NonPagedPool, Packet->TotalSize);
    if (Mem == NULL) {
        DbgPrint("FragmentReceive: Couldn't allocate memory!?!\n");
        return 0;
    }
    RtlCopyMemory(Mem, Packet->Data, Packet->TotalSize);
    
    //
    // We're going to put it on one of the Fragment lists.
    // Allocate shim to hold link field.
    //
    Shim = ExAllocatePool(NonPagedPool, sizeof(*Shim));
    if (Shim == NULL) {
        // Out of memory!  Drop packet.
        return 0;
    }
    Shim->This = Mem;
    Shim->Len = Packet->TotalSize;
    Shim->Offset = FragOffset;
    Shim->Next = NULL;

    //
    // Determine where this fragment fits among the previous ones.
    //
    if (FragOffset == Reass->Marker) {
        //
        // This fragment extends the contiguous list.
        //
        if (Reass->ContigList == NULL) {
            //
            // We're first on the list.
            //
            ASSERT(FragOffset == 0);
            Reass->ContigList = Shim;

            //
            // We use info from the offset zero fragment to recreate
            // original datagram.
            //

            // Save the next header value.
            Reass->NextHeader = Frag->NextHeader;

            // Grab the unfragmentable data, i.e. the extension headers that 
            // preceded the fragment header. 
            //
            // BUGBUG: We're assuming that the unfragmentable headers exist
            // BUGBUG: in contiguous data after the IP header - dangerous*.
            //
            Reass->UnfragmentLength = (Packet->Position - FRAG_HDR_LEN) -
                (Packet->IPPosition + IPv6_HDR_LEN);

            if (Reass->UnfragmentLength) {
                Reass->UnfragData = 
                    ExAllocatePool(NonPagedPool, Reass->UnfragmentLength);
                RtlCopyMemory(Reass->UnfragData, 
                              (uchar*)((uchar *)(Packet->IP) + IPv6_HDR_LEN),
                              Reass->UnfragmentLength);
            }
            
            // Store the NTE pointer because it's needed if reassembly 
            // timer expires.
            Reass->RecvNTE = RecvNTE;

        } else {
            //
            // Add us to the end of the list.
            //
            Reass->ContigEnd->Next = Shim;
        }
        Reass->ContigEnd = Shim;
        
        //
        // Increment our contiguous extent marker.
        //
        Reass->Marker += Packet->TotalSize;

        //
        // Now peruse the non-contiguous list here to see if we already
        // have the next fragment to extend the contiguous list, and if so,
        // move it on over.  Repeat until we can't.
        //
        MoveShim = &Reass->GapList;
        while ((ThisShim = *MoveShim) != NULL) {
            if (ThisShim->Offset == Reass->Marker) {
                //
                // This fragment now extends the contiguous list.
                // Add it to the end of the list.
                //
                Reass->ContigEnd->Next = ThisShim;
                Reass->ContigEnd = ThisShim;
                Reass->Marker += ThisShim->Len;

                //
                // Remove it from non-contiguous list.
                //
                *MoveShim = ThisShim->Next;
                ThisShim->Next = NULL;

                //
                // Start over from beginning of the non-contigous list.
                // REVIEW: Sort the GapList to avoid this?
                //
                MoveShim = &Reass->GapList;
                continue;
            }
            MoveShim = &ThisShim->Next;
        }
        
        //
        // Check to see if contiguous list now holds entire packet.
        //
        if (Reass->Marker == Reass->DataLength) {
            ReassembleDatagram(Reass, Packet, Reass->DataLength, RecvNTE);
        }
    } else {
        //
        // Exile this fragment to the non-contiguous list.
        // REVIEW: We don't keep the ordered.  Probably should.
        //
        Shim->Next = Reass->GapList;
        Reass->GapList = Shim;
    }
    return 0;
}
    

//* FragmentLookup - look for record of previous fragments from this datagram.
//
//  A datagram is uniquely identified by its {source address, destination
//  address, identification field} triple.  This function checks our
//  reassembly list for previously received fragments of a given datagram.
//
Reassembly *
FragmentLookup(
    IPv6Addr *Src,  // Source address to match.
    IPv6Addr *Dst,  // Destination address to match.
    ulong Id)       // Fragment identification field to match.
{
    Reassembly *Reass;

    Reass = ReassemblyList;

    //
    // Just linearly scan the list for now.  If this proves to be too slow,
    // we can always hash on the Id.
    //
    while (Reass != NULL) {
        if ((Reass->Id == Id) && (IP6_ADDR_EQUAL(&Reass->IPHdr.Source, Src)) &&
            (IP6_ADDR_EQUAL(&Reass->IPHdr.Dest, Dst))) {
            break;
        }
        Reass = Reass->Next;
    }
    return Reass;
}


//* RemoveFromReassemblyList - remove reassembly record from list.
//
void
RemoveFromReassemblyList(
    Reassembly *Reass)  // Reassembly record to remove.
{
    Reassembly **PrevReass = &ReassemblyList;
    Reassembly *ThisReass; 
    PacketShim *ThisShim, *PrevShim;
    
    KIRQL OldIrql;

    // Acquire the lock to protect the Reassembly List
    KeAcquireSpinLock(&ReassemblyListLock, &OldIrql);
    
    while (ThisReass = *PrevReass) {
        if (ThisReass == Reass) {   
            // We've found the record to remove, so unlink it from this list.
            *PrevReass = Reass->Next;

            // Free ContigList if populated.
            PrevShim = ThisShim = Reass->ContigList;
            while(ThisShim != NULL) {
                PrevShim = ThisShim;
                ThisShim = ThisShim->Next;
                ExFreePool(PrevShim->This);
                ExFreePool(PrevShim);
            }

            // Free GapList if populated.
            PrevShim = ThisShim = Reass->GapList;
            while(ThisShim != NULL) {
                PrevShim = ThisShim;
                ThisShim = ThisShim->Next;
                ExFreePool(PrevShim->This);
                ExFreePool(PrevShim);
            }

            // Free unfragmentable data.
            if (Reass->UnfragData) {
                ExFreePool(Reass->UnfragData);
            }

            ExFreePool(Reass);
            break;
        }
        PrevReass = &ThisReass->Next;
    }
    KeReleaseSpinLock(&ReassemblyListLock, OldIrql);
}


//* ReassembleDatagram - put all the fragments together.
//
//  Called when we have all the fragments to complete a datagram.
//  Patch them together and pass the packet up.
//
void
ReassembleDatagram(
    Reassembly *Reass,       // Reassembly record for datagram to put together.
    IPv6Packet *Packet,      // The last received fragment.
    uint OrigLen,            // Original length of datagram.
    NetTableEntry *RecvNTE)  // NTE we received (all of?!?) the fragments on.
{
    uint TotalLength;
    int memptr = IPv6_HDR_LEN;
    PacketShim *ThisShim, *PrevShim;
    IPv6Packet OPacket;
    void *OMem;
    NDIS_STATUS Status;            
    int (*Handler)();
    int PacketRefs;


    TotalLength = OrigLen + Reass->UnfragmentLength + IPv6_HDR_LEN;

    //
    // Allocate memory for buffer and copy fragment data into it.
    //
    OMem = ExAllocatePool(NonPagedPool, TotalLength);
    if (OMem == NULL) {
        DbgPrint("FragmentReceive: Couldn't allocate memory!?!\n");
        return;
    }

    //
    // Copy the original IP hdr, unfragmentable data, and fragment data into 
    // new packet.
    //
    //
    Reass->IPHdr.PayloadLength = OrigLen + Reass->UnfragmentLength;
    Reass->IPHdr.NextHeader = Reass->NextHeader;
    RtlCopyMemory(OMem, (uchar *)&Reass->IPHdr, IPv6_HDR_LEN);
    
    if (Reass->UnfragmentLength) {
        RtlCopyMemory((uchar *)OMem + memptr, 
                      (uchar *)(Packet->IP + IPv6_HDR_LEN), 
                      Reass->UnfragmentLength);
        memptr += Reass->UnfragmentLength;
    }

    //
    // Run through the contiguous list, copying data over to our new packet.
    // REVIEW: Rather than copy, should we turn existing memory regions
    // REVIEW: into NDIS buffers?
    //
    PrevShim = ThisShim = Reass->ContigList;
    while(ThisShim != NULL) {
        RtlCopyMemory((uchar *)OMem + memptr, ThisShim->This, 
                      ThisShim->Len);
        memptr += ThisShim->Len;
        if (memptr > (int) (OrigLen + IPv6_HDR_LEN)) {
            KdPrint(("ReassembleDatagram: packets don't add up\n"));
            return;
        }
        PrevShim = ThisShim;
        ThisShim = ThisShim->Next;

        // Free shim and its buffer.
        ExFreePool(PrevShim->This);
        ExFreePool(PrevShim);
    }

    // Explicitly null out the ContigList which was free above
    Reass->ContigList = NULL;

    // Set IPv6Packet parameters.
    OPacket.Next = NULL;
    OPacket.IP = OMem;
    OPacket.Data = (uchar*)OMem + IPv6_HDR_LEN;
    OPacket.Position = IPv6_HDR_LEN;
    OPacket.ContigSize = TotalLength - IPv6_HDR_LEN;
    OPacket.TotalSize = TotalLength - IPv6_HDR_LEN;
    OPacket.NdisPacket = NULL;
    OPacket.Auxiliary = NULL;
    OPacket.Flags = 0;
    OPacket.IPPosition = 0;
    
    // Send reassembled packet to upper layers.
    Handler = ProtocolSwitchTable[Reass->NextHeader].DataReceive;
    if (Handler == NULL) {
        //
        // If we don't have a handler for this header type,
        // send an ICMP error message and drop the packet.
        // ICMP Pointer value is the offset from the start of the
        // incoming packet's IPv6 header to the offending field.
        //
        // REVIEW: The offset we use is correct for the reassembled packet.
        // REVIEW: Or should we be complaining about the original fragment?
        //
        KdPrint(("FragmentReceive: Next Header type %d isn't supported.\n",
                 Reass->NextHeader));
        SendICMPv6Error(RecvNTE, &OPacket,
                        ICMPv6_PARAMETER_PROBLEM, 
                        ICMPv6_UNRECOGNIZED_NEXT_HEADER,
                        FIELD_OFFSET(IPv6Header, NextHeader));
        return;
    }

    if ((*Handler)(RecvNTE, &OPacket) == 0) {
        ExFreePool(OMem);
        // BUGBUG: Should remove Opacket here?
    } // else
      // Some upper layer kept a reference to the packet!
      // REVIEW: O.K. to do nothing in that case?
 
    RemoveFromReassemblyList(Reass);

    return;
}


//* ReassemblyTimeout - Handle a reassembly timer event.
//
//  This routine is called periodically by IPv6Timeout to check for
//  timed out fragments.
//
void
ReassemblyTimeout(void)
{
    Reassembly **PrevReass = &ReassemblyList;
    Reassembly *ThisReass; 
    PacketShim *ThisShim, *PrevShim;

    KIRQL OldIrql;

    // Acquire the lock to protect the Reassembly List.
    KeAcquireSpinLock(&ReassemblyListLock, &OldIrql);

    //
    // Scan the ReassemblyList checking for expired reassembly contexts.
    //
    while (ThisReass = *PrevReass) {
        IPv6Packet *Packet;
        //
        // First decrement the timer then check if it has expired.  If so,
        // remove the reassembly record.  This is basically the same code 
        // as in RemoveFromReassemblyList().
        //
        ThisReass->Timer--;
        
        if (ThisReass->Timer == 0) {
            //
            // This Reassembly context has expired, send ICMP error IF we
            // have received the first fragment (ie that with a zero offset).
            //
            if (ThisReass->Marker) {
                if (Packet = CreateFragmentPacket(ThisReass)) {
                    SendICMPv6Error(ThisReass->RecvNTE, Packet,
                                    ICMPv6_TIME_EXCEEDED,
                                    ICMPv6_REASSEMBLY_TIME_EXCEEDED,
                                    0);

                }
                ExFreePool(Packet->IP);  // Start of memory region.
                ExFreePool(Packet);
            }

            KdPrint(("ReassemblyTimeout Id = %d\n", ThisReass->Id));

            // Remove reassembly entry from the list. 
            *PrevReass = ThisReass->Next;

            // Free ContigList if populated.
            PrevShim = ThisShim = ThisReass->ContigList;
            while(ThisShim != NULL) {
                PrevShim = ThisShim;
                ThisShim = ThisShim->Next;
                ExFreePool(PrevShim->This);
                ExFreePool(PrevShim);
            }

            // Free GapList if populated.
            PrevShim = ThisShim = ThisReass->GapList;
            while(ThisShim != NULL) {
                PrevShim = ThisShim;
                ThisShim = ThisShim->Next;
                ExFreePool(PrevShim->This);
                ExFreePool(PrevShim);
            }

            ExFreePool(ThisReass);

        } else
            PrevReass = &ThisReass->Next;
    }
    KeReleaseSpinLock(&ReassemblyListLock, OldIrql);
}
     

//* DestinationOptionsReceive - Handle a IPv6 Destination optiions.
// 
//  This is the routine called to process a Destination Options Header,
//  a next header value of 60.
// 
int
DestinationOptionsReceive(
    NetTableEntry *RecvNTE,     // NTE that we received the packet on.
    IPv6Packet *Packet)         // Packet handed to us by the link layer.
{
    int PacketRefs;
    uint ExtLen;
    IPv6DestOptionsHeader *DestOpt;
    Options Opts;
    ICMPv6MsgType Err;
    int (*Handler)();

    
    //
    // Verify that we have enough contiguous data to overlay a Destination
    // Options Header structure on the incoming packet.  Then do so.
    //
    if (Packet->ContigSize < EXT_HDR_LEN) {
        if (PacketPullup(Packet, EXT_HDR_LEN) == 0) {
            // Pullup failed.
            KdPrint(("DestinationOptionsReceive: Incoming packet too small"
                     " to contain destination options header\n"));
            return 0;  // Drop packet.
        }
    }
    DestOpt = (IPv6DestOptionsHeader UNALIGNED *) Packet->Data;
    
    //
    // Check that length of destination options also fit in remaining data.
    //
    ExtLen = DestOpt->HeaderExtLength * EXT_LEN_UNIT + DEFAULT_EXT_HDR_LEN;
    if (ExtLen > Packet->ContigSize) {
        if (PacketPullup(Packet, ExtLen) == 0) {
            // Pullup failed.
            KdPrint(("DestinationOptionsReceive: Incoming packet too small"
                     " to contain destination options\n"));
            return 0;  // Drop packet.
        }
    }

    //
    // Process options in this extension header.  If an error occurs 
    // while processing the options, discard packet.
    //
    if (OptionProcessing((uchar *)Packet->Data + EXT_HDR_LEN, Packet,
                         RecvNTE, ExtLen - EXT_HDR_LEN, &Opts) == 1) {
        return 0; // Discard
    }
  
    //
    // Switch to higher layer protocol.
    //
    Handler = ProtocolSwitchTable[DestOpt->NextHeader].DataReceive;
    if (Handler == NULL) {
        //
        // If we don't have a handler for this header type,
        // send an ICMP error message and drop the packet.
        // ICMP Pointer value is the offset from the start of the
        // incoming packet's IPv6 header to the offending field.
        //
        KdPrint(("DestOptionsReceive: Header type %d isn't supported.\n",
                 DestOpt->NextHeader));
        SendICMPv6Error(RecvNTE, Packet,
                        ICMPv6_PARAMETER_PROBLEM, 
                        ICMPv6_UNRECOGNIZED_NEXT_HEADER,
                        Packet->Position - Packet->IPPosition +
                        FIELD_OFFSET(IPv6DestOptionsHeader, NextHeader));
        return 0;
    }

    ADJUST_PACKET_PARAMS(Packet, ExtLen);
    (uchar *)Packet->Data += ExtLen;

    PacketRefs = (*Handler)(RecvNTE, Packet);

    return PacketRefs;
}


//* HopBPyHopOptionsReceive - Handle a IPv6 Hop-by-Hop Options.
//                                
// 
//  This is the routine called to process a Hop-by-Hop Options Header,
//  next header value of 0.
//  
int
HopByHopOptionsReceive(
    NetTableEntry *RecvNTE,     // NTE that we received the packet on.
    IPv6Packet *Packet)         // Packet handed to us by the link layer.
{
    int PacketRefs;
    uint PayloadLength;
    uint ExtLen;
    IPv6HopByHopHeader *HopByHop;
    Options Opts;
    ICMPv6MsgType Err;
    int (*Handler)();

    // Check that this is indeed the first extension header
    if (Packet->IP->NextHeader != IP_PROTOCOL_HOP_BY_HOP) {
        //
        // REVIEW: Should send an ICMP, but RFC 1883 does not specify 
        // REVIEW: what type.  We send a "PARAMETER PROBLEM" error
        // REVIEW: with the pointer offset set to beginning of this
        // REVIEW: HOP_BY_HOP header.
        //
        SendICMPv6Error(RecvNTE, Packet,
                        ICMPv6_PARAMETER_PROBLEM, 
                        ICMPv6_ERRONEOUS_HEADER_FIELD,
                        Packet->Position - Packet->IPPosition);
        return 0;
    }

    PayloadLength = net_short(Packet->IP->PayloadLength);

    //
    // Verify that we have enough contiguous data to overlay a minimum
    // length Hop-by-Hop Options Header.  Then do so.
    //
    if (Packet->ContigSize < EXT_HDR_LEN) {
        if (PacketPullup(Packet, EXT_HDR_LEN) == 0) {
            // Pullup failed.
            KdPrint(("HopByHopOptionsReceive: Incoming packet too small"
                     " to contain destination options header\n"));
            return 0;  // Drop packet.
        }
    }
    HopByHop = (IPv6HopByHopHeader UNALIGNED *) Packet->Data;
    
    // 
    // Check that length of the entire Hop-by-Hop options also fits in 
    // remaining data.
    //
    ExtLen = HopByHop->HeaderExtLength * EXT_LEN_UNIT + DEFAULT_EXT_HDR_LEN;
    if (Packet->ContigSize < ExtLen) {
        if (PacketPullup(Packet, ExtLen) == 0) {
            // Pullup failed.
            KdPrint(("HopByHopOptionsReceive: Incoming packet too small"
                     " to contain Hop-by-Hop options\n"));
            return 0;  // Drop packet.
        }
    }

    //
    // Process options in this extension header. If an error occurs 
    // while processing the options, discard packet.
    //
    if (OptionProcessing((uchar *)Packet->Data + EXT_HDR_LEN, Packet,
                         RecvNTE, ExtLen - EXT_HDR_LEN, &Opts) == 1) {
        return 0; // Discard
    }

    // 
    // If we have a valid Jumbo Payload Option, use it's value as
    // the packet PayloadLength
    //
    if (Opts.JumboLength) {
        //
        // Check that the amount of payload specified in the Jumbo
        // Payload value fits in the buffer handed to us.
        //
        if (Opts.JumboLength > Packet->TotalSize) {
            // Silently discard data
            KdPrint(("IPv6 Receive: Jumbo payload length value doesn't"
                     " match data"));
            return 0;
        }
        PayloadLength = (uint) Opts.JumboLength;
        
        //
        // As in IPv6Receive(), adjust the ContigSize to be exacltly the 
        // IP payload size.
        //
        // BUGBUG: Should we check here if Packet->ContigSize < PayloadLength ?
        //
        if (Packet->ContigSize > (PayloadLength - IPv6_HDR_LEN))
            Packet->ContigSize = PayloadLength - IPv6_HDR_LEN;

        // Set the jumbo option packet flag.
        Packet->Flags |= PACKET_JUMBO_OPTION;
    }

    //
    // Switch to higher layer protocol.
    //
    Handler = ProtocolSwitchTable[HopByHop->NextHeader].DataReceive;
    if (Handler == NULL) {
        //
        // If we don't have a handler for this header type,
        // send an ICMP error message and drop the packet.
        // ICMP Pointer value is the offset from the start of the
        // incoming packet's IPv6 header to the offending field.
        //
        KdPrint(("HopByHopOptionsReceive: Header type %d isn't supported.\n",
                 HopByHop->NextHeader));
        SendICMPv6Error(RecvNTE, Packet,
                        ICMPv6_PARAMETER_PROBLEM, 
                        ICMPv6_UNRECOGNIZED_NEXT_HEADER,
                        Packet->Position - Packet->IPPosition +
                        FIELD_OFFSET(IPv6HopByHopHeader, NextHeader));
        return 0;
    }

    ADJUST_PACKET_PARAMS(Packet, ExtLen);
    (uchar *)Packet->Data += ExtLen;

    PacketRefs = (*Handler)(RecvNTE, Packet);

    return PacketRefs;
}


//* OptionProcessing - Routine for generic header options processsing. 
//
// Return value: 1 if 'unacceptable' error occured, 0 otherwise.
int  
OptionProcessing(
    uchar *Buffer,  // Points to start of the ext hdr containing the options.
    IPv6Packet *Packet,     // The packet handed to us by the link layer.
    NetTableEntry *RecvNTE, // NTE that we received the packet on.
    uint OptsLength,        // Length of the entire options area.
    Options *Opts)          // Options struct pointer returned to caller.
{
    OptionHeader *OptHdr;
    uchar *BufPtr = Buffer;
    uint Len, Marker = 0;

    // Zero out the Options struct that is returned.
    RtlZeroMemory(Opts, sizeof(Options));

    while (Marker < OptsLength) {
        OptHdr = (OptionHeader *) BufPtr;

        switch(OptHdr->Type) {

        case OPT6_PAD_1:
            // This is a special pad option which is just a one byte field, 
            // i.e. it has no length or data field.
            BufPtr++;
            Marker++;
            break;

        case OPT6_PAD_N:
            Len = OptHdr->DataLength + OPT_HDR_LEN;
            if (Len > OptsLength) {
                // bad length, discard packet.
                return 1;
            }
            Marker += Len;
            BufPtr += Len;
            break;

        case OPT6_JUMBO_PAYLOAD:
            Opts->JumboLength = net_long((ulong) *(BufPtr + OPT_HDR_LEN));
            if (Opts->JumboLength <= MAX_IPv6_PAYLOAD) {
                //
                // Jumbo payload option was not actually needed,
                // since the size could have fit in the PayloadLength field.
                // ICMP pointer is set to offset of jumbo payload len field.
                //
                SendICMPv6Error(RecvNTE, Packet,
                                ICMPv6_PARAMETER_PROBLEM,
                                ICMPv6_ERRONEOUS_HEADER_FIELD,
                                Packet->Position - Packet->IPPosition +
                                EXT_HDR_LEN + OPT_HDR_LEN + Marker);
                return 1; // discard packet
            }
            if (net_short(Packet->IP->PayloadLength) != 0) {
                //
                // Jumbo option encoutered when IP payload is not set to zero.
                // Send ICMP error, set pointer to offset of this option type.
                //
                SendICMPv6Error(RecvNTE, Packet,
                                ICMPv6_PARAMETER_PROBLEM,
                                ICMPv6_ERRONEOUS_HEADER_FIELD,
                                Packet->Position - Packet->IPPosition +
                                Marker);
                return 1; // discard packet
            }
            // Ignore the option data length, assume it's 6 bytes incl. hdr.
            Marker += 6;
            BufPtr += 6;
            break;

        case OPT6_ROUTER_ALERT:
            // Return the pointer to the router alert struct.
            Opts->Alert = (int) BufPtr;

            // Ignore the option data length, assume it's 4 bytes incl. hdr.
            Marker += 4;
            BufPtr += 4;
            break;

        default:
            switch(OPT6_ACTION(OptHdr->Type)) {
            case OPT6_A_SKIP:
                Len = OptHdr->DataLength + OPT_HDR_LEN;
                if (Len > OptsLength) {
                    // bad length, drop packet.
                    DbgPrint("bad opt len discard\n");
                    return 1;
                }
                Marker += Len;
                BufPtr += Len;
                DbgPrint("bad opt skip\n");
                break;

            case OPT6_A_DISCARD:
                return 1; // discard packet.

            case OPT6_A_SEND_ICMP_ALL:
                goto badopt;

            case OPT6_A_SEND_ICMP_NON_MULTI:
                if (IsMulticast(&Packet->IP->Dest)) {
                    return 1;  // discard.
                } else {
                    goto badopt;
                }
            }
            break;
        }
    }
    return 0;

  badopt:
    SendICMPv6Error(RecvNTE, Packet,
                    ICMPv6_PARAMETER_PROBLEM,
                    ICMPv6_UNRECOGNIZED_OPTION,
                    Packet->Position - Packet->IPPosition +
                    EXT_HDR_LEN + Marker);
    return 1; // discard packet
}


//* ExtHdrControlReceive 
// 
//  Routine for processing the extension headers in an ICMP error message
//  before delivering the error message to the upper-layer protocol.
//
int
ExtHdrControlReceive(
    NetTableEntry *RecvNTE,     // NTE that we received the packet on.
    IPv6Packet *Packet,         // Packet handed to us by link layer.
    StatusArg *StatArg)         // ICMP Error code and offset pointer.
{
    ProtoControlRecvProc *CtlProc = NULL;

    uint NextHdr = StatArg->Other;
    int Length;
    
    // BUGBUG: need to check for contiguous data here.

    while (IS_EXTENSION_HEADER(NextHdr)) {
        switch(NextHdr)	{
        case IP_PROTOCOL_HOP_BY_HOP: {
            IPv6HopByHopHeader *ExtHdr;

            ExtHdr = (IPv6HopByHopHeader *) Packet->Data;
            Length =  (ExtHdr->HeaderExtLength + 1) * EXT_LEN_UNIT;
            NextHdr = ExtHdr->NextHeader;
            }
            break;
        case IP_PROTOCOL_DEST_OPTS: {
            IPv6DestOptionsHeader *ExtHdr;

            ExtHdr = (IPv6DestOptionsHeader *) Packet->Data;
            Length =  (ExtHdr->HeaderExtLength + 1) * EXT_LEN_UNIT;
            NextHdr = ExtHdr->NextHeader;
            }
            break;
        case IP_PROTOCOL_FRAGMENT: {
            ExtensionHeader *ExtHdr;

            ExtHdr = (ExtensionHeader *) Packet->Data;
            Length =  DEFAULT_EXT_HDR_LEN;
            NextHdr = ExtHdr->NextHeader;
            }
            break;
        case IP_PROTOCOL_ROUTING: {
            IPv6RoutingHeader *ExtHdr;

            ExtHdr = (IPv6RoutingHeader *) Packet->Data;
            Length =  (ExtHdr->HeaderExtLength + 1) * EXT_LEN_UNIT;
            NextHdr = ExtHdr->NextHeader;
            }
            break;
        default:
            return 0; // done with packet.
        }

        (uchar *)Packet->Data += Length;
        ADJUST_PACKET_PARAMS(Packet, Length);
    }
    
    //
    // Deliver ICMP Error to higher layers.
    //

    CtlProc = ProtocolSwitchTable[NextHdr].ControlReceive;
    if (CtlProc) {
        (*CtlProc)(RecvNTE, Packet, StatArg);
    }
    return 0; // done with packet, so return a zero packet reference.
}


//* CreateFragmentPacket
// 
//  Recreates the first fragment packet for purposes of notifying a source
//  of a 'fragment reassembly time exceeded'.
//
//  REVIEW: It's not clear in specs whether a full fragment should be created.
//
IPv6Packet *
CreateFragmentPacket(
    Reassembly *Reass)
{
    IPv6Packet *Packet;
    FragmentHeader FragHdr;
    int Len, memptr = IPv6_HDR_LEN;
    void *OMem;
  
    //
    // Allocate memory for creating the first fragment, i.e. the first
    // buffer in our contig list.
    //
    Len = IPv6_HDR_LEN + FRAG_HDR_LEN + Reass->UnfragmentLength + 
        Reass->ContigList->Len;
    OMem = ExAllocatePool(NonPagedPool, Len);
    if (OMem == NULL) {
        DbgPrint("CreateFragmentPacket: Couldn't allocate memory!?!\n");
        return NULL;
    }

    //
    // Copy the original IP hdr, Unfragmentable data, fragment header,
    // and data into new packet.
    // 

    RtlCopyMemory(OMem, (uchar *)&Reass->IPHdr, IPv6_HDR_LEN);
    if (Reass->UnfragmentLength) {
        RtlCopyMemory((uchar *)OMem + memptr, (uchar *)Reass->UnfragData, 
                      Reass->UnfragmentLength);
        memptr += Reass->UnfragmentLength;
    }

    FragHdr.NextHeader = Reass->NextHeader;
    FragHdr.Reserved = 0;
    FragHdr.OffsetFlag = 0;
    FragHdr.Id = Reass->Id;

    RtlCopyMemory((uchar *)OMem + memptr, (uchar *) &FragHdr, 
                  FRAG_HDR_LEN);
    
    memptr += FRAG_HDR_LEN;

    RtlCopyMemory((uchar *)OMem + memptr, Reass->ContigList->This,
                  Reass->ContigList->Len);
    
    // Set IPv6Packet parameters.
    Packet = ExAllocatePool(NonPagedPool, sizeof(IPv6Packet));
    Packet->Next = NULL;
    Packet->IP = OMem;
    Packet->IPPosition = 0;
    Packet->Data = (uchar*)OMem + IPv6_HDR_LEN;
    Packet->Position = IPv6_HDR_LEN;
    Packet->ContigSize = Len - IPv6_HDR_LEN;
    Packet->TotalSize = Len - IPv6_HDR_LEN;
    Packet->NdisPacket = NULL;
    Packet->Auxiliary = NULL;
    Packet->Flags = 0;

    return Packet;
}


//* SendICMPv6Error
//
void
SendICMPv6Error(
    NetTableEntry *RecvNTE,
    IPv6Packet *Packet,
    uint Type,
    uint Code,
    uint Pointer)
{
    ICMPv6MsgType Err;
    
    Err.Type = Type;
    Err.Code = Code;
    Err.Value = Pointer;
    ICMPv6Reply(RecvNTE, Packet, &Err);
}
