// -*- mode: C++; tab-width: 4; indent-tabs-mode: nil -*- (for GNU Emacs)
//
// Copyright (c) 1998 Microsoft Corporation
//
// This file is part of the Microsoft Research IPv6 Network Protocol Stack.
// You should have received a copy of the Microsoft End-User License Agreement
// for this software along with this release; see the file "license.txt".
// If not, please see http://www.research.microsoft.com/msripv6/license.htm,
// or write to Microsoft Research, One Microsoft Way, Redmond, WA 98052-6399.
//
// Abstract:
//
// Routing routines for Internet Protocol Version 6.
//


#include "oscfg.h"
#include "ndis.h"
#include "ip6imp.h"
#include "ip6def.h"
#include "route.h"
#include "neighbor.h"


//
// The Route Cache contains RCEs. RCEs with reference count of one
// can still be cached, but they may also be reclaimed.
// (The lone reference is from the cache itself.)
//
// An RCE records a route from an NTE (think source address)
// to a destination address. It contains a pointer to an NCE
// (next-hop neighbor) that should be used to reach the destination.
// It can also record path MTU.
//
// The current implementation is a very simple linked-list of RCEs.
//
KSPIN_LOCK RouteCacheLock;
RouteCacheEntry *RouteCache = NULL;
RouterListEntry *RouterList = NULL;
PrefixListEntry *PrefixList = NULL;

//* ValidateRCE
//
//  Check that an RCE is still valid, and if not, releases
//  the RCE and returns a reference for a new RCE.
//  In any case, returns a pointer to an RCE.
//
//  The supplied NTE indicates the source address
//  that the caller intends to use. It may be NULL,
//  to indicate that the caller will use the preferred
//  source address (RCE->NTE).
//
RouteCacheEntry *
ValidateRCE(NetTableEntry *NTE, RouteCacheEntry *RCE)
{
    if (((RCE->ValidateNTE != RCE->NTE->IF->Validate) ||
         (RCE->ValidateNCE != RCE->NCE->Validate)) &&
        (RCE->State != RCE_STATE_PERMANENT)) {
        RouteCacheEntry *NewRCE;
        IP_STATUS Status;

        //
        // Get a new RCE to replace the current RCE.
        // If this fails, then continue to use the current RCE.
        // This way our callers don't have to check for errors.
        //
        Status = RouteToDestination(&RCE->Destination, NTE, &NewRCE);
        if (Status == IP_SUCCESS) {
            ReleaseRCE(RCE);
            RCE = NewRCE;
        }
        else {
            DbgPrint("ValidateRCE(%x): RouteToDestination: %x\n", RCE, Status);
        }
    }

    return RCE;
}



//* FindMulticastCapableInterface
//
//  This function loops through the list of interfaces to find
//  one which supports multicast addresses.  The address of the
//  first acceptable interface is returned, or NULL is returned
//  if none of the interfaces are multicast capable.
//
Interface*
FindMulticastCapableInterface()
{
    Interface *IF;

    //
    // BUGBUG: The interface list should be locked.
    //
    for (IF = IFList; IF != NULL; IF = IF->Next) {
        if (IsIFMulticastCapable(IF))
            break;
    }

    return IF;
}



//* CreateOrReuseRoute
//
//  Creates a new RCE. Attempts to reuse an existing RCE.
//
//  Called with the route cache lock held.
//  Callable from a thread or DPC context.
//
//  Returns NULL if a new RCE can not be allocated.
//  The RefCnt field in the returned RCE is initialized to one.
//
//  REVIEW: Immediately reuses any RCE with one reference.
//  This will aid debugging. But it would be better to cache
//  unused RCEs for some time.
//
RouteCacheEntry *
CreateOrReuseRoute(void)
{
    RouteCacheEntry **Prev;
    RouteCacheEntry *RCE;

    //
    // First search for an unused RCE.
    //
    for (Prev = &RouteCache; (RCE = *Prev) != NULL; Prev = &RCE->Next) {

        if (RCE->RefCnt == 1) {
            //
            // We can reuse this RCE.
            //
            *Prev = RCE->Next;
            ReleaseNCE(RCE->NCE);
            ReleaseNTE(RCE->NTE);
            return RCE;
        }
    }

    //
    // Create a new RCE.
    //
    RCE = ExAllocatePool(NonPagedPool, sizeof *RCE);
    if (RCE == NULL)
        return NULL;

    RCE->RefCnt = 1;
    return RCE;
}

//* RouteToDestination - Find a route to a particular destination.
//
//  The optional NTE argument in effect specifies the source address
//  (and hence local interface) which should be used to reach the destination.
//  If this is a "weak model" host, then it can be ignored.
//
//  Note the NTE parameter is not optional when the destination is link-local,
//  because there is no other way to tell which interface/link to use.
//  BUGBUG: Strong/Weak host model issues need to be considered.
//
IP_STATUS  // Returns: whether call was sucessful and/or why not.
RouteToDestination(
    IPv6Addr *Dest,               // Destination address to route to.
    NetTableEntry *NTE,           // Net to send from (optional, may be NULL).
    RouteCacheEntry **ReturnRCE)  // Returns pointer to cached route info.
{
    RouteCacheEntry *RCE, **PrevRCE;
    RouteCacheEntry *OldRCE = NULL;
    Interface *IF;
    NeighborCacheEntry *NCE;
    KIRQL OldIrql, OldIrql2;
    IP_STATUS ReturnValue;

    KeAcquireSpinLock(&RouteCacheLock, &OldIrql);
    // KdPrint(("RTD: got lock.\n"));
    if (NTE == NULL) {
        //
        // If we're to route to a link-local address, we need to know which
        // interface the address is local to.  Get this from NTE.
        //
        if (IsLinkLocal(Dest)) {
            DbgPrint("RouteToDestination - ll dst, null NTE?\n");
            ReturnValue = IP_PARAMETER_PROBLEM;
            goto ReturnError;
        }
    }
    else {
        //
        // Disallow NTEs that are invalid (eg because of DAD).
        // Note that the NTE could be valid now and immediately
        // become invalid; not a problem.
        //
        if (!IsValidNTE(NTE)) {
            DbgPrint("RouteToDestination: invalid NTE\n");
            ReturnValue = IP_BAD_REQ;
            goto ReturnError;
        }
    }

    //
    // Check for an existing route cache entry.
    //

    PrevRCE = &RouteCache;
    while ((RCE = *PrevRCE) != NULL) {

        if (IP6_ADDR_EQUAL(Dest, &RCE->Destination) &&
            ((NTE == NULL) || (NTE->IF == RCE->NTE->IF))) {

            //
            // We must check that the RCE is still valid.
            //
            if (((RCE->ValidateNTE == RCE->NTE->IF->Validate) &&
                 (RCE->ValidateNCE == RCE->NCE->Validate)) ||
                (RCE->State == RCE_STATE_PERMANENT))
                goto ReturnRCE;

            if (RCE->ValidateNCE == RCE->NCE->Validate) {
                NetTableEntry *NewNTE;
                //
                // The NTE is the possible problem. Check if the
                // preferred source address is in fact unchanged.
                // This should be reasonably common for many NTE state
                // changes, and it will prevent needless massive
                // turnover in the route cache.
                //
                NewNTE = FindBestSourceAddress(RCE->NTE->IF, Dest);
                if (NewNTE != NULL) {
                    ReleaseNTE(NewNTE);

                    if (NewNTE == RCE->NTE) {
                        //
                        // The preferred source address is unchanged.
                        // Just update the validation counter.
                        //
                        DbgPrint("RouteToDestination - updating RCE %x\n", RCE);
                        RCE->ValidateNTE = NewNTE->IF->Validate;
                        goto ReturnRCE;
                    }
                }
            }

            //
            // REVIEW - There are more possible optimizations here,
            // like checking if the next-hop neighbor hasn't changed,
            // or updating the RCE in-place if it only has one reference.
            // My intuition right now is that they aren't worth the bother;
            // the fall-back of creating a new RCE with state from the old
            // one is good enough.
            //

            //
            // Not valid, we keep looking for a valid matching RCE.
            // We remove the invalid RCE from the route cache.
            // However we hold a reference to the old RCE: we will probably
            // create a new RCE, and we don't want to lose the interesting
            // state (like path MTU) that the old RCE contains.
            //
            DbgPrint("RouteToDestination - RCE %x is invalid\n", RCE);
            if (OldRCE == NULL) {
                *PrevRCE = RCE->Next;
                OldRCE = RCE;
                continue;
            }
        }

        //
        // Keep looking.
        //
        PrevRCE = &RCE->Next;
    }

    //
    // We shouldn't ever create new loop-back routes here.
    // They should be created via AddRouteToDestination.
    // The source address of packets seen outside the machine
    // can not be loop-back address.
    //
    if (NTE == LoopNTE) {
        DbgPrint("RouteToDestination - loop-back source?\n");
        ReturnValue = IP_PARAMETER_PROBLEM;
        goto ReturnError;
    }

    //
    // Get a new route cache entry.
    //
    RCE = CreateOrReuseRoute();
    if (RCE == NULL) {
        ReturnValue = IP_NO_RESOURCES;
        goto ReturnError;
    }

    //
    // Fill in newly created route cache entry with appropriate information.
    //
    if (IsV4Compatible(Dest)) {
        //
        // The destination address is v4-compatible.
        // Get a pseudo-NCE from the tunnel module.
        // (This can fail if tunneling is not enabled.)
        //
        NCE = TunnelGetPseudoNCE();
        if (NCE == NULL)
            goto ErrorNoRoute;

        IF = NCE->IF;

        //
        // Check that the source address (if specified) is assigned
        // to the outgoing pseudo-interface, eg, is v4-compatible.
        // This is enforcing strong host model - is this appropriate??
        //
        if ((NTE != NULL) && (NTE->IF != IF)) {
            DbgPrint("RouteToDestination: v4 dest but not source?\n");
            ReleaseNCE(NCE);
            goto ErrorNoRoute;
        }
    }
    else if (IsMulticast(Dest)) {
        //
        // This is a multicast packet.  If the NTE has been specified, use the
        // interface which owns the NTE.  Otherwise look for an interface 
        // capable of supporting multicast addresses.
        //
        if (NTE != NULL) {
            IF = NTE->IF;
            if (!IsIFMulticastCapable(IF)) {
                KdPrint(("RouteToDestination: NTE->IF not multicast capable.\n"));
                goto ErrorNoRoute;
            }
        }
        else {
            IF = FindMulticastCapableInterface();
            if (IF == NULL) {
                KdPrint(("RouteToDestination: no multicast capable IF found.\n"));
                goto ErrorNoRoute;
            }
        }

        //
        // We need an NCE to hold the multicast link-layer address. 
        //
        NCE = FindOrCreateNeighbor(IF, Dest);
        if (NCE == NULL) {
            ReturnValue = IP_NO_RESOURCES;
            goto ReturnErrorCleanup;
        }
    }    
    else if ((IF = FindOnLinkInterface(Dest, NTE)) != NULL) {
        //
        // Check that this is not a pseudo-interface.
        // For example, if we receive an Echo Request
        // with v4-compatible destination and link-local source,
        // then this would happen.
        //
        if (IF->DisableND) {
            DbgPrint("RouteToDestination: dest on-link pseudo-interface?\n");
            goto ErrorNoRoute;
        }

        // This destination address is on-link.
        // Search the interface's neighbor cache for the destination.
        //
        NCE = FindOrCreateNeighbor(IF, Dest);
        if (NCE == NULL) {
            ReturnValue = IP_NO_RESOURCES;
            goto ReturnErrorCleanup;
        }
    }
    else if ((NCE = FindDefaultRouter(NTE, Dest)) != NULL) {
        //
        // We have a router that we can use to reach the destination.
        //
        IF = NCE->IF;
    }
    else {
        //
        // We seem to have no way of reaching this destination.
        // REVIEW - RFC 1970 says to assume the destination is on-link.
        // But what interface?
        //
      ErrorNoRoute:
        ReturnValue = IP_DEST_NO_ROUTE;
        goto ReturnErrorCleanup;
    }

    //
    // Find the best source address for this destination.
    // (The NTE from our caller might not be the best.)
    //
    NTE = FindBestSourceAddress(IF, Dest);
    if (NTE == NULL) {
        //
        // We have no valid source address to use!
        //
        ReleaseNCE(NCE);
        ReturnValue = IP_DEST_NO_ROUTE;
        goto ReturnErrorCleanup;
    }

    //
    // FindOrCreateNeighbor/FindDefaultRouter gave us a reference
    // for the NCE. We donate that reference to the RCE.
    // Similarly, FindBestSourceAddress gave us a reference
    // for the NTE and we donate the reference to the RCE.
    //
    RCE->NCE = NCE;
    RCE->ValidateNCE = NCE->Validate;
    RCE->NTE = NTE;
    RCE->ValidateNTE = IF->Validate;
    RCE->PathMTU = IF->LinkMTU;
    RCE->Destination = *Dest;
    RCE->State = RCE_STATE_COMPUTED;

    //
    // Copy state from a previous RCE for this destination,
    // if we have it and the state is relevant.
    //
    if (OldRCE != NULL) {
        //
        // PathMTU is relevant if the next-hop neighbor is unchanged.
        //
        if (RCE->NCE == OldRCE->NCE)
            RCE->PathMTU;

        //
        // Other state, like ICMP rate-limiting information,
        // should also be copied.
        //
        ReleaseRCE(OldRCE);
    }
    
    //
    // Add the new route cache entry to the cache.
    //
    RCE->Next = RouteCache;
    RouteCache = RCE;

  ReturnRCE:
    AddRefRCE(RCE);
    KeReleaseSpinLock(&RouteCacheLock, OldIrql);
    // KdPrint(("RTD: release lock, RCE: %x\n", RCE));
    *ReturnRCE = RCE;
    return IP_SUCCESS;

  ReturnErrorCleanup:
    ExFreePool(RCE);
  ReturnError:
    KeReleaseSpinLock(&RouteCacheLock, OldIrql);
    if (OldRCE != NULL)
        ReleaseRCE(OldRCE);
    return ReturnValue;
}

//* AddRouteToDestination
//
//  Adds a permanent route to the destination.
//
//  The new RCE has two references - one for being alive
//  and a member of the cache, and the other for our caller.
//
RouteCacheEntry *
AddRouteToDestination(
    IPv6Addr *Dest,
    NetTableEntry *NTE,
    NeighborCacheEntry *NCE)
{
    KIRQL OldIrql;
    RouteCacheEntry *RCE;

    RCE = ExAllocatePool(NonPagedPool, sizeof *RCE);
    if (RCE == NULL)
        return NULL;

    //
    // We create the RCE with an extra initial reference,
    // for our caller.
    //
    RCE->RefCnt = 2;
    RCE->Destination = *Dest;
    RCE->NTE = NTE;
    RCE->ValidateNTE = NTE->IF->Validate;
    RCE->NCE = NCE;
    RCE->ValidateNCE = NCE->Validate;
    RCE->PathMTU = NCE->IF->LinkMTU;
    RCE->State = RCE_STATE_PERMANENT;
    KeQueryTickCount(&RCE->LastICMPInfoSend);
    KeQueryTickCount(&RCE->LastICMPErrorSend);

    KeAcquireSpinLock(&RouteCacheLock, &OldIrql);
    RCE->Next = RouteCache;
    RouteCache = RCE;
    KeReleaseSpinLock(&RouteCacheLock, OldIrql);

    //
    // Take references for the RCE to hold.
    //
    AddRefNCE(NCE);
    AddRefNTE(NTE);

    return RCE;
}

//* RemoveRouteToDestination
//
//  Removes a route from the route cache.
//
void
RemoveRouteToDestination(RouteCacheEntry *RCE)
{
    RouteCacheEntry **PrevRCE, *ThisRCE;
    KIRQL OldIrql;

    KeAcquireSpinLock(&RouteCacheLock, &OldIrql);

    PrevRCE = &RouteCache;
    while ((ThisRCE = *PrevRCE) != NULL) {

        if (ThisRCE == RCE) {
            //
            // Found it - remove from the route cache and
            // release the cache's reference for the RCE.
            //
            DbgPrint("Removing RCE %x from route cache\n", RCE);
            *PrevRCE = RCE->Next;
            ReleaseRCE(RCE);
            break;
        }

        PrevRCE = &ThisRCE->Next;
    }

    KeReleaseSpinLock(&RouteCacheLock, OldIrql);
}

//* ReleaseRCE
//
//  Releases a reference to an RCE.
//  Sometimes called with the route cache lock held.
//
void
ReleaseRCE(RouteCacheEntry *RCE)
{
    // KdPrint(("RelRCE: enter %x\n",RCE));
    if (InterlockedDecrement(&RCE->RefCnt) == 0) {
        //
        // This RCE should be deallocated.
        // It has already been removed from the cache.
        //
        ReleaseNTE(RCE->NTE);
        ReleaseNCE(RCE->NCE);
        KdPrint(("Freeing RCE: %x\n",RCE));
        ExFreePool(RCE);
    }
}



//* FindNetworkWithAddress - Locate NTE with corresponding address.
//
//  Convert a source address to an NTE by scanning
//  the list of NTEs, looking for an NTE with this address.
//  Returns NULL of no matching NTE is found.
//  If found, returns a reference for the NTE.
//
NetTableEntry *
FindNetworkWithAddress(IPv6Addr *Source)
{
    NetTableEntry *NTE;
    KIRQL OldIrql;

    KeAcquireSpinLock(&NetTableListLock, &OldIrql);

    for (NTE = NetTableList; ; NTE = NTE->NextOnNTL) {

        if (NTE == NULL)
            goto Return;

        if (IP6_ADDR_EQUAL(&NTE->Address, Source)) {
            //
            // Check that this NTE is valid.
            // (For example, an NTE still doing DAD is invalid.)
            //
            if (!IsValidNTE(NTE)) {
                DbgPrint("FindNetworkWithAddress: invalid NTE\n");
                NTE = NULL;
                goto Return;
            }
            break;
        }
    }

    AddRefNTE(NTE);
  Return:
    KeReleaseSpinLock(&NetTableListLock, OldIrql);
    return NTE;
}

//* InvalidateRouter
//
//  Called when we know a neighbor is no longer a router.
//
//  Callable from a thread or DPC context.
//  Called with no locks held.
//
void
InvalidateRouter(NeighborCacheEntry *NCE)
{
    RouterListEntry *RLE, **PrevRLE;
    KIRQL OldIrql;

    KeAcquireSpinLock(&RouteCacheLock, &OldIrql);

    DbgPrint("Invalidating router with NCE=%x\n", NCE);

    for (PrevRLE = &RouterList;
         (RLE = *PrevRLE) != NULL;
         PrevRLE = &RLE->Next) {

        if (RLE->NCE == NCE) {
            //
            // Remove the RLE from the list.
            //
            DbgPrint("Removed router RLE %x from router list\n", RLE);
            *PrevRLE = RLE->Next;

            //
            // Release the RLE.
            //
            ReleaseNCE(NCE);
            ExFreePool(RLE);
            
            break;
        }
    }

    //
    // Remove all cached routes using this neighbor as a router.
    // Note we must do this whether or not the neighbor
    // was a default router.
    //
    InvalidateRouteCacheForNCE(NCE);

    KeReleaseSpinLock(&RouteCacheLock, OldIrql);
}

//* RouterListUpdate - Update the default router list.
//
//  Called when we receive a valid Router Advertisement.
//
//  Note that the RouterLifetime may be ROUTER_LIFETIME_INFINITE,
//  whereas Neighbor Discovery does not allow an infinite value
//  for router lifetimes on the wire.
//
void
RouterListUpdate(NeighborCacheEntry *NCE, uint RouterLifetime)
{
    RouterListEntry *RLE;

    KeAcquireSpinLockAtDpcLevel(&RouteCacheLock);

    //
    // Search for an existing Router List Entry.
    //
    for (RLE = RouterList; ; RLE = RLE->Next) {

        if (RLE == NULL) {
            //
            // No existing entry for this router.
            // Create an entry if the lifetime is non-zero.
            //
            if (RouterLifetime != 0) {

                RLE = ExAllocatePool(NonPagedPool, sizeof *RLE);
                if (RLE == NULL)
                    break;

                RLE->NCE = NCE;
                AddRefNCE(NCE);
                RLE->RouterLifetime = RouterLifetime;

                //
                // Add the new entry to the router list.
                //
                RLE->Next = RouterList;
                RouterList = RLE;
            }
            break;
        }

        if (RLE->NCE == NCE) {
            //
            // Found an existing entry.
            // Reset its invalidation timer value.
            // (If the new lifetime is zero, it will timeout and
            // be removed immediately.)
            //
            RLE->RouterLifetime = RouterLifetime;
            break;
        }
    }

    KeReleaseSpinLockFromDpcLevel(&RouteCacheLock);
}

//* PrefixListUpdate - Update the list of on-link prefixes.
//
//  Called when we receive a valid Router Advertisement
//  with a Prefix Information option that has the L (on-link) bit set.
//  Our caller is responsible for checking that the prefix
//  is not the link-local prefix.
//
//  REVIEW - Should we do anything special when one prefix is
//  a prefix of another prefix for the same interface?
//  Should we check for interfaces having the same prefix?
//
void
PrefixListUpdate(
    Interface *IF,
    IPv6Addr *Prefix,
    uint PrefixLength,
    uint ValidLifetime)
{
    PrefixListEntry *PLE;

    KeAcquireSpinLockAtDpcLevel(&RouteCacheLock);

    //
    // Search for an existing Prefix List Entry.
    //
    for (PLE = PrefixList; ; PLE = PLE->Next) {

        if (PLE == NULL) {
            //
            // No existing entry for this prefix.
            // Create an entry if the lifetime is non-zero.
            //
            if (ValidLifetime != 0) {

                PLE = ExAllocatePool(NonPagedPool, sizeof *PLE);
                if (PLE == NULL)
                    break;

                PLE->IF = IF;
                PLE->Prefix = *Prefix;
                PLE->PrefixLength = PrefixLength;
                PLE->ValidLifetime = ValidLifetime;

                //
                // Add the new entry to the prefix list.
                //
                PLE->Next = PrefixList;
                PrefixList = PLE;
            }
            break;
        }

        if ((PLE->IF == IF) &&
            IP6_ADDR_EQUAL(&PLE->Prefix, Prefix) && 
            (PLE->PrefixLength == PrefixLength)) {
            //
            // Found an existing entry.
            // Reset its invalidation timer value.
            // (If the new lifetime is zero, it will timeout and
            // be removed immediately.)
            //
            PLE->ValidLifetime = ValidLifetime;
            break;
        }
    }

    KeReleaseSpinLockFromDpcLevel(&RouteCacheLock);
}

//* FindOnLinkInterface
//
//  Given a destination address, checks the list of on-link prefixes
//  to decide if this destination is on-link. If so, returns the
//  on-link interface.
//
//  If the optional NTE (source address) is supplied, then
//  this constrains the check. (Unless this is a "weak model" host.)
//
//  The RouteCacheLock must already be held.
//
Interface *
FindOnLinkInterface(
    IPv6Addr *Dest,
    NetTableEntry *NTE)
{
    PrefixListEntry *PLE;

    if (IsLinkLocal(Dest)) {
        ASSERT(NTE != NULL);
        return NTE->IF;
    }

    for (PLE = PrefixList; PLE != NULL; PLE = PLE->Next) {

        if (HasPrefix(Dest, &PLE->Prefix, PLE->PrefixLength) &&
            ((NTE == NULL) || (NTE->IF == PLE->IF)))
            return PLE->IF;
    }

    return NULL;
}

//* LongestMatchingPrefixLength
//
//  Looking only at prefixes for the given interface,
//  returns the length of the prefix that best matches
//  the supplied address.
//
//  The RouteCacheLock must already be held.
//
uint
LongestMatchingPrefixLength(
    Interface *IF,
    IPv6Addr *Addr)
{
    PrefixListEntry *PLE;
    uint Length, Longest = 0;

    for (PLE = PrefixList; PLE != NULL; PLE = PLE->Next) {

        if (PLE->IF == IF) {
            Length = CommonPrefixLength(Addr, &PLE->Prefix);
            if (Length > PLE->PrefixLength)
                Length = PLE->PrefixLength;

            if (Length > Longest)
                Longest = Length;
        }
    }

    return Longest;
}

//* FindDefaultRouter
//
//  Checks the list of routers and finds an appropriate default router.
//
//  If the optional NTE (source address) is supplied, then
//  this constrains the check. (Unless this is a "weak model" host.)
//
//  If the optional destination address is supplied,
//  then it is used to pick among otherwise equivalent routers
//  by comparing address prefixes.
//
//  The RouteCacheLock must already be held.
//
NeighborCacheEntry *
FindDefaultRouter(NetTableEntry *NTE, IPv6Addr *Dest)
{
    RouterListEntry *RLE;
    NeighborCacheEntry *NCE, *BestNCE = NULL;
    uint Length, BestLength;
    uint State, BestState;

    //
    // BUGBUG - Should probably round-robin among routers,
    // at least if there aren't any known to be reachable.
    // See section 6.3.6 of RFC  1970.
    // BUGBUG - Strong/weak host model issues need to be considered.
    //
    // In comparing routers, we treat the STALE, REACHABLE,
    // and PERMANENT states as equivalent.
    //
    // If the Neighbor Discovery states are equivalent,
    // and a destination address has been supplied,
    // then we look for longest prefix matching the destination.
    // This is a very poor substitute for real routing,
    // but it solves our current 6bone connection issues.
    //

    for (RLE = RouterList; RLE != NULL; RLE = RLE->Next) {
        NCE = RLE->NCE;

        //
        // Is this router suitable for the source NTE?
        //
        if ((NTE == NULL) || (NTE->IF == NCE->IF)) {
            //
            // Compute equivalent ND state for this router.
            //
            State = NCE->NDState;
            if (State > ND_STATE_STALE)
                State = ND_STATE_STALE;

            //
            // Compute matching prefix length for this router.
            // If this a router for the tunnel pseudo-interface,
            // then the router has a meaningful v6 address.
            // Otherwise the router's address is link-local and
            // doesn't tell us anything; we have to look at the
            // prefix list to get the information we want.
            //
            if (Dest != NULL)
                if (NCE->IF->DisableND)
                    Length = CommonPrefixLength(Dest, &NCE->NeighborAddress);
                else
                    Length = LongestMatchingPrefixLength(NCE->IF, Dest);
            else
                Length = 0;

            if (BestNCE == NULL) {
                //
                // This is the first suitable router,
                // so remember it.
                //
                AddRefNCE(NCE);
              RememberIt:
                BestNCE = NCE;
                BestState = State;
                BestLength = Length;
            }
            else if ((State > BestState) ||
                     ((State == BestState) &&
                      (Length > BestLength))) {
                //
                // This router looks better.
                //
                AddRefNCE(NCE);
                ReleaseNCE(BestNCE);
                goto RememberIt;
            }
        }
    }

    return BestNCE;
}

//* FindBestSourceAddress
//
//  Given an outgoing interface and a destination address,
//  finds the best source address (NTE) to use.
//
//  May be called with the route cache lock held.
//
//  If found, returns a reference for the NTE.
//
NetTableEntry *
FindBestSourceAddress(
    Interface *IF,
    IPv6Addr *Dest)
{
    NetTableEntry *BestNTE = NULL;
    uint Length, BestLength;
    NetTableEntry *NTE;
    KIRQL OldIrql;

    KeAcquireSpinLock(&IF->Lock, &OldIrql);

    for (NTE = IF->NTE; NTE != NULL; NTE = NTE->NextOnIF) {
        //
        // We look for the source address which best matches
        // the destination address. If possible, we should use
        // a source address with the appropriate scope for
        // the destination. For example, if the destination address
        // is global, the source address should be global.
        // A link-local source address will not work, unless
        // the destination happens to be on-link.
        // Site-local addresses have the same concerns.
        //
        // If we have a choice of addresses that match equally,
        // then we take a preferred address over a deprecated address.
        // We will use a deprecated address if it is the best match.
        //
        // REVIEW - Possibly should be looking at the addresses' scope
        // instead of the prefix match. Should we prefer one global
        // source address over another just because the common prefix
        // is slightly longer?
        //
        if (IsValidNTE(NTE)) { // PREFERRED or DEPRECATED
            Length = CommonPrefixLength(Dest, &NTE->Address);
            if (BestNTE == NULL) {
                AddRefNTE(NTE);
                BestNTE = NTE;
                BestLength = Length;
            }
            else if ((Length > BestLength) ||
                     ((Length == BestLength) &&
                      (BestNTE->DADState == DAD_STATE_DEPRECATED))) {
                AddRefNTE(NTE);
                ReleaseNTE(BestNTE);
                BestNTE = NTE;
                BestLength = Length;
            }
        }
    }

    KeReleaseSpinLock(&IF->Lock, OldIrql);

    return BestNTE;
}

//* RouterListTimeout
//
//  Called periodically from IPv6Timeout.
//  Handles lifetime expiration of default routers.
//
void
RouterListTimeout(void)
{
    RouterListEntry *RLE, **PrevRLE;

    KeAcquireSpinLockAtDpcLevel(&RouteCacheLock);

    PrevRLE = &RouterList;
    while ((RLE = *PrevRLE) != NULL) {

        if (RLE->RouterLifetime == 0) {
            DbgPrint("RouterListTimeout: RLE %x using NCE %x timed out\n",
                     RLE, RLE->NCE);

            //
            // Remove the RLE from the list.
            //
            *PrevRLE = RLE->Next;

            //
            // Invalidate all cached routes using this router.
            //
            InvalidateRouteCacheForNCE(RLE->NCE);

            //
            // Release the RLE.
            //
            ReleaseNCE(RLE->NCE);
            ExFreePool(RLE);
        } else {
            if (RLE->RouterLifetime != ROUTER_LIFETIME_INFINITE)
                RLE->RouterLifetime--;

            PrevRLE = &RLE->Next;
        }
    }

    KeReleaseSpinLockFromDpcLevel(&RouteCacheLock);
}

//* PrefixListTimeout
//
//  Called periodically from IPv6Timeout.
//  Handles lifetime expiration of on-link prefixes.
//
//  Note that when prefixes expire, we do not remove
//  entries from the route cache - the normal Neighbor Discovery
//  mechanisms suffice to remove them if they are invalid.
//  See RFC 1970, section 6.3.5.
//
void
PrefixListTimeout(void)
{
    PrefixListEntry *PLE, **PrevPLE;

    KeAcquireSpinLockAtDpcLevel(&RouteCacheLock);

    PrevPLE = &PrefixList;
    while ((PLE = *PrevPLE) != NULL) {

        if (PLE->ValidLifetime == 0) {
            //
            // Remove the PLE from the list.
            //
            *PrevPLE = PLE->Next;

            //
            // Release the PLE.
            //
            ExFreePool(PLE);
        }
        else {
            if (PLE->ValidLifetime != DAD_LIFETIME_INFINITE)
                PLE->ValidLifetime--;

            PrevPLE = &PLE->Next;
        }
    }

    KeReleaseSpinLockFromDpcLevel(&RouteCacheLock);
}

//* UpdatePathMTU
//
//  Update the route cache with a new MTU obtained
//  from a Packet Too Big message.
//
//  Callable from DPC context, not from thread context.
//  Called with no locks held.
//
void
UpdatePathMTU(
    IPv6Addr *Dest,
    NetTableEntry *NTE,
    uint MTU)
{
    Interface *IF = NTE->IF;
    RouteCacheEntry *RCE;

    KeAcquireSpinLockAtDpcLevel(&RouteCacheLock);

    //
    // Search the route cache for the appropriate RCE.
    // There will be at most one.
    //

    for (RCE = RouteCache; RCE != NULL; RCE = RCE->Next) {

        if (IP6_ADDR_EQUAL(&RCE->Destination, Dest) &&
            (RCE->NTE->IF == IF)) {

            //
            // As a sanity check, exclude loop-back RCEs.
            //
            if (RCE->NCE == LoopNCE)
                break;

            //
            // Update the path MTU.
            //
            if (MTU < RCE->PathMTU) {
                KdPrint(("UpdatePathMTU(RCE %x): new MTU %d for %ws\n",
                         RCE, MTU, FormatV6Address(Dest)));
                RCE->PathMTU = MTU;
            }
            break;
        }
    }

    KeReleaseSpinLockFromDpcLevel(&RouteCacheLock);
}

//* RedirectRouteCache
//
//  Update the route cache to reflect a Redirect message.
//
//  Callable from DPC context, not from thread context.
//  Called with no locks held.
//
void
RedirectRouteCache(
    IPv6Addr *Source,           // Source of the redirect.
    IPv6Addr *Dest,
    NetTableEntry *NTE,
    NeighborCacheEntry *NCE)
{
    Interface *IF = NTE->IF;
    RouteCacheEntry *RCE, **PrevRCE;
    RouteCacheEntry *OldRCE = NULL;
#if DBG
    WCHAR Buffer1[40], Buffer2[40];
#endif

    //
    // Our caller guarantees this.
    //
    ASSERT(IF == NCE->IF);

    KeAcquireSpinLockAtDpcLevel(&RouteCacheLock);

    //
    // Search the route cache for the appropriate RCE.
    // There will be at most one.
    //

    for (PrevRCE = &RouteCache;
         (RCE = *PrevRCE) != NULL;
         PrevRCE = &RCE->Next) {

        if (IP6_ADDR_EQUAL(&RCE->Destination, Dest) &&
            (RCE->NTE->IF == IF)) {

            //
            // We must check that the source of the redirect
            // is the current next-hop neighbor.
            // (This prevents other nodes on a link from hijacking.)
            //
            if (!IP6_ADDR_EQUAL(&RCE->NCE->NeighborAddress, Source)) {
                DbgPrint("RedirectRouteCache - hijack attempt\n");
                break;
            }

            //
            // No need to check for a null redirect (NCE == RCE->NCE).
            // Shouldn't happen and if it does nothing bad happens.
            // As an optimization, we could check if the RCE
            // has only one reference and if so, update it in place.
            // However I suspect that it isn't worth the code.
            //

            //
            // Cause anyone who is using/caching this RCE to notice
            // that it is no longer valid, so they will find
            // the new RCE that we create below.
            //
            RCE->ValidateNCE--;

            //
            // Remove the RCE from the cache (CreateOrReuseRoute below
            // will not find it) but hold onto it for the moment.
            //
            *PrevRCE = RCE->Next;
            OldRCE = RCE;
            break;
        }
    }

#if DBG
    FormatV6AddressWorker(Buffer1, Dest);
    FormatV6AddressWorker(Buffer2, &NCE->NeighborAddress);
    DbgPrint("RedirectRouteCache - redirect %ws to %ws\n", Buffer1, Buffer2);
#endif

    //
    // Create a route cache entry for the redirect.
    // (If we get any errors, we just ignore the redirect,
    // except for removing the old RCE.)
    //
    RCE = CreateOrReuseRoute();
    if (RCE == NULL)
        goto Return;

    //
    // Find the best source address for this destination.
    // (The NTE from our caller might not be the best.)
    //
    NTE = FindBestSourceAddress(IF, Dest);
    if (NTE == NULL) {
        //
        // We have no valid source address to use!
        //
        ExFreePool(RCE);
        goto Return;
    }

    AddRefNCE(NCE);
    RCE->NCE = NCE;
    RCE->ValidateNCE = NCE->Validate;
    RCE->NTE = NTE;     // Donate ref from FindBestSourceAddress.
    RCE->ValidateNTE = IF->Validate;
    RCE->PathMTU = IF->LinkMTU;
    RCE->Destination = *Dest;
    RCE->State = RCE_STATE_REDIRECT;

    //
    // Copy state from a previous RCE for this destination,
    // if we have it and the state is relevant.
    //
    if (OldRCE != NULL) {
        //
        // PathMTU state is not relevant, since the new RCE
        // presumably specifies a new route.
        // Other state, like ICMP rate-limiting information,
        // should be copied.
        //
    }
    
    //
    // Add the new route cache entry to the cache.
    //
    RCE->Next = RouteCache;
    RouteCache = RCE;

  Return:
    KeReleaseSpinLockFromDpcLevel(&RouteCacheLock);
    if (OldRCE != NULL)
        ReleaseRCE(OldRCE);     // The cache's reference.
}

//* InitRouting - Initialize the routing module.
//
void
InitRouting(void)
{
    KeInitializeSpinLock(&RouteCacheLock);
}
