// -*- mode: C++; tab-width: 4; indent-tabs-mode: nil -*- (for GNU Emacs)
//
// Copyright (c) 1998 Microsoft Corporation
//
// This file is part of the Microsoft Research IPv6 Network Protocol Stack.
// You should have received a copy of the Microsoft End-User License Agreement
// for this software along with this release; see the file "license.txt".
// If not, please see http://www.research.microsoft.com/msripv6/license.htm,
// or write to Microsoft Research, One Microsoft Way, Redmond, WA 98052-6399.
//
// Abstract:
//
// Routing code external definitions for Internet Protocol Version 6.
//


#ifndef ROUTE_INCLUDED
#define ROUTE_INCLUDED 1

extern void InitRouting(void);

//
// Structure of a route cache entry.
//
// A route cache entry (RCE) primarily caches two computations:
// next-hop determination and source address selection.
// An RCE also caches other information related to the destination,
// like path MTU.
//
// An RCE can also be created as a result of receiving an Redirect
// ICMP message.
//
// There is at most one RCE per destination address / interface pair.
// Our route cache corresponds to the destination cache
// mentioned in RFC 1970's conceptual data structures,
// with the addition of support for multi-homed nodes.
//
// The primary lookup key for RCEs is the destination address.
// The current implementation just searches a list of all RCEs,
// but a hash table or tree data structure would be preferable.
//
// Some nodes (like busy servers) might have many thousands of RCEs
// but only tens of NCEs, because most destinations are reached
// through only a few neighbor routers. Some nodes (like busy routers)
// will have relatively few RCEs and hundreds of NCEs, because
// forwarding does not use an RCE.
//
// The three major components of an RCE are the destination address,
// NTE (indicates both the interface, and the best source address
// on that interface to use for this destination), and NCE
// (neighbor to which to send packets for this destination).
//
// Once an RCE is created, these three components are read-only
// and anyone who holds a reference for the RCE can rely on
// them not changing. The RCE holds references for the NTE and NCE.
// This allows code that holds an RCE to access the important
// fields without acquiring any locks. Fields like the path MTU
// can also be safely read without a lock.
//
// When an RCE becomes invalid, it is removed from the route cache
// but it is not deallocated until it has zero references.
// The route cache itself holds one reference on RCEs in the cache.
//
// Because an RCE caches the result of two computations, RCEs can
// become invalid (stale) for two reasons: the preferred source
// address should be recomputed, or the next-hop neighbor should be
// recomputed.
//
// Source addresses need to be recomputed or checked when the NTEs
// on the RCE's interface change state - for example a new address
// is created, a preferred address becomes deprecated, etc.
// In practice, these should be relatively infrequent situations.
//
// Next-hop determination needs to be redone in several situations:
// a neighbor is not reachable, a neighbor stops being a router,
// a neighbor is removed from the default router list.
// Again, these should be relatively infrequent situations.
//
// To avoid undue time & memory overheads (for example maintaining
// a linked list of all RCEs that point to an NCE and a linked list
// of all RCEs on a given interface, so that the right RCEs can
// be immediately found when something changes), we use a "lazy" approach
// based on validation counters.
//
// Each interface has an validation counter that is incremented
// when a relevant NTE state change occurs. Each RCE has a snapshot
// of that counter value. If the RCE's snapshot differs from the current
// value, then source address selection must be redone.
//
// Similarly, each NCE has an validation counter and each RCE has
// a snapshot of its NCE's counter value. The NCE's counter is incremented
// when the NCE changes state in a way that affects routing.
// If the RCE's snapshot differs from the current NCE counter value,
// then next-hop determination must be redone.
//
// For efficiency, some code may cache an RCE reference for a "long"
// time, for example in a connection control block. Before using
// the cached RCE, such code should check the two counter values
// to ensure that the RCE is still valid. The ValidateRCE function
// performs this check.
//
// If a cached RCE is invalid, then it must be released and a new RCE
// acquired. Because the important fields in an RCE are read-only,
// an RCE can only be updated in-place if it has no external references.
//
struct RouteCacheEntry {
    struct RouteCacheEntry *Next;    // Next RCE in cache list.
    long RefCnt;
    IPv6Addr Destination;
    long ValidateNTE;                // Snapshot of NTE validation counter.
    struct NetTableEntry *NTE;       // Preferred source address/interface.
    long ValidateNCE;                // Snapshot of NCE validation counter.
    struct NeighborCacheEntry *NCE;  // First-hop neighbor.
    uint PathMTU;                    // MTU of path to destination.
    LARGE_INTEGER LastICMPErrorSend; // Time when last ICMP error msg was sent.
    LARGE_INTEGER LastICMPInfoSend;  // Time when last ICMP info  msg was sent.
    uint State;
};

#define RCE_STATE_PERMANENT     0
#define RCE_STATE_COMPUTED      1
#define RCE_STATE_REDIRECT      2

__inline void
AddRefRCE(RouteCacheEntry *RCE)
{
    InterlockedIncrement(&RCE->RefCnt);
}

extern void
ReleaseRCE(RouteCacheEntry *RCE);

extern RouteCacheEntry *
ValidateRCE(NetTableEntry *NTE, RouteCacheEntry *RCE);

extern IP_STATUS RouteToDestination(IPv6Addr *Destination, NetTableEntry *NTE,
                                    RouteCacheEntry **RCE);
extern NetTableEntry *FindNetworkWithAddress(IPv6Addr *Source);

extern RouteCacheEntry *
AddRouteToDestination(
    IPv6Addr *Dest,
    NetTableEntry *NTE,
    NeighborCacheEntry *NCE);

extern void
RemoveRouteToDestination(RouteCacheEntry *RCE);

extern void
RouterListUpdate(NeighborCacheEntry *NCE, uint RouterLifetime);

extern void
PrefixListUpdate(
    Interface *IF,
    IPv6Addr *Prefix,
    uint PrefixLength,
    uint ValidLifetime);

//
// Structure of an entry in the list of default routers.
//
typedef struct RouterListEntry {
    struct RouterListEntry *Next;
    NeighborCacheEntry *NCE;
    uint RouterLifetime;                // In ticks
} RouterListEntry;

//
// Structure of an entry in the list of on-link prefixes.
//
typedef struct PrefixListEntry {
    struct PrefixListEntry *Next;
    Interface *IF;
    IPv6Addr Prefix;
    uint PrefixLength;
    uint ValidLifetime;                 // In ticks
} PrefixListEntry;

extern Interface *
FindOnLinkInterface(
    IPv6Addr *Dest,
    NetTableEntry *NTE);

extern NeighborCacheEntry *
FindDefaultRouter(NetTableEntry *NTE, IPv6Addr *Dest);

extern NetTableEntry *
FindBestSourceAddress(
    Interface *IF,
    IPv6Addr *Dest);

extern void
RouterListTimeout(void);

extern void
PrefixListTimeout(void);

extern void
InvalidateRouter(NeighborCacheEntry *NCE);

extern void
UpdatePathMTU(
    IPv6Addr *Dest,
    NetTableEntry *NTE,
    uint MTU);

extern void
RedirectRouteCache(
    IPv6Addr *Source,
    IPv6Addr *Dest,
    NetTableEntry *NTE,
    NeighborCacheEntry *NCE);

extern KSPIN_LOCK RouteCacheLock;
extern RouteCacheEntry *RouteCache;
extern RouterListEntry *RouterList;
extern PrefixListEntry *PrefixList;

#endif  // ROUTE_INCLUDED
