// -*- mode: C++; tab-width: 4; indent-tabs-mode: nil -*- (for GNU Emacs)
//
// Copyright (c) 1985-1998 Microsoft Corporation
//
// This file is part of the Microsoft Research IPv6 Network Protocol Stack.
// You should have received a copy of the Microsoft End-User License Agreement
// for this software along with this release; see the file "license.txt".
// If not, please see http://www.research.microsoft.com/msripv6/license.htm,
// or write to Microsoft Research, One Microsoft Way, Redmond, WA 98052-6399.
//
// Abstract:
//
// Common code for dealing with datagrams.  This code is common to
// both UDP and Raw IP.
//


#include "oscfg.h"
#include "ndis.h"
#include "ip6imp.h"
#include "ip6def.h"
#include "tdi.h"
#include "tdistat.h"
#include "tdint.h"
#include "queue.h"
#include "transprt.h"
#include "addr.h"
#include "datagram.h"

#define NO_TCP_DEFS 1
#include "tcpdeb.h"

//
// REVIEW: shouldn't this be in an include file someplace?  Isn't it already?
//
#ifdef POOL_TAGGING

#ifdef ExAllocatePool
#undef ExAllocatePool
#endif

#define ExAllocatePool(type, size) ExAllocatePoolWithTag(type, size, 'dPCT')

#endif // POOL_TAGGING


#define NUM_DG_HEADERS 5
#define DG_MAX_HDRS 0xffff

ulong DGCurrentSendFree = 0;
ulong DGMaxSendFree = DG_MAX_HDRS;

extern KSPIN_LOCK AddrObjTableLock;

DGSendReq *DGSendReqFree;
KSPIN_LOCK DGSendReqLock;

SLIST_HEADER DGRcvReqFree;

KSPIN_LOCK DGRcvReqFreeLock;

#if DBG
uint NumSendReq = 0;
uint NumRcvReq = 0;
#endif

//
// Information for maintaining the DG Header structures and pending queue.
//
uint DGHeaderSize;
PNDIS_BUFFER DGHeaderList;
Queue DGHeaderPending;

Queue DGDelayed;
WORK_QUEUE_ITEM DGDelayedWorkItem;

typedef struct DGHdrBPoolEntry {
    struct DGHdrBPoolEntry *uhe_next;
    NDIS_HANDLE uhe_handle;
    uchar *uhe_buffer;
} DGHdrBPoolEntry;

DGHdrBPoolEntry *DGHdrBPoolList = NULL;

//
// All of the init code can be discarded.
//
#ifdef ALLOC_PRAGMA

int InitDG(uint MaxHeaderSize);

#pragma alloc_text(INIT, InitDG)

#endif // ALLOC_PRAGMA


//* GrowDGHeaderList - Try to grow the DG header list.
//
//  Called when we run out of buffers on the DG header list, and need
//  to grow it.  We look to see if we're already at the maximum size, and
//  if not we'll allocate the need structures and free them to the list.
//  This routine must be called with the SendReq lock held.
//
PNDIS_BUFFER  // Returns: A new DG header buffer if we have one, or NULL.
GrowDGHeaderList(void)
{
    DGHdrBPoolEntry *NewEntry;
    NDIS_STATUS Status;
    uint HeaderSize;
    uchar *DGSendHP;
    uint i;
    PNDIS_BUFFER Buffer;
    PNDIS_BUFFER ReturnBuffer = NULL;

    if (DGCurrentSendFree < DGMaxSendFree) {

        // Still room to grow the list.
        NewEntry = ExAllocatePool(NonPagedPool, sizeof(DGHdrBPoolEntry));

        if (NewEntry == NULL) {
            // Couldn't get the memory.
            return NULL;
        }

        NdisAllocateBufferPool(&Status, &NewEntry->uhe_handle, NUM_DG_HEADERS);

        if (Status != NDIS_STATUS_SUCCESS) {
            // Couldn't get a new set of buffers. Fail.
            ExFreePool(NewEntry);
            return NULL;
        }

        HeaderSize = DGHeaderSize + MaxLinkLevelHeader;

        DGSendHP = ExAllocatePool(NonPagedPool, HeaderSize * NUM_DG_HEADERS);

        if (DGSendHP == NULL) {
            NdisFreeBufferPool(NewEntry->uhe_handle);
            ExFreePool(NewEntry);
            return NULL;
        }

        NewEntry->uhe_buffer = DGSendHP;

        for (i = 0; i < NUM_DG_HEADERS; i++) {
            NdisAllocateBuffer(&Status, &Buffer, NewEntry->uhe_handle,
                               DGSendHP + (i * HeaderSize), HeaderSize);
            if (Status != NDIS_STATUS_SUCCESS) {
                NdisFreeBufferPool(NewEntry->uhe_handle);
                ExFreePool(NewEntry);
                ExFreePool(DGSendHP);
                return NULL;
            }
            if (i != 0)
                FreeDGHeader(Buffer);
            else
                ReturnBuffer = Buffer;
        }

        DGCurrentSendFree += NUM_DG_HEADERS;
        NewEntry->uhe_next = DGHdrBPoolList;
        DGHdrBPoolList = NewEntry;

    } else {
        // At the limit already.
        ReturnBuffer = NULL;
    }

    return ReturnBuffer;
}


//* GetDGHeader - Get a DG header buffer.
//
//  The get header buffer routine.  Called with the SendReqLock held.
//
_inline PNDIS_BUFFER  // Returns: a header buffer, or NULL if out of space.
GetDGHeader(void)
{
    PNDIS_BUFFER NewBuffer;

    NewBuffer = DGHeaderList;
    if (NewBuffer != NULL)
        DGHeaderList = NDIS_BUFFER_LINKAGE(NewBuffer);
    else
        NewBuffer = GrowDGHeaderList();

    return NewBuffer;
}

//* FreeDGHeader - Free a DG header buffer.
//
//  The free header buffer routine.  Called with the SendReqLock held.
//
void                           // Returns: Nothing.
FreeDGHeader(
    PNDIS_BUFFER FreedBuffer)  // Buffer to be freed.
{
    NDIS_BUFFER_LINKAGE(FreedBuffer) = DGHeaderList;
    DGHeaderList = FreedBuffer;
}

//* PutPendingQ - Put an address object on the pending queue.
//
//  Called when we've experienced a header buffer out of resources condition,
//  and want to queue an AddrObj for later processing.  We put the specified
//  address object on the DGHeaderPending queue, set the OOR flag and clear
//  the 'send request' flag.  It is invariant in the system that the send
//  request flag and the OOR flag are not set at the same time.
//
//  This routine assumes that the caller holds the DGSendReqLock and the
//  lock on the particular AddrObj.
//
void                      // Returns: Nothing.
PutPendingQ(
    AddrObj *QueueingAO)  // Address object to be queued.
{
    CHECK_STRUCT(QueueingAO, ao);

    if (!AO_OOR(QueueingAO)) {
        CLEAR_AO_REQUEST(QueueingAO, AO_SEND);
        SET_AO_OOR(QueueingAO);

        ENQUEUE(&DGHeaderPending, &QueueingAO->ao_pendq);
    }
}

//* GetDGSendReq - Get a DG send request.
//
//  Called when someone wants to allocate a DG send request.  We assume
//  the send request lock is held when we are called.
//
//  Note: This routine and the corresponding free routine might
//      be good candidates for inlining.
//
DGSendReq *  // Returns: Pointer to the SendReq, or NULL if none.
GetDGSendReq()
{
    DGSendReq *NewReq;

    NewReq = DGSendReqFree;
    if (NewReq != NULL) {
        CHECK_STRUCT(NewReq, dsr);
        DGSendReqFree = (DGSendReq *)NewReq->dsr_q.q_next;
    } else {
        // Couldn't get a request, grow it.  This is one area where we'll try
        // to allocate memory with a lock held.  Because of this, we've
        // got to be careful about where we call this routine from.

        NewReq = ExAllocatePool(NonPagedPool, sizeof(DGSendReq));
        if (NewReq != NULL) {
#if DBG
            NewReq->dsr_sig = dsr_signature;
            NumSendReq++;
#endif
        }
    }

    return NewReq;
}

//* FreeDGSendReq - Free a DG send request.
//
//  Called when someone wants to free a DG send request.  It's assumed
//  that the caller holds the SendRequest lock.
//
void                     // Returns: Nothing.
FreeDGSendReq(
    DGSendReq *SendReq)  // SendReq to be freed.
{
    CHECK_STRUCT(SendReq, dsr);

    *(DGSendReq **)&SendReq->dsr_q.q_next = DGSendReqFree;
    DGSendReqFree = SendReq;
}

//* GetDGRcvReq - Get a DG receive request.
//
//  Called when we need to get a DG receive request.
//
DGRcvReq *  // Returns: Pointer to new request, or NULL if none.
GetDGRcvReq()
{
    DGRcvReq *NewReq;
    PSINGLE_LIST_ENTRY BufferLink;
    Queue *QueuePtr;

    BufferLink = ExInterlockedPopEntrySList(&DGRcvReqFree, &DGRcvReqFreeLock);

    if (BufferLink != NULL) {
        QueuePtr = CONTAINING_RECORD(BufferLink, Queue, q_next);
        NewReq = CONTAINING_RECORD(QueuePtr, DGRcvReq, drr_q);
        CHECK_STRUCT(NewReq, drr);
    } else {
        // Couldn't get a request, grow it.
        NewReq = ExAllocatePool(NonPagedPool, sizeof(DGRcvReq));
        if (NewReq != NULL) {
#if DBG
            NewReq->drr_sig = drr_signature;
            ExInterlockedAddUlong(&NumRcvReq, 1, &DGRcvReqFreeLock);
#endif
        }
    }

    return NewReq;
}

//* FreeDGRcvReq - Free a DG rcv request.
//
//  Called when someone wants to free a DG rcv request.
//
void                   // Returns: Nothing.
FreeDGRcvReq(
    DGRcvReq *RcvReq)  // RcvReq to be freed.
{
    PSINGLE_LIST_ENTRY BufferLink;

    CHECK_STRUCT(RcvReq, drr);

    BufferLink = CONTAINING_RECORD(&(RcvReq->drr_q.q_next), SINGLE_LIST_ENTRY,
                                   Next);
    ExInterlockedPushEntrySList(&DGRcvReqFree, BufferLink, &DGRcvReqFreeLock);
}


//* DGDelayedWorker - Routine to handle a delayed work item.
//
//  This is the work item callback routine, used for out-of-resources
//  conditions on AddrObjs.  We pull from the delayed queue, and if the addr
//  obj is not already busy we'll send the datagram.
//
void                  // Returns: Nothing.
DGDelayedWorker(
    void *Context)    // Nothing.
{
    KIRQL Irql0, Irql1;  // One per lock nesting level.
    AddrObj *SendingAO;
    DGSendProc SendProc;

    KeAcquireSpinLock(&DGSendReqLock, &Irql0);
    while (!EMPTYQ(&DGDelayed)) {
        DEQUEUE(&DGDelayed, SendingAO, AddrObj, ao_pendq);
        CHECK_STRUCT(SendingAO, ao);

        KeAcquireSpinLock(&SendingAO->ao_lock, &Irql1);

        CLEAR_AO_OOR(SendingAO);
        if (!AO_BUSY(SendingAO)) {
            DGSendReq *SendReq;

            if (!EMPTYQ(&SendingAO->ao_sendq)) {
                DEQUEUE(&SendingAO->ao_sendq, SendReq, DGSendReq, dsr_q);

                CHECK_STRUCT(SendReq, dsr);
                ASSERT(SendReq->dsr_header != NULL);

                SendingAO->ao_usecnt++;
                SendProc = SendingAO->ao_dgsend;
                KeReleaseSpinLock(&SendingAO->ao_lock, Irql1);
                KeReleaseSpinLock(&DGSendReqLock, Irql0);

                (*SendProc)(SendingAO, SendReq);
                DEREF_AO(SendingAO);
                KeAcquireSpinLock(&DGSendReqLock, &Irql0);
            } else {
                ASSERT(FALSE);
                KeReleaseSpinLock(&SendingAO->ao_lock, Irql1);
            }

        } else {
            SET_AO_REQUEST(SendingAO, AO_SEND);
            KeReleaseSpinLock(&SendingAO->ao_lock, Irql1);
        }
    }

    KeReleaseSpinLock(&DGSendReqLock, Irql0);
}

//* DGSendComplete - DG send complete handler.
//
//  This is the routine called by IP when a send completes.  We
//  take the context passed back as a pointer to a SendRequest
//  structure, and complete the caller's send.
//
void                      // Returns: Nothing.
DGSendComplete(
    PNDIS_PACKET Packet)  // Packet that was sent.
{
    DGSendReq *FinishedSR;
    KIRQL Irql0, Irql1;  // One per lock nesting level.
    RequestCompleteRoutine Callback;  // Request's Completion routine.
    PVOID CallbackContext;  // User context.
    ushort SentSize;
    AddrObj *AO;
    PNDIS_BUFFER BufferChain;  // Chain of buffers sent.
    PVOID Unused1;
    UINT Unused2, Unused3;

    //
    // Pull values we care about out of the packet structure.
    //
    FinishedSR = (DGSendReq *) PC(Packet)->CompletionData;
    NdisGetFirstBufferFromPacket(Packet, &BufferChain, &Unused1, &Unused2,
                                 &Unused3);

    CHECK_STRUCT(FinishedSR, dsr);
    KeAcquireSpinLock(&DGSendReqLock, &Irql0);

    Callback = FinishedSR->dsr_rtn;
    CallbackContext = FinishedSR->dsr_context;
    SentSize = FinishedSR->dsr_size;

    // If there's nothing on the header pending queue, just free the
    // header buffer.  Otherwise pull from the pending queue, give him the
    // resource, and schedule an event to deal with him.
    if (EMPTYQ(&DGHeaderPending)) {
        FreeDGHeader(BufferChain);
    } else {
        DEQUEUE(&DGHeaderPending, AO, AddrObj, ao_pendq);
        CHECK_STRUCT(AO, ao);
        KeAcquireSpinLock(&AO->ao_lock, &Irql1);
        if (!EMPTYQ(&AO->ao_sendq)) {
            DGSendReq *SendReq;

            PEEKQ(&AO->ao_sendq, SendReq, DGSendReq, dsr_q);
            SendReq->dsr_header = BufferChain;  // Give him this buffer.

            ENQUEUE(&DGDelayed, &AO->ao_pendq);
            KeReleaseSpinLock(&AO->ao_lock, Irql1);
            ExQueueWorkItem(&DGDelayedWorkItem, CriticalWorkQueue);
        } else {
            // On the pending queue, but no sends!
            KdBreakPoint();
            CLEAR_AO_OOR(AO);
            KeReleaseSpinLock(&AO->ao_lock, Irql1);
        }
    }

    NdisFreePacket(Packet);
    FreeDGSendReq(FinishedSR);
    KeReleaseSpinLock(&DGSendReqLock, Irql0);
    if (Callback != NULL)
        (*Callback)(CallbackContext, TDI_SUCCESS, (uint)SentSize);
}


//
// NT supports cancellation of DG send/receive requests.
//

#define TCP_DEBUG_SEND_DGRAM    0x00000100
#define TCP_DEBUG_RECEIVE_DGRAM 0x00000200

extern ULONG TCPDebug;


VOID
TdiCancelSendDatagram(
    AddrObj *SrcAO,
    PVOID Context)
{
    KIRQL OldIrql;
    DGSendReq *sendReq = NULL;
    Queue *qentry;
    BOOLEAN found = FALSE;

    CHECK_STRUCT(SrcAO, ao);

    KeAcquireSpinLock(&SrcAO->ao_lock, &OldIrql);

    // Search the send list for the specified request.
    for (qentry = QNEXT(&(SrcAO->ao_sendq)); qentry != &(SrcAO->ao_sendq);
         qentry = QNEXT(qentry)) {

        sendReq = CONTAINING_RECORD(qentry, DGSendReq, dsr_q);

        CHECK_STRUCT(sendReq, dsr);

        if (sendReq->dsr_context == Context) {
            //
            // Found it.  Dequeue.
            //
            REMOVEQ(qentry);
            found = TRUE;

            IF_TCPDBG(TCP_DEBUG_SEND_DGRAM) {
                KdPrint(("TdiCancelSendDatagram: Dequeued item %lx\n",
                         Context));
            }

            break;
        }
    }

    KeReleaseSpinLock(&SrcAO->ao_lock, OldIrql);

    if (found) {
        //
        // Complete the request and free its resources.
        //
        (*sendReq->dsr_rtn)(sendReq->dsr_context, (uint) TDI_CANCELLED, 0);

        KeAcquireSpinLock(&DGSendReqLock, &OldIrql);

        if (sendReq->dsr_header != NULL) {
            FreeDGHeader(sendReq->dsr_header);
        }

        FreeDGSendReq(sendReq);

        KeReleaseSpinLock(&DGSendReqLock, OldIrql);
    }

} // TdiCancelSendDatagram


VOID
TdiCancelReceiveDatagram(
    AddrObj *SrcAO,
    PVOID Context)
{
    KIRQL OldIrql;
    DGRcvReq *rcvReq = NULL;
    Queue *qentry;
    BOOLEAN found = FALSE;

    CHECK_STRUCT(SrcAO, ao);

    KeAcquireSpinLock(&SrcAO->ao_lock, &OldIrql);

    // Search the send list for the specified request.
    for (qentry = QNEXT(&(SrcAO->ao_rcvq)); qentry != &(SrcAO->ao_rcvq);
         qentry = QNEXT(qentry)) {

        rcvReq = CONTAINING_RECORD(qentry, DGRcvReq, drr_q);

        CHECK_STRUCT(rcvReq, drr);

        if (rcvReq->drr_context == Context) {
            //
            // Found it.  Dequeue.
            //
            REMOVEQ(qentry);
            found = TRUE;

            IF_TCPDBG(TCP_DEBUG_SEND_DGRAM) {
                KdPrint(("TdiCancelReceiveDatagram: Dequeued item %lx\n",
                         Context));
            }

            break;
        }
    }

    KeReleaseSpinLock(&SrcAO->ao_lock, OldIrql);

    if (found) {
        //
        // Complete the request and free its resources.
        //
        (*rcvReq->drr_rtn)(rcvReq->drr_context, (uint) TDI_CANCELLED, 0);

        FreeDGRcvReq(rcvReq);
    }

} // TdiCancelReceiveDatagram


//* TdiSendDatagram - TDI send datagram function.
//
//  This is the user interface to the send datagram function.  The caller
//  specified a request structure, a connection info structure containing
//  the address, and data to be sent.  This routine gets a DG Send request
//  structure to manage the send, fills the structure in, and calls DGSend
//  to deal with it.
//
TDI_STATUS  // Returns: Status of attempt to send.
TdiSendDatagram(
    PTDI_REQUEST Request,                  // Request structure.
    PTDI_CONNECTION_INFORMATION ConnInfo,  // ConnInfo for remote address.
    uint DataSize,                         // Size in bytes of data to be sent.
    uint *BytesSent,                       // Return location for bytes sent.
    PNDIS_BUFFER Buffer)                   // Buffer chain to send.
{
    AddrObj *SrcAO;          // Pointer to AddrObj for src.
    DGSendReq *SendReq;      // Pointer to send req for this request.
    KIRQL Irql0, Irql1;      // One per lock nesting level.
    TDI_STATUS ReturnValue;
    DGSendProc SendProc;

    // First, get a send request.  We do this first because of MP issues
    // if we port this to NT.  We need to take the SendRequest lock before
    // we take the AddrObj lock, to prevent deadlock and also because
    // GetDGSendReq might yield, and the state of the AddrObj might
    // change on us, so we don't want to yield after we've validated it.

    KeAcquireSpinLock(&DGSendReqLock, &Irql0);
    SendReq = GetDGSendReq();

    // Now get the lock on the AO, and make sure it's valid.  We do this
    // to make sure we return the correct error code.

    SrcAO = Request->Handle.AddressHandle;

    CHECK_STRUCT(SrcAO, ao);

    KeAcquireSpinLock(&SrcAO->ao_lock, &Irql1);

    if (AO_VALID(SrcAO)) {

        // Make sure the size is reasonable.
        if (DataSize <= SrcAO->ao_maxdgsize)  {
            //
            // The AddrObj is valid.  Now fill the address into the send
            // request if we've got one.  If this works, we'll continue
            // with the send.
            //
            if (SendReq != NULL) {
                // Got a send request.
                if (GetAddress(ConnInfo->RemoteAddress, &SendReq->dsr_addr,
                    &SendReq->dsr_port)) {

                    SendReq->dsr_rtn = Request->RequestNotifyObject;
                    SendReq->dsr_context = Request->RequestContext;
                    SendReq->dsr_buffer = Buffer;
                    SendReq->dsr_size = (ushort)DataSize;

                    // We've filled in the send request. If the AO isn't
                    // already busy, try to get a DG header buffer and send
                    // this. If the AO is busy, or we can't get a buffer, queue
                    // until later. We try to get the header buffer here, as
                    // an optimazation to avoid having to retake the lock.

                    if (!AO_OOR(SrcAO)) {
                        // AO isn't out of resources.

                        if (!AO_BUSY(SrcAO)) {
                            // or busy.

                            if ((SendReq->dsr_header = GetDGHeader())
                                != NULL) {
                                REF_AO(SrcAO);      // Lock out exclusive
                                                    // activities.
                                SendProc = SrcAO->ao_dgsend;

                                KeReleaseSpinLock(&SrcAO->ao_lock, Irql1);
                                KeReleaseSpinLock(&DGSendReqLock, Irql0);

                                // Allright, just send it.
                                (*SendProc)(SrcAO, SendReq);

                                // See if any pending requests occured during
                                // the send. If so, call the request handler.
                                DEREF_AO(SrcAO);

                                return TDI_PENDING;
                            } else {
                                // We couldn't get a header buffer.  Put this
                                // guy on the pending queue, and then fall
                                // through to the 'queue request' code.
                                PutPendingQ(SrcAO);
                            }
                        } else {
                            // AO is busy, set request for later.
                            SET_AO_REQUEST(SrcAO, AO_SEND);
                        }
                    }

                    // AO is busy, or out of resources.  Queue the send request
                    // for later.
                    SendReq->dsr_header = NULL;
                    ENQUEUE(&SrcAO->ao_sendq, &SendReq->dsr_q);
                    SendReq = NULL;
                    ReturnValue = TDI_PENDING;
                } else {
                    // The remote address was invalid.
                    ReturnValue = TDI_BAD_ADDR;
                }
            } else {
                // Send request was null, return no resources.
                ReturnValue = TDI_NO_RESOURCES;
            }
        } else {
            // Buffer was too big, return an error.
            ReturnValue = TDI_BUFFER_TOO_BIG;
        }
    } else {
        // The addr object is invalid, possibly because it's deleting.
        ReturnValue = TDI_ADDR_INVALID;
    }

    KeReleaseSpinLock(&SrcAO->ao_lock, Irql1);

    if (SendReq != NULL)
        FreeDGSendReq(SendReq);

    KeReleaseSpinLock(&DGSendReqLock, Irql0);

    return TDI_ADDR_INVALID;  // BUGBUG? Why not return ReturnValue?
}

//* TdiReceiveDatagram - TDI receive datagram function.
//
//  This is the user interface to the receive datagram function.  The
//  caller specifies a request structure, a connection info structure
//  that acts as a filter on acceptable datagrams, a connection info
//  structure to be filled in, and other parameters.  We get a DGRcvReq
//  structure, fill it in, and hang it on the AddrObj, where it will be
//  removed later by the incoming datagram handler.
//
TDI_STATUS  // Returns: Status of attempt to receive.
TdiReceiveDatagram(
    PTDI_REQUEST Request,
    PTDI_CONNECTION_INFORMATION ConnInfo,    // ConnInfo for remote address.
    PTDI_CONNECTION_INFORMATION ReturnInfo,  // ConnInfo to be filled in.
    uint RcvSize,                            // Total size in bytes of Buffer.
    uint *BytesRcvd,                         // Where to return bytes recv'd.
    PNDIS_BUFFER Buffer)                     // Buffer chain to fill in.
{
    AddrObj *RcvAO;          // AddrObj that is receiving.
    DGRcvReq *RcvReq;        // Receive request structure.
    KIRQL OldIrql;
    uchar AddrValid;

    RcvReq = GetDGRcvReq();
    RcvAO = Request->Handle.AddressHandle;
    CHECK_STRUCT(RcvAO, ao);

    KeAcquireSpinLock(&RcvAO->ao_lock, &OldIrql);
    if (AO_VALID(RcvAO)) {

        IF_TCPDBG(TCP_DEBUG_RAW) {
            KdPrint(("posting receive on AO %lx\n", RcvAO));
        }

        if (RcvReq != NULL) {
            if (ConnInfo != NULL && ConnInfo->RemoteAddressLength != 0) {
                AddrValid = GetAddress(ConnInfo->RemoteAddress,
                                       &RcvReq->drr_addr, &RcvReq->drr_port);
            } else {
                AddrValid = TRUE;
                RcvReq->drr_addr = UnspecifiedAddr;
                RcvReq->drr_port = 0;
            }

            if (AddrValid) {
                //
                // Everything is valid.
                // Fill in the receive request and queue it.
                //
                RcvReq->drr_conninfo = ReturnInfo;
                RcvReq->drr_rtn = Request->RequestNotifyObject;
                RcvReq->drr_context = Request->RequestContext;
                RcvReq->drr_buffer = Buffer;
                RcvReq->drr_size = RcvSize;
                ENQUEUE(&RcvAO->ao_rcvq, &RcvReq->drr_q);
                KeReleaseSpinLock(&RcvAO->ao_lock, OldIrql);

                return TDI_PENDING;
            } else {
                // Have an invalid filter address.
                KeReleaseSpinLock(&RcvAO->ao_lock, OldIrql);
                FreeDGRcvReq(RcvReq);
                return TDI_BAD_ADDR;
            }
        } else {
            // Couldn't get a receive request.
            KeReleaseSpinLock(&RcvAO->ao_lock, OldIrql);
            return TDI_NO_RESOURCES;
        }
    } else {
        // The AddrObj isn't valid.
        KeReleaseSpinLock(&RcvAO->ao_lock, OldIrql);
    }

    // The AddrObj is invalid or non-existent.
    if (RcvReq != NULL)
        FreeDGRcvReq(RcvReq);

    return TDI_ADDR_INVALID;
}

#pragma BEGIN_INIT

//* InitDG - Initialize the DG stuff.
//
//  Called during init time to initalize the DG code.  We initialize
//  our locks and request lists.
//
int                      // Returns: True if we succeed, False if we fail.
InitDG(
    uint MaxHeaderSize)  // Maximum size of a datagram transport header.
{
    PNDIS_BUFFER Buffer;
    KIRQL OldIrql;

    DGHeaderSize = MaxHeaderSize + sizeof(IPv6Header);

    KeInitializeSpinLock(&DGSendReqLock);
    KeInitializeSpinLock(&DGRcvReqFreeLock);

    DGSendReqFree = NULL;

    ExInitializeSListHead(&DGRcvReqFree);

    KeAcquireSpinLock(&DGSendReqLock, &OldIrql);

    Buffer = GrowDGHeaderList();

    if (Buffer != NULL) {
        FreeDGHeader(Buffer);
        KeReleaseSpinLock(&DGSendReqLock, OldIrql);
    } else {
        KeReleaseSpinLock(&DGSendReqLock, OldIrql);
        return FALSE;
    }

    INITQ(&DGHeaderPending);
    INITQ(&DGDelayed);

    //
    // Prepare a work-queue item which we may later enqueue for a system
    // worker thread to handle.  Here we associate our callback routine
    // (DGDelayedWorker) with the work item.
    // 
    ExInitializeWorkItem(&DGDelayedWorkItem, DGDelayedWorker, NULL);

    return TRUE;
}

#pragma END_INIT
