// -*- mode: C++; tab-width: 4; indent-tabs-mode: nil -*- (for GNU Emacs)
//
// Copyright (c) 1985-1998 Microsoft Corporation
//
// This file is part of the Microsoft Research IPv6 Network Protocol Stack.
// You should have received a copy of the Microsoft End-User License Agreement
// for this software along with this release; see the file "license.txt".
// If not, please see http://www.research.microsoft.com/msripv6/license.htm,
// or write to Microsoft Research, One Microsoft Way, Redmond, WA 98052-6399.
//
// Abstract:
//
// Test tool for sending different types raw IPv6 packets.
//

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <sys/types.h>
#include <winsock2.h>
#include <ws2ip6.h>

#include <testsend.h>

#define IP_VERSION 0x00000060

static ulong Id = 0;  // used for fragment ids.

//
// Network byte-order is big-endian.
// NT runs in little-endian mode on all supported architectures.
//
__inline ushort
net_short(ushort x)
{
    return (((x & 0xff) << 8) | ((x & 0xff00) >> 8));
}


ProtoTblStruct ProtocolTbl[] = {
#if 0
    { "icmp", IP_PROTOCOL_ICMPv6, CreateICMPMsg },
#endif
    { "ip", IP_PROTOCOL_NONE, CreateIPMsg },
    { NULL, 0, NULL }
};

MesgTypeTblStruct MesgTypeTbl[] = {
    { "frag", "fragment", FRAG, IP_PROTOCOL_FRAGMENT },
    { "frag0", "fragment0", FRAG0, IP_PROTOCOL_FRAGMENT }, // last fragment.
    { "hop", "hopbyhop", HOP, IP_PROTOCOL_HOP_BY_HOP },
    { "dest", "destopts", DEST, IP_PROTOCOL_DEST_OPTS },
    { "route", "routing", RTG, IP_PROTOCOL_ROUTING },
    { NULL, NULL, 0, 0 }
};

ErrorTypeTblStruct ErrorTypeTbl[] = {
    { "param", "parameter_problem", PARAM_PROB },
    { "unreach", "dest_unreachable", DEST_UNREACH },
    { "time", "time_exceeded", TIME_EXCEED },
    { "big", "too_big", TOO_BIG },
    { NULL, NULL, 0 }
};

void _CRTAPI1
main(int argc, char **argv)
{
    int Error, i;
    WSADATA WsaData;
    SOCKET Sock;
	char TempBuf[1000];
	IPv6Header *IP;
	int Count, ch, HdrLen = 0;
	uchar HopLimit, IPProto;
    struct sockaddr_in6 Us, Peer;
    short Port;
    char *Host;
    char *arg, *strarg;
    char *proto = NULL;
    char *mesg_type = NULL;
    char *error_type = NULL;
    ushort TypeFlag = 0, ErrorFlag = 0;
    ProtoTblStruct *ProtoEntry;

	Count = 29;                  // Amount of raw data to send after IPv6 hdr.
	Port = 4000;                 // Arbitrary (and meaningless).
	HopLimit = 10;               // Arbitrary.
	IPProto = IP_PROTOCOL_NONE;  // The "No Next Header" type.
    
    
    if (argc <= 2) usage();
    
    //
    // Process command line.
    //
    for (i=1; i < argc; i++) {
        arg = argv[i];

        if (arg[0] == '-') {
            switch(arg[1]) {
            case '?':
                usage();
                goto error_exit;

            case 'p':
                if (!param(&proto, argv, argc, i)) {
                    goto error_exit;
                }
                break;

            case 't':
                if (!param(&mesg_type, argv, argc, i)) {
                    goto error_exit;
                }
                if (!verify_type(mesg_type, &TypeFlag)) {
                    goto error_exit;
                }
                break;
            case 'e':
                if (!param(&error_type, argv, argc, i)) {
                    goto error_exit;
                }
                if (!verify_error(error_type, &ErrorFlag)) {
                    goto error_exit;
                }
                break;

#if 0
            case 's':
                // potential option for source addr.
                break;
#endif
            default:
                printf("%s is not a valid command option.\n", argv[i]);
                usage();
                break;

            }
        }
    }
    
    //
    // Create the specific protocol headers.
    //
    if (proto) {
        if (!(ProtoEntry = GetProtoFunc(proto))) {
            printf("testsend: %s is an unsupported protocol.\n", proto);
            usage();
        }
        IPProto = (*ProtoEntry->cfunc)(TempBuf + IPv6_HDR_LEN, 
                                       TypeFlag, ErrorFlag, &HdrLen);
    } else {
        if (ErrorFlag & PARAM_PROB) {
            IPProto = IP_PROTOCOL_BOGUS;
        }
    }

    
    //
    // Initialize winsock.
    //
    Error = WSAStartup(0x0101, &WsaData);
    if (Error == SOCKET_ERROR) {
        printf("testsend: WSAStartup failed %ld:", WSAGetLastError());
    }

    //
    // Establish our peer's address.
    //
	memset((char *)&Peer, 0, sizeof(Peer));
	Host = argv[i-2];     // REVIEW: a little fragile to do this.
	if (!parse_addr(Host, &Peer))
        err("bad hostname");
	Peer.sin6_port = htons(Port);
	
    //
    // Establish our address.
    //
	memset((char *)&Us, 0, sizeof(Us));
	Host = argv[i-1];
	if (!parse_addr(Host, &Us))
        err("bad hostname");
	Us.sin6_port = 0;
	Us.sin6_family = AF_INET6;

    //
    // Create a raw IPv6 socket.
    //
	Sock = socket(PF_INET6, SOCK_RAW, 0);
	if (Sock == INVALID_SOCKET) {
        printf("socket() failed: %ld\n", GetLastError( ) );
        exit(1);
	}

    //
    // Bind our source address to the socket.
    // For raw send, this has the effect of picking an outgoing interface.
    //
    if (bind(Sock, (PSOCKADDR)&Us, sizeof(Us)) < 0)
        err("bind");

    //
    // Put an IPv6 header at the beginning of our packet buffer.
    //
    IP = (IPv6Header *)((uchar *)TempBuf);
    IP->VersClassFlow = IP_VERSION;
    IP->NextHeader = IPProto;
    IP->HopLimit = HopLimit;
    IP->Source = Us.sin6_addr;
    IP->Dest = Peer.sin6_addr;
    IP->PayloadLength = net_short((ushort)(Count+HdrLen));

    //
    // Send the packet.
    //
	if (sendto(Sock, (char *)TempBuf, sizeof(IPv6Header) + Count + HdrLen, 0,
               (PSOCKADDR)&Peer, sizeof(Peer)) < 0) {
        printf("sendto() failed: %ld\n", GetLastError( ) );
        exit(1);
    }

    //
    // Cleanup.
    //
    closesocket(Sock);
    WSACleanup();
    exit(0);


  error_exit:
    printf("testsend: invalid option argument.\n");
    usage();
}



void
usage()
{
	printf("Usage: testsend [-p protocol] [-t mesg_type] [-e error_type] dest src\n");
    printf("\n       where protocol   = {ip}\n");
    printf("             mesg_type  = {frag | hop | dest}\n");
    printf("             error_type = {param}\n");
    
    exit(1);
}

//
// Print an error message.
//
void err(s)
char *s;
{
        fprintf(stderr,"testsend: ");
        perror(s);
        fprintf(stderr,"errno=%ld\n", WSAGetLastError( ));
        WSACleanup();
        exit(1);
}


//
// Does not currently support DNS names, but should.
// (If WSAStringToAddress fails, then try DNS.)
//
int
parse_addr(char *s, struct sockaddr_in6 *sin6)
{
    struct hostent *h;
    
    sin6->sin6_family = AF_INET6;
    
    if (!inet6_addr(s, &sin6->sin6_addr)) {
        // Presumably the user gave us a DNS name.

        h = getnodebyname(s, AF_INET6, AI_DEFAULT);
        if (h == NULL)
            return FALSE;

        memcpy(&sin6->sin6_addr, h->h_addr, sizeof(struct in6_addr));
        freehostent(h);
    }

    return TRUE;
}

//* CreateICMPMsg
uint
CreateIPMsg(uchar *TheBuf, ushort TypeFlag, ushort ErrorType, int *Len)
{
    uint proto = 1;         // Value of 1 means protocol not specified.
    uchar *Buf = TheBuf;
    uchar *tmp_hdr = NULL;
    
    if (TypeFlag & HOP) {
        ExtensionHeaderPadded *hdr;

        // Create a minimum length, no-content, HopByHop Header.
        hdr = (ExtensionHeaderPadded*) Buf;        
        hdr->NextHeader = IP_PROTOCOL_NONE;
        hdr->HeaderExtLength = 0;  // Length is the min 8 bytes.
        hdr->Pad0 = 0;     
        hdr->Pad1 = 0;
        hdr->Pad2 = 0;
        
        // Save pointer to this header's next header value.
        tmp_hdr = Buf;

        // Increment buffer pointer.
        Buf += sizeof(ExtensionHeaderPadded);
        *Len += sizeof(ExtensionHeaderPadded);
        
        // Overwrite previous and make this the IP->NextHeader.
        proto = IP_PROTOCOL_HOP_BY_HOP;
    } 
    if (TypeFlag & DEST) {
        ExtensionHeaderPadded *hdr;

        // Create a minimum length, no-content, Destination Options Header.
        hdr = (ExtensionHeaderPadded*) Buf;        
        hdr->NextHeader = IP_PROTOCOL_NONE;
        hdr->HeaderExtLength = 0;  // Length is the min 8 bytes.
        hdr->Pad0 = 0;     // just padding.
        hdr->Pad1 = 0;
        hdr->Pad2 = 0;

        // Set DEST_OPT as next header in the previous header.
        if (tmp_hdr) {
            *tmp_hdr = IP_PROTOCOL_DEST_OPTS;
        }
        tmp_hdr = Buf;

        // Increment buffer pointer.
        Buf += sizeof(ExtensionHeaderPadded);
        *Len += sizeof(ExtensionHeaderPadded);


        // If protocol not yet set, make this the IP->NextHeader.
        if (proto == 1) {
            proto = IP_PROTOCOL_DEST_OPTS;
        }
    } 
    if ((TypeFlag & FRAG) || (TypeFlag & FRAG0)) {
        FragmentHeader *hdr;
        
        hdr = (FragmentHeader*) Buf;        
        hdr->NextHeader = IP_PROTOCOL_NONE;
        hdr->Reserved = 0;
        if (TypeFlag & FRAG) {
            hdr->OffsetFlag = net_short(1);   // indicate more fragments.
        } else {
            hdr->OffsetFlag = 0;              // indicate last fragment.
        }
        hdr->Id = Id++;

        // Set last header's next header pointer to this header type.
        if (tmp_hdr) {
            *tmp_hdr = IP_PROTOCOL_FRAGMENT;
        }
        tmp_hdr = Buf;

        // If protocol not yet set, make this the IP->NextHeader.
        if (proto == 1) {
            proto = IP_PROTOCOL_FRAGMENT;
        }

        *Len += sizeof(FragmentHeader);

        // We don't increment the buffer ptr here, leave it pointing
        // to the beginning of the Fragment header.
    } 


    // Introduce error into message.
    //
    // REVIEW: Currently, only paramenter problem errors are generated.
    //
    if (ErrorType & PARAM_PROB) {
        *tmp_hdr = (uchar) IP_PROTOCOL_BOGUS;
    }

    return proto;
}

//* CreateICMPMsg
uint
CreateICMPMsg(uchar *Buf, ushort TypeFlag, ushort ErrorType, int *Len)
{
    ICMPv6Header Hdr;
    uint proto;

#if 0    
    // Just create ICMP error message.
    Hdr = (ICMPv6Header *)((uchar *)TempBuf);
#endif

    return IP_PROTOCOL_NONE;
}


// * GetProtoFunc
//
ProtoTblStruct *GetProtoFunc(char *proto_name)
{
    ProtoTblStruct *p;

    for(p = ProtocolTbl; p->Name; p++) {
        //printf("getprotofunc: %s %s\n", proto_name, p->Name);
        if (strcmp(proto_name, p->Name) == 0)
            return  p;
    }
    return NULL;
}

BOOLEAN
param(
    char **strarg,
    char **argv,
    int argc,
    int current)
{
    if (current == (argc - 1)) {
        printf("A value must be supplied for option %s.\n", argv[current]);
        return(FALSE);
    }
    *strarg = argv[current+1];
    //printf("param: %s \n", *strarg);
    return TRUE;
}

BOOLEAN verify_type(char *type, ushort *Flag)
{
    MesgTypeTblStruct *m;
    
    for(m = MesgTypeTbl; m->Name0; m++) {
        if ((strcmp(type, m->Name0) == 0) || (strcmp(type, m->Name1) == 0)) {
            *Flag |= m->Value;
            return TRUE;
        }
    }
    return FALSE;
}

BOOLEAN verify_error(char *type, ushort *Flag)
{
    ErrorTypeTblStruct *m;
    
    for(m = ErrorTypeTbl; m->Name0; m++) {
        if ((strcmp(type, m->Name0) == 0) || (strcmp(type, m->Name1) == 0)) {
            *Flag |= m->Value;
            return TRUE;
        }
    }
    return FALSE;
}
