// -*- mode: C++; tab-width: 4; indent-tabs-mode: nil -*- (for GNU Emacs)
//
// Copyright (c) 1985-1998 Microsoft Corporation
//
// This file is part of the Microsoft Research IPv6 Network Protocol Stack.
// You should have received a copy of the Microsoft End-User License Agreement
// for this software along with this release; see the file "license.txt".
// If not, please see http://www.research.microsoft.com/msripv6/license.htm,
// or write to Microsoft Research, One Microsoft Way, Redmond, WA 98052-6399.
//


#include <windows.h>
#include <winsock2.h>
#include <ws2ip6.h>
#include <svcguid.h>

#define T_AAAA  28
#define C_IN    1

#define MAX_ADDRS       64      // The maximum we can return.

#define HOSTENT_LEN (sizeof(struct hostent) +                           \
                     (MAX_ADDRS * (sizeof(struct in6_addr) +            \
                                   sizeof(struct in6_addr *))) +        \
                     sizeof(struct in6_addr *))

static int
ParseReply(uchar *data, uint size, struct in6_addr **paddrs, uint *psize);

//* getnodebyname
//
//  A wrapper around WSALookupServiceBegin/Next/End,
//  that is like gethostbyname but returns AF_INET6 addresses.
//
//  The NT4 DNS name space resolver (rnr20.dll) does not
//  cache replies when you request a specific RR type.
//  This means that every call to getnodebyname
//  results in DNS message traffic. There is no caching!
//
struct hostent * WSAAPI
getnodebyname(const char *name, int af, int flags)
{
    // The TCP port number is ignored.
    static GUID guid = SVCID_TCP_RR(0, T_AAAA);

    char buffer[sizeof(WSAQUERYSETA) + 2048];
    uint Size;
    PWSAQUERYSETA query = (PWSAQUERYSETA) buffer;
    HANDLE handle;
    struct hostent *h;
    uint hsize;
    struct in6_addr *addrs;
    struct in6_addr **array;
    int err;
    int i;

    //
    // The current implementation doesn't support any options.
    //
    if ((af != AF_INET6) || (flags != AI_DEFAULT))
        return NULL;

    //
    // Build a Winsock DNS query.
    // The guid specifies that we want AAAA records.
    //
    memset(query, 0, sizeof *query);
    query->dwSize = sizeof *query;
    query->lpszServiceInstanceName = (char *)name;
    query->lpServiceClassId = &guid;
    query->dwNameSpace = NS_DNS;

    //
    // Initiate the DNS query.
    // We ask for raw DNS replies to be returned.
    // (LUP_RETURN_ADDR does not work because the NT4 DNS
    // name space provider does not understand AAAA records.)
    //
    err = WSALookupServiceBeginA(query,
                                 LUP_RETURN_BLOB,
                                 &handle);
    if (err) {
        return NULL;
    }

    //
    // We return a static hostent structure.
    //
    h = (struct hostent *) LocalAlloc(0, HOSTENT_LEN);
    if (h == NULL)
        return NULL;
    h->h_name = (char *)name;
    h->h_aliases = NULL;
    h->h_addrtype = AF_INET6;
    h->h_length = sizeof(struct in6_addr);

    //
    // hsize tracks the remaining space in the hostent buffer,
    // which is used to return addresses. Each returned address
    // takes sizeof(struct in6_addr) bytes,
    // plus sizeof(struct in6_addr *) bytes for an entry in
    // the h_addr_list array to point to the address.
    // Plus there needs to be a null address to terminate h_addr_list.
    //
    hsize = HOSTENT_LEN - sizeof *h - sizeof(struct in6_addr *);
    addrs = (struct in6_addr *)(h + 1);

    for (;;) {
        Size = sizeof buffer;
        err = WSALookupServiceNextA(handle, 0, &Size, query);
        if (err) {
            err = WSAGetLastError();
            if (err == WSA_E_NO_MORE)
                break;

            (void) WSALookupServiceEnd(handle);
            LocalFree(h);
            return NULL;
        }

        if (query->lpBlob == NULL) {
            (void) WSALookupServiceEnd(handle);
            LocalFree(h);
            return NULL;
        }

        //
        // Parse the DNS reply message, looking for AAAA records.
        // This will increment addrs and decrement hsize appropriately.
        //
        if (! ParseReply(query->lpBlob->pBlobData,
                         query->lpBlob->cbSize,
                         &addrs, &hsize)) {
            (void) WSALookupServiceEnd(handle);
            LocalFree(h);
            return NULL;
        }
    }

    (void) WSALookupServiceEnd(handle);

    //
    // If there were no AAAA records, we must fail.
    //
    if (addrs == (struct in6_addr *)(h + 1)) {
        LocalFree(h);
        return NULL;
    }

    //
    // Construct the h_addr_list array.
    //
    array = (struct in6_addr **) addrs;
    for (i = 0; i < addrs - (struct in6_addr *)(h + 1); i++)
        array[i] = &((struct in6_addr *)(h + 1))[i];
    array[i] = NULL;

    h->h_addr_list = (char **) array;
    return h;
}

void WSAAPI
freehostent(struct hostent *ptr)
{
    LocalFree(ptr);
}

static int
ParseReply(uchar *data, uint size,
           struct in6_addr **paddrs, uint *psize)
{
    ushort id, codes, qdcount, ancount, nscount, arcount;
    struct in6_addr *addr;

    //
    // The DNS message starts with six two-byte fields.
    //
    if (size < sizeof(ushort) * 6)
        return FALSE;

    id = ntohs(((ushort * UNALIGNED)data)[0]);
    codes = ntohs(((ushort * UNALIGNED)data)[1]);
    qdcount = ntohs(((ushort * UNALIGNED)data)[2]);
    ancount = ntohs(((ushort * UNALIGNED)data)[3]);
    nscount = ntohs(((ushort * UNALIGNED)data)[4]);
    arcount = ntohs(((ushort * UNALIGNED)data)[5]);

    data += sizeof(ushort) * 6;
    size -= sizeof(ushort) * 6;

    //
    // Skip over the question records.
    // Each question record has a QNAME, a QTYPE, and a QCLASS.
    // The QNAME is a sequence of labels, where each label
    // has a length byte. It is terminated by a zero-length label.
    // The QTYPE and QCLASS are two bytes each.
    //
    while (qdcount > 0) {

        //
        // Skip over the QNAME labels.
        //
        for (;;) {
            uchar length;

            if (size < sizeof length)
                return FALSE;
            length = *data;
            if ((length & 0xc) == 0xc) {
                //
                // This is a pointer to labels elsewhere.
                //
                if (size < sizeof(ushort))
                    return FALSE;
                data += sizeof(ushort);
                size -= sizeof(ushort);
                break;
            }

            data += sizeof length;
            size -= sizeof length;

            if (length == 0)
                break;

            if (size < length)
                return FALSE;
            data += length;
            size -= length;
        }

        //
        // Skip over QTYPE and QCLASS.
        //
        if (size < sizeof(ushort) * 2)
            return FALSE;
        data += sizeof(ushort) * 2;
        size -= sizeof(ushort) * 2;

        qdcount--;
    }

    //
    // Examine the answer records, looking for AAAA records.
    // Each answer record has a name, followed by several values:
    // TYPE, CLASS, TTL, RDLENGTH, and then RDLENGTH bytes of data.
    // TYPE, CLASS, and RDLENGTH are two bytes; TTL is four bytes.
    //
    while (ancount > 0) {
        ushort type, class, rdlength;
        uint ttl;

        //
        // Skip over the name.
        //
        for (;;) {
            uchar length;

            if (size < sizeof length)
                return FALSE;
            length = *data;
            if ((length & 0xc0) == 0xc0) {
                //
                // This is a pointer to labels elsewhere.
                //
                if (size < sizeof(ushort))
                    return FALSE;
                data += sizeof(ushort);
                size -= sizeof(ushort);
                break;
            }
            data += sizeof length;
            size -= sizeof length;

            if (length == 0)
                break;

            if (size < length)
                return FALSE;
            data += length;
            size -= length;
        }

        if (size < sizeof(ushort) * 3 + sizeof(uint))
            return FALSE;
        type = ntohs(((ushort * UNALIGNED)data)[0]);
        class = ntohs(((ushort * UNALIGNED)data)[1]);
        ttl = ntohl(((uint * UNALIGNED)data)[1]);
        rdlength = ntohs(((ushort * UNALIGNED)data)[4]);

        data += sizeof(ushort) * 3 + sizeof(uint);
        size -= sizeof(ushort) * 3 + sizeof(uint);

        addr = (struct in6_addr *)data;

        if (size < rdlength)
            return FALSE;

        data += rdlength;
        size -= rdlength;

        if ((type == T_AAAA) &&
            (class == C_IN) &&
            (rdlength == sizeof(struct in6_addr))) {
            //
            // We have found a valid AAAA record.
            // Save the address.
            //
            if (*psize < sizeof(struct in6_addr) + sizeof(struct in6_addr *))
                return FALSE;

            memcpy(*paddrs, addr, sizeof(struct in6_addr));
            (*paddrs)++;
            *psize -= sizeof(struct in6_addr) + sizeof(struct in6_addr *);
        }

        ancount--;
    }

    return TRUE;
}
